/*
 *	bsyscalls.c
 *	Blocking system calls
 *	Copyright (C) 2000 F.R.M. Barnes
 *
 *	This program is free software; you can redistribute it and/or modify
 *	it under the terms of the GNU General Public License as published by
 *	the Free Software Foundation; either version 2 of the License, or
 *	(at your option) any later version.
 *
 *	This program is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU General Public License for more details.
 *
 *	You should have received a copy of the GNU General Public License
 *	along with this program; if not, write to the Free Software
 *	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */


#ifdef BLOCKING_SYSCALLS
#define __BSYSCALLS_C

#include <stdio.h>
#include <string.h>
/* #include <stdlib.h> */
#include <stdarg.h>
#include <setjmp.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/mman.h>
#include <sys/types.h>
#include <sched.h>
#include <malloc.h>
#include <sys/ipc.h>
#include <sys/sem.h>
#include <signal.h>
#include <errno.h>

#include "sched_types.h"
#include "sched_consts.h"
#include "sched.h"
#include "sched_asm_inserts.h"
#include "occam_io.h"
#include "queue_defs.h"

#include "bsyscalls.h"
#include "atomics.h"
#include "rts.h"

/* Make sure we have a nice definition of semaphore control.. */
#if defined(__GNU_LIBRARY__) && !defined(_SEM_SEMUN_UNDEFINED)
	/* union semun is defined in sys/sem.h */
#else
	union semun {
		int val;
		struct semid_ds *buf;
		unsigned short int *array;
		struct seminfo *__buf;
	};
#endif

/* Variables */
static volatile bsc_thread	dispatching;
static volatile bsc_thread	*v_dispatching;
static spl_t			dispatch_lock;
static volatile int		clone_pids[MAX_NUM_CLONES];
static volatile bsc_thread	*volatile clone_arry[MAX_NUM_CLONES];
static int			num_clones = 0;
static int			clone_sem;
static int			occam_proc_pid;

/* one semaphore in the set 'clone_sem' */
#define NOTIFY_SEM 0		/* to notify a clone */


spl_t			queue_lock;
volatile word		*volatile q_fptr, *volatile q_bptr;
int			num_spare;
int			num_queued;

/* Prototypes */

static int create_clone (void);
static int clone_entry (void *arg);
static int set_semaphores (int sem_id, int n_sems, ...);
static int claim_semaphore (int semid, int semnum);
static int safe_claim_semaphore (int semid, int semnum);
static int notify_semaphore (int semid, int semnum);
static void clone_bad_signal (int sig);
static void clone_vbad_signal (int sig);
static void clone_abort (int sig);

#define CLONE_FSTR	"P-%d,T-%d,S-%d,WS-%p,R-%p,A-%p,F-%p"
#define CLONE_PARAMS(X)	X->pid, X->thr_num, X->par_sync, X->ws_ptr, X->raddr, X->ws_arg, X->func

/*
 *	int bsyscalls_create_clones (void)
 *	Creates the clones (and sets them off)
 *	Returns 0 on success, -1 on failure
 */
int bsyscalls_create_clones (void)
{
	int			x;

	spl_init (&dispatch_lock);
	spl_init (&queue_lock);
	spl_lock (&queue_lock);
	num_spare = 0;
	q_fptr = q_bptr = NULL;
	spl_unlock (&queue_lock);
	v_dispatching = &dispatching;

	occam_proc_pid = getpid ();
	clone_sem = semget (IPC_PRIVATE, 2, 0600);
	if (clone_sem < 0) {
		fprintf (stderr, "KRoC: bsyscalls: unable to create semaphores: %s\n", strerror(errno));
		return -1;
	}
	if (set_semaphores (clone_sem, 1, 0, 0)) {
		fprintf (stderr, "KRoC: bsyscalls: unable to set semaphore values.\n");
		return -1;
	}
	for (x=0; x<NUM_CLONES; x++) {
		if (create_clone ()) {
			return -1;
		}
	}
	return 0;
}


/*
 *	void bsyscalls_destroy_clones (void)
 *	Destroyes the clones (sends a SIGTERM)
 */
void bsyscalls_destroy_clones (void)
{
	int x;
	union semun dummy;

	/* Sanity ? -- we could get in here by a signal which causes entry to somewhere in the RTS */
	if (getpid() != occam_proc_pid) {
		_exit (0);
	}
	/* This should release processes blocking on the semaphore */
	if (semctl (clone_sem, 0, IPC_RMID, dummy) < 0) {
		#ifdef BLOCKING_DEBUG
			fprintf (stderr, "KRoC: bsyscalls (%d): unable to destroy semaphore set: %s\n", getpid(), strerror(errno));
		#endif
	}
	/* Go and kill the others */
	for (x=0; x<num_clones; x++) {
		#ifdef BLOCKING_DEBUG
			fprintf (stderr, "KRoC: bsyscalls (%d): destroying clone %d (pid=%d)\n", getpid(), x, clone_pids[x]);
		#endif
		if (clone_pids[x] > 0) {
			kill (clone_pids[x], SIGKILL);
		}
	}
	return;
}


/*
 *	int create_clone (void)
 *	Creates a clone process, sets it up and runs into the occam kernel
 */
static int create_clone (void)
{
	void *c_stack;
	volatile bsc_thread *new;

	if (num_clones == MAX_NUM_CLONES) {
		return -1;
	}
	/* c_stack = mmap ((void *)0, CLONE_STACK_SIZE, PROT_READ|PROT_WRITE, MAP_ANONYMOUS|MAP_SHARED, -1, 0); */
	/* ... as to why mmap() didn't work, I have no idea .. */
	c_stack = (void *)malloc (CLONE_STACK_SIZE);
	if (!c_stack) {
		fprintf (stderr, "KRoC: bsyscalls: unable to allocate %d bytes for clone stack: %s\n", CLONE_STACK_SIZE, strerror(errno));
		return -1;
	}
	new = (bsc_thread *)c_stack;
	new->pid = 0;
	new->thr_num = num_clones;
	new->terminated = 0;
	new->cancel = 0;
	new->jbuf = (sigjmp_buf *)(c_stack + sizeof(bsc_thread) + 16);
	if ((int)(new->jbuf) % sizeof(void *)) {
		(int)(new->jbuf) -= ((int)(new->jbuf) % sizeof(void *));
	}
	new->user_ptr = (void *)((char *)(new->jbuf) + sizeof(sigjmp_buf) + 16);
	if ((int)(new->user_ptr) % sizeof(void *)) {
		(int)(new->user_ptr) -= ((int)(new->user_ptr) % sizeof(void *));
	}
	new->clone_cleanup = NULL;
	clone_arry[num_clones] = new;
	clone_pids[num_clones] = -1;
	barrier ();
	if (__clone (clone_entry, (c_stack + (CLONE_STACK_SIZE - 4)), CLONE_VM|CLONE_FS|CLONE_FILES,
		(void *)new) == -1) {
		fprintf (stderr, "KRoC: bsyscalls: calling __clone() failed because: %s\n", strerror(errno));
		return -1;
	}
	barrier ();
	spl_lock (&queue_lock);
	num_clones++;
	num_spare++;
	spl_unlock (&queue_lock);
	barrier ();
	return 0;
}


/*
 *	void clone_abort (int sig)
 *	SIGUSR1 handler for clone
 */
static void clone_abort (int sig)
{
	int n_cln, x, me;
	volatile bsc_thread *volatile tmp;

	n_cln = num_clones;
	me = getpid ();
	for (x=0; x<n_cln; x++) {
		if (clone_pids[x] == me) {
			break;
		}
	}
	if (x == n_cln) {
		fprintf (stderr, "KRoC: FATAL: clone not found in clone_abort ()\n");
		kill (occam_proc_pid, SIGSEGV);
		pause ();
		_exit (1);
	}
	tmp = clone_arry[x];
	siglongjmp (*(tmp->jbuf), 1);
}


/*
 *	void clone_vbad_signal (int sig)
 *	very bad signal..
 */
static void clone_vbad_signal (int sig)
{
	fprintf (stderr, "KRoC: clone (%d) got bad signal %d\n", getpid (), sig);
	_exit (sig);
}


/*
 *	int clone_entry (void *arg)
 *	Clone enters here -- in fact, it never leaves...
 */
static int		clone_entry (void *arg)
{
	volatile bsc_thread *volatile me;
	int need_sig, *ptr, e_withsig;
	int xflag, do_cleanup;
	volatile int *volatile xptr;
	#ifdef BLOCKING_PROFILE
		long long t_time, x_time;
	#endif
	struct sigaction act, oact;
	sigset_t just_usr1, old_mask;

	ptr = NULL;
	me = (bsc_thread *)arg;
	clone_pids[me->thr_num] = getpid();
	me->pid = getpid ();
	/* Boring signals first */
	signal (SIGTERM, clone_vbad_signal);
	signal (SIGHUP, clone_vbad_signal);
	signal (SIGCHLD, clone_vbad_signal);
	signal (SIGSEGV, clone_vbad_signal);
	signal (SIGBUS, clone_vbad_signal);
	signal (SIGPIPE, clone_vbad_signal);
	signal (SIGUSR2, clone_vbad_signal);
	/* Setup SIGUSR1 (terminate signal) */
	act.sa_handler = clone_abort;
	sigemptyset (&act.sa_mask);
	act.sa_flags = 0;
	if (sigaction (SIGUSR1, &act, &oact) < 0) {
		fprintf (stderr, "KRoC: bsyscalls: clone %d unable to set handler for SIGUSR1: %s\n", me->thr_num, strerror(errno));
		_exit (1);
	}
	/* Block the signal :) */
	sigemptyset (&just_usr1);
	sigaddset (&just_usr1, SIGUSR1);
	sigprocmask (SIG_BLOCK, &just_usr1, &old_mask);
	#ifdef BLOCKING_DEBUG
		fprintf (stderr, "Clone [" CLONE_FSTR "] started.\n", CLONE_PARAMS(me));
	#endif
	/* Go! */
	for (;;) {
		/* Block on semaphore - access dispatching */
		#ifdef BLOCKING_DEBUG
			fprintf (stderr, "Clone [" CLONE_FSTR "] about to wait on semaphore.\n", CLONE_PARAMS(me));
		#endif
		barrier ();
		if (claim_semaphore (clone_sem, NOTIFY_SEM) < 0) {
			_exit (0);
		}
		barrier ();
		#ifdef BLOCKING_DEBUG
			fprintf (stderr, "Clone [" CLONE_FSTR "] woke up (1).\n", CLONE_PARAMS(me));
		#endif
		#ifdef BLOCKING_PROFILE
			x_time = v_dispatching->time_dispatch;
		#endif
		me->ws_ptr = v_dispatching->ws_ptr;
		me->ws_arg = v_dispatching->ws_arg;
		me->func = v_dispatching->func;
		me->raddr = v_dispatching->raddr;
		me->adjustment = v_dispatching->adjustment;
		me->clone_cleanup = NULL;
		/* go! */
		spl_unlock (&dispatch_lock);
		#ifdef BLOCKING_PROFILE
			LOADTIME (t_time);
			fprintf (stderr, "dispatch %d %Ld %Ld\n", me->thr_num, x_time, t_time);
		#endif
		me->cancel = 0;
		me->terminated = 0;
		#ifdef BLOCKING_DEBUG
			fprintf (stderr, "Clone [" CLONE_FSTR "] woke up (2).\n", CLONE_PARAMS(me));
		#endif
		if (me->adjustment) {
			/* Place a pointer to ourselves in the workspace (this is the magic occam channel) */
			ptr = (int *)(me->ws_arg[-1]);
			*ptr = (int)me;
		}
		#ifdef BLOCKING_DEBUG
			fprintf (stderr, "Clone [" CLONE_FSTR "] before barrier()\n", CLONE_PARAMS(me));
		#endif
		barrier ();
		/* Deal with the signal jump stuff... */
		if (!sigsetjmp (*(me->jbuf), 1)) {
			/* We are the first time */
			sigprocmask (SIG_UNBLOCK, &just_usr1, NULL);
			barrier ();
			#ifdef BLOCKING_DEBUG
				fprintf (stderr, "Clone [" CLONE_FSTR "] about to call 0x%x (0x%x)\n", CLONE_PARAMS(me), (unsigned int)me->func, (unsigned int)me->ws_arg);
			#endif
			me->func (me->ws_arg);
			#ifdef BLOCKING_DEBUG
				fprintf (stderr, "Clone [" CLONE_FSTR "] done calling 0x%x (0x%x)\n", CLONE_PARAMS(me), (unsigned int)me->func, (unsigned int)me->ws_arg);
			#endif
			me->terminated = 1;
			sigprocmask (SIG_BLOCK, &just_usr1, NULL);
			e_withsig = 0;
		} else {
			/* We are the interrupted instance -- SIGUSR1 block from siglongjmp :-) */
			me->terminated = 1;
			e_withsig = 1;
		}
		barrier ();
		#ifdef BLOCKING_PROFILE
			LOADTIME (t_time);
		#endif
		do_cleanup = 0;
		/* me->cancel check */
		xflag = 1;
		xptr = &(me->cancel);
		XCHG (xflag, xptr);
		if (xflag) {
			/* cancellation was started */
			if (!e_withsig) {
				#ifdef BLOCKING_DEBUG
					fprintf (stderr, "Clone [" CLONE_FSTR "] hasn\'t seen signal yet -- going into sigsuspend ()\n", CLONE_PARAMS(me));
				#endif
				sigsuspend (&old_mask);
			}
			/* We got it */
			do_cleanup = 1;
			#ifdef BLOCKING_PROFILE
				fprintf (stderr, "killed %d %Ld %Ld\n", me->thr_num, me->time_kill, t_time);
			#endif
		} else {
			#ifdef BLOCKING_PROFILE
				fprintf (stderr, "finish %d %Ld\n", me->thr_num, t_time);
			#endif
		}
		if (me->adjustment) {
			/* Clear channel parameter (in ptr) */
			ptr = (int *)(me->ws_arg[-1]);
			*ptr = 0;
		}
		barrier ();
		/* run cleanup function if necessary */
		if (me->clone_cleanup && do_cleanup) {
			me->clone_cleanup (me->user_ptr);
		}
		/* add to queue */
		#ifdef BLOCKING_DEBUG
			fprintf (stderr, "Clone [" CLONE_FSTR "] adding workspace to queue.\n", CLONE_PARAMS(me));
		#endif
		need_sig = 0;
		barrier ();
		spl_lock (&queue_lock);
		if (!q_fptr) {
			q_fptr = q_bptr = me->ws_ptr;
			need_sig = 1;
		} else {
			q_bptr[Link] = (word)me->ws_ptr;
			q_bptr = me->ws_ptr;
		}
		num_queued++;
		((int *)(me->ws_ptr))[Iptr] = (int)me->raddr;
		barrier ();
		spl_unlock (&queue_lock);
		/* signal occam process */
		#ifdef BLOCKING_DEBUG
			fprintf (stderr, "Clone [" CLONE_FSTR "] signalling occam process.\n", CLONE_PARAMS(me));
		#endif
		if (need_sig) {
			kill (occam_proc_pid, SIGBLOCKING);
		}
		barrier ();
	}
	_exit (0);
}


/*
 *	void _killcall (int w[])
 *	PROC C.killcall (CHAN OF INT c, INT status)
 *	terminates a blocking system call (gulp)
 */
void _killcall (int w[])
{
	int *ptr, *result;
	volatile bsc_thread *volatile tmp;
	int xflag;
	volatile int *volatile xptr;
	#ifdef BLOCKING_PROFILE
		long long t_time;
	#endif

	ptr = (int *)(w[0]);
	result = (int *)(w[1]);
	tmp = (bsc_thread *)(*ptr);
	if (!tmp) {
		/* Huh -- blocking syscall hasn't got here yet... */
		sched_yield ();
		tmp = (bsc_thread *)(*ptr);
		if (!tmp) {
			/* Oh, ok, it already finished. */
			*result = -1;
			return;
		}
	}
	if (tmp->terminated) {
		/* Thread has already terminated */
		*result = 1;
		*ptr = 0;
		return;
	}
	#ifdef BLOCKING_PROFILE
		LOADTIME (t_time);
		tmp->time_kill = t_time;
	#endif
	xflag = 1;
	xptr = &(tmp->cancel);
	XCHG (xflag, xptr);
	if (xflag) {
		/* thread termination already started */
		*result = 2;
		*ptr = 0;
		return;
	}
	/* Signal it.. */
	kill (tmp->pid, SIGUSR1);
	*result = 0;
	*ptr = 0;
	return;
}


/*
 *	void *bsyscalls_set_cleanup (void (*cleanup_fcn)(void *))
 *	sets the cleanup function for a clone
 */
void *bsyscalls_set_cleanup (void (*cleanup_fcn)(void *))
{
	int n_cln, x, me;
	volatile bsc_thread *volatile tmp;

	n_cln = num_clones;
	me = getpid ();
	for (x=0; x<n_cln; x++) {
		if (clone_pids[x] == me) {
			break;
		}
	}
	if (x == n_cln) {
		fprintf (stderr, "KRoC: FATAL: clone not found in ... ()\n");
		kill (occam_proc_pid, SIGSEGV);
		pause ();
		_exit (1);
	}
	tmp = clone_arry[x];
	/* me is something valid */
	tmp->clone_cleanup = cleanup_fcn;
	return tmp->user_ptr;
}


/*
 *	void clone_bad_signal (int sig)
 *	Called when clone gets an unexpected signal
 */
static void		clone_bad_signal (int sig)
{
	fprintf (stderr, "KRoC: bsyscalls: clone got unexpected/unhandled signal %d.\n", sig);
	if (sig == SIGSEGV) {
		kill (occam_proc_pid, SIGSEGV);
		_exit (0);
	}
	signal (sig, clone_bad_signal);
	return;
}


/*
 *	int set_semaphores (int sem_id, int n_sems, ...)
 *	arguments are pairs of [sem_num, val]
 *	Sets semaphores to specified values
 *	Returns 0 on success, -1 on failure.
 */
static int		set_semaphores (int sem_id, int n_sems, ...)
{
	int			sem_num, sem_val;
	union semun		sem_parm;
	va_list			ap;
	int			res;

	va_start (ap, n_sems);
	res = 0;
	while (n_sems) {
		sem_num = va_arg (ap, int);
		sem_val = va_arg (ap, int);
		sem_parm.val = sem_val;
		if (semctl (sem_id, sem_num, SETVAL, sem_parm) < 0) {
			n_sems = 0;
			res = -1;
		}
		n_sems--;
	}
	va_end (ap);
	return res;
}


/*
 *	int claim_semaphore (int semid, int semnum)
 *	Claims a semaphore.  Returns 0 on success (ie, claimed), -1 if an error occured
 */
static int		claim_semaphore (int semid, int semnum)
{
	struct sembuf sem_parm;

	sem_parm.sem_num = semnum;
	sem_parm.sem_flg = 0L;
	sem_parm.sem_op = -1;
	#ifdef BLOCKING_DEBUG
		fprintf (stderr, "bsyscalls: [%d] claim_semaphore (%d, %d)\n", getpid(), semid, semnum);
	#endif
	if (semop (semid, &sem_parm, 1) == -1) {
		#ifdef BLOCKING_DEBUG
			fprintf (stderr, "bsyscalls: [%d] claim_semaphore (%d, %d) failed: %s\n", getpid(), semid, semnum, strerror(errno));
		#endif
		return -1;
	}
	#ifdef BLOCKING_DEBUG
		fprintf (stderr, "bsyscalls: [%d] claim_semaphore (%d, %d) succeeded\n", getpid(), semid, semnum);
	#endif
	return 0;
}


/*
 *	int safe_claim_semaphore (int semid, int semnum)
 *	Claims a semaphore (verbose failing)
 */
static int safe_claim_semaphore (int semid, int semnum)
{
	struct sembuf sem_parm;
	union semun sem_union;
	int retval;

	sem_parm.sem_num = semnum;
	sem_parm.sem_flg = 0L;
	sem_parm.sem_op = -1;
	#ifdef BLOCKING_DEBUG
		fprintf (stderr, "bsyscalls: [%d] claim_semaphore (%d, %d)\n", getpid(), semid, semnum);
	#endif
	while (((retval = semop (semid, &sem_parm, 1)) == -1) && (errno == EINTR)) {
		fprintf (stderr, "safe_claim_semaphore (): interrupted!\n");
		if (sf.combined_sync != 0) {
			fprintf (stderr, "something set sync flag to 0x%8.8x\n", sf.combined_sync);
		}
		/* value of semaphore... ? */
		retval = semctl (semid, semnum, GETVAL, sem_union);
		fprintf (stderr, "safe_claim_semaphore (): value is %d\n", retval);
		sem_parm.sem_num = semnum;
		sem_parm.sem_flg = 0L;
		sem_parm.sem_op = -1;
		sched_yield ();
	}
	if (retval == -1) {
		#ifdef BLOCKING_DEBUG
			fprintf (stderr, "bsyscalls: [%d] claim_semaphore (%d, %d) failed: %s\n", getpid(), semid, semnum, strerror(errno));
		#endif
		return -1;
	}
	#ifdef BLOCKING_DEBUG
		fprintf (stderr, "bsyscalls: [%d] claim_semaphore (%d, %d) succeeded\n", getpid(), semid, semnum);
	#endif
	return 0;
}


/*
 *	int notify_semaphore (int semid, int semnum)
 *	Notifies a semaphore.  Returns 0 on success (ie, notified), -1 if an error occured
 */
static int		notify_semaphore (int semid, int semnum)
{
	struct sembuf sem_parm;

	sem_parm.sem_num = semnum;
	sem_parm.sem_flg = SEM_UNDO;
	sem_parm.sem_op = 1;
	#ifdef BLOCKING_DEBUG
		fprintf (stderr, "bsyscalls: [%d] notify_semaphore (%d, %d)\n", getpid(), semid, semnum);
	#endif
	if (semop (semid, &sem_parm, 1) == -1) {
		return -1;
	}
	return 0;
}


/*
 *	void bsyscalls_dispatch (word *ws, char *raddr, word *ws_arg, void (*func)(word *), int adjustment)
 *	Dispatches a thread to handle a blocking system call
 */
void bsyscalls_dispatch (word *ws, char *raddr, word *ws_arg, void (*func)(word *), int adjustment)
{
	int no_clone, deferred_new;
	#ifdef BLOCKING_PROFILE
		long long t_time, x_time;
	#endif

	#ifdef BLOCKING_DEBUG
		fprintf (stderr, "bsyscalls_dispatch () : %d clones, %d spare\n", num_clones, num_spare);
	#endif
	#ifdef BLOCKING_PROFILE
		LOADTIME (t_time);
	#endif
	no_clone = 0;
	if (num_spare == 1) {
		deferred_new = 1;
	} else if (!num_spare) {
		/* fprintf (stderr, "bsyscalls_dispatch () : creating new clone\n"); */
		if (create_clone () < 0) {
			deferred_new = 0;
			no_clone = 1;
		} else {
			deferred_new = 1;
		}
	} else {
		deferred_new = 0;
	}
	/* if !no_clone, clone is waiting */
	if (no_clone) {
		fprintf (stderr, "KRoC: bsyscalls: out of clones.\n");
		occam_exit (1, (int)raddr);
	}
	barrier ();
	while (spl_lock_or_fail (&dispatch_lock)) {
		/* avoid extended spinning by rescheduling */
		sched_yield ();
	}
	#ifdef BLOCKING_PROFILE
		v_dispatching->time_dispatch = t_time;
	#endif
	v_dispatching->ws_ptr = (int *)ws;
	/* Adjust for the magic channel parameter (if applicable) */
	v_dispatching->ws_arg = (int *)((char *)ws_arg + adjustment);
	v_dispatching->func = func;
	v_dispatching->raddr = raddr;	/* to store in ws_ptr[Iptr] later */
	v_dispatching->adjustment = adjustment;
	num_spare--;
	barrier ();
/* fprintf (stderr, "dispatch: notify ()\n"); */
	if (notify_semaphore (clone_sem, NOTIFY_SEM) < 0) {
		fprintf (stderr, "KRoC: error! -- notify_semaphore () failed with %s\n", strerror(errno));
	}
/* fprintf (stderr, "dispatch: done notify ()\n"); */
	barrier ();
	#ifdef BLOCKING_PROFILE
		LOADTIME (x_time);
		fprintf (stderr, "trigger %Ld %Ld\n", t_time, x_time);
	#endif
	#ifdef BLOCKING_DEBUG
		fprintf (stderr, "bsyscalls_dispatch (0x%lx, 0x%lx, 0x%lx) : notified clone\n", ws, ws_arg, func);
	#endif
/* fprintf (stderr, "bsyscalls_dispatch (0x%lx, 0x%lx, 0x%lx) : notified clone\n", ws, ws_arg, func); */
	if (deferred_new) {
		create_clone ();
	}
	return;
}


/*
 *	int bsyscalls_pending (void)
 *	Returns number of clones currently blocked
 */
int bsyscalls_pending (void)
{
	return (num_clones - num_spare);
}

#else	/* !BLOCKING_SYSCALLS */
	#if defined(BLOCKING_SYSCALLS_UNSUPPORTED)
		void _blocking_syscalls_unsupported (void)
	#else
		void _blocking_syscalls_disabled (void)
	#endif
	{
		return;
	}
#endif	/* !BLOCKING_SYSCALLS */

