/*
 *	deadlock.c -- run-time deadlock debugging
 *	Also includes run-time Floating Point error stuff
 *	Copyright (C) 2000 Fred Barnes  <frmb2@ukc.ac.uk>
 *
 *	This program is free software; you can redistribute it and/or modify
 *	it under the terms of the GNU General Public License as published by
 *	the Free Software Foundation; either version 2 of the License, or
 *	(at your option) any later version.
 *
 *	This program is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU General Public License for more details.
 *
 *	You should have received a copy of the GNU General Public License
 *	along with this program; if not, write to the Free Software
 *	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <stdio.h>

#include "sched_types.h"
#include "sched_consts.h"
#include "occam_io.h"
#include "timer.h"

#ifndef RTS_H
	#define RTS_H
	#include "rts.h"
#endif

typedef void (*FUNCPTR)();

extern void _occam_start(); /* occam code entry point */
extern word _wsbytes; /* occam workspace size */
extern word _vsbytes; /* occam vectorspace size */

static void dump_process_workspace( word *lwptr );
static int deadlock_debug (void);
static void segv_handler (int);
static void debug_deadlock_out (FILE *stream, char *ops, char *proc, char *file, int line);

static char *ws_ptr;
static char *vs_ptr;
static int debug_dead;
static volatile int got_segv;

void (*old_segv_handler)(int);

#define INS_X86_JMPIMMS		(0xeb)

/* These are ours, set from occam_entry.c */

word *dead_workspace;
int dead_size;

/*
 * void give_wsvs (char *ws, char *vs)
 * Called by _enter() in occam_entry.c to give us ws and vs pointers
 * -- also our initialisation point
 */
void give_wsvs (char *ws, char *vs)
{
	debug_dead = 0;
	if (!ws) {
		fprintf (stderr, "KRoC: %s: ws given as 0.  Not enabling deadlock debugging.\n", __FILE__);
		return;
	}
	if (!vs) {
		fprintf (stderr, "KRoC: %s: vs given as 0.  Not enabling deadlock debugging.\n", __FILE__);
		return;
	}
	debug_dead = 1;
	ws_ptr = ws;
	vs_ptr = vs;
	return;
}


/*
 * int deadlock_debug (void)
 * Performs the WS search etc.
 * Returns 0 on success, -1 on failure (ie, program compiled without -X5)
 */
static int deadlock_debug (void)
{
	char *search, *debug_filename, *debug_procname;
	unsigned char *codeptr;
	int *search_int;
	int *search_limit;
	int found;
	int debug_dlop, debug_kentry, debug_line, debug_file, debug_proc;
	int did_print;
	unsigned char ins_1, ins_2;
	static char *dlops[] = { "INVALID", "IN", "OUT", "OUTBYTE", "OUTWORD", "ALTWT", "TALTWT" };
	static int num_dlops = 7;

	found = 0;
	did_print = 0;
	if ((int)ws_ptr | 0x03) {
		search = (char *)(((unsigned int)ws_ptr & 0xfffffffc) + 4);
	} else {
		search = (char *)ws_ptr;
	}
	got_segv = 0;
	old_segv_handler = signal (SIGSEGV, segv_handler);
	/* Nicely aligned workspace */
	search_int = (int *)((int)search + 8);
	search_limit = (int *)((int)search + _wsbytes);
	#ifdef DEBUG_DEBUG
		fprintf (stderr, "KRoC: %s: searching from 0x%lx to 0x%lx...\n", __FILE__, (unsigned long int)search_int, (unsigned long int)search_limit);
	#endif
	for (; search_int < search_limit; search_int++) {
		if ((int *)search_int[-2] == search_int) {
			/* Possibility */
			codeptr = (unsigned char *)search_int[-1];
			if (codeptr < (unsigned char *)_occam_start) {
				/* Nope.. */
				#ifdef DEBUG_DEBUG
					fprintf (stderr, "KRoC: %s: codeptr (0x%lx) < _occam_start (0x%lx)\n", __FILE__, (unsigned long int)codeptr,
						(unsigned long int)_occam_start);
				#endif
				continue;
			}
			/* Try dereferencing it... */
			ins_1 = *codeptr;
			ins_2 = *(codeptr + 1);
			if (got_segv) {
				/* Nope, obviously not.. */
				got_segv = 0;
				#ifdef DEBUG_DEBUG
					fprintf (stderr, "KRoC: %s: SEGV while dereferencing at 0x%lx.  _occam_start = 0x%lx\n", __FILE__,
						(unsigned long int)codeptr, (unsigned long int)_occam_start);
				#endif
				continue;
			}
			/* They should be two halves */
			if (ins_1 != INS_X86_JMPIMMS) {
				#ifdef DEBUG_DEBUG
					fprintf (stderr, "KRoC: %s: INS_X86_JMPIMMS=0x%2.2x, read 0x%2.2x.\n", __FILE__, INS_X86_JMPIMMS, ins_1);
				#endif
				continue;
			}
			if (ins_2 != 0x11) {
				#ifdef DEBUG_DEBUG
					fprintf (stderr, "KRoC: %s: jump (0x%2.2x) out of range (should be 0x11)\n", __FILE__, ins_2);
				#endif
				continue;
			}
			/* Could be likely.. */
			codeptr += 2;
			/* codeptr should be pointing at 12 bytes of debugging info + 5 bytes of jump code */
			ins_1 = codeptr[3];
			ins_2 = codeptr[2];
			debug_kentry = ins_2;
			debug_dlop = ins_1;
			ins_1 = codeptr[0];
			ins_2 = codeptr[1];
			debug_line = (ins_2 << 8) + ins_1;
			if (got_segv) {
				got_segv = 0;
				continue;
			}
			if ((debug_dlop >= num_dlops) || (debug_dlop <= 0)) {
				/* Well..  nearly */
				#ifdef DEBUG_DEBUG
					fprintf (stderr, "KRoC: %s: DLOP read as %d.  Bytes are: %2.2X %2.2X %2.2X %2.2X\n", __FILE__, debug_dlop, codeptr[0], codeptr[1],
						codeptr[2], codeptr[3]);
				#endif
				continue;
			}
			/* Next word is filename/procname */
			codeptr += 4;
			ins_1 = codeptr[0];
			ins_2 = codeptr[1];
			debug_proc = ((int)ins_2 << 8) + ins_1;
			ins_1 = codeptr[2];
			ins_2 = codeptr[3];
			debug_file = ((int)ins_2 << 8) + ins_1;
			if (got_segv) {
				got_segv = 0;
				#ifdef DEBUG_DEBUG
					fprintf (stderr, "KRoC: %s: SEGV while deferencing at 0x%lx.  _occam_start = 0x%lx\n", __FILE__,
						(unsigned long int)codeptr, (unsigned long int)_occam_start);
				#endif
				continue;
			}
			/* Next word is some magic */
			codeptr += 4;
			ins_1 = codeptr[0];
			ins_2 = codeptr[1];
			if ((ins_1 != 0xde) || (ins_2 != 0xad)) {
				/* Incorrect first magic */
				#ifdef DEBUG_DEBUG
					fprintf (stderr, "KRoC: %s: Read magic as 0x%2.2x%2.2x, should have been 0xdead\n", __FILE__, ins_1, ins_2);
				#endif
				continue;
			}
			ins_1 = codeptr[2];
			ins_2 = codeptr[3];
			if ((ins_1 != 0xbe) || (ins_2 != 0xef)) {
				/* Incorrect second magic */
				#ifdef DEBUG_DEBUG
					fprintf (stderr, "KRoC: %s: Read magic as 0x%2.2x%2.2x, should have been 0xbeef\n", __FILE__, ins_1, ins_2);
				#endif
				continue;
			}
			/* Jump into codeptr... */
			codeptr += 4;
			__asm__ __volatile__ ("\
				call *%%eax\n\t\
				" : "=a" (debug_filename), "=b" (debug_procname)\
				  : "a" (codeptr)\
				  : "cc", "memory" );
			if (got_segv) {
				#ifdef DEBUG_DEBUG
					fprintf (stderr, "KRoC: %s: SEGV while executing at 0x%lx\n", __FILE__, (unsigned long int)codeptr);
				#endif
				continue;
			}
			if (debug_filename < (char *)codeptr) {
				#ifdef DEBUG_DEBUG
					fprintf (stderr, "KRoC: %s: filename table (0x%lx) < codeptr (0x%lx)\n", __FILE__,
						(unsigned long int)debug_filename, (unsigned long int)codeptr);
				#endif
				continue;
			}
			if (debug_procname < (char *)codeptr) {
				#ifdef DEBUG_DEBUG
					fprintf (stderr, "KRoC: %s: procname table (0x%lx) < codeptr (0x%lx)\n", __FILE__,
						(unsigned long int)debug_procname, (unsigned long int)codeptr);
				#endif
				continue;
			}
			/* Check them */
			if (debug_file >= *(int *)(debug_filename)) {
				#ifdef DEBUG_DEBUG
					fprintf (stderr, "KRoC: %s: filename %d out of range of table (%d)\n", __FILE__,
						debug_file, *(int *)(debug_filename));
				#endif
				continue;
			}
			if (debug_proc >= *(int *)(debug_procname)) {
				#ifdef DEBUG_DEBUG
					fprintf (stderr, "KRoC: %s: procname %d out of range of table (%d)\n", __FILE__,
						debug_proc, *(int *)(debug_procname));
				#endif
				continue;
			}
			/* Okay.. */
			debug_filename += *(int *)(debug_filename + ((1 + debug_file) * 4));
			debug_procname += *(int *)(debug_procname + ((1 + debug_proc) * 4));
			found++;
			if (!did_print) {
				fprintf (stderr, "KRoC: Deadlock:\n");
				did_print = 1;
			}
			debug_deadlock_out (stderr, dlops[debug_dlop], debug_procname, debug_filename, debug_line);
		}
	}
	signal (SIGSEGV, old_segv_handler);
	if (!found) {
		fprintf (stderr, "KRoC: %s: No valid processes found in workspace.\n", __FILE__);
		return -1;
	}
	return 0;
}


/*
 * void debug_deadlock_out (FILE *stream, char *ops, char *proc, char *file, int line)
 * Sends out debugging info on 'stream'
 */
static void debug_deadlock_out (FILE *stream, char *ops, char *proc, char *file, int line)
{
	#if defined(DEADLOCK_TERSE)
		fprintf (stream, "%s(%d)-%s  PROC %s (...)\n", file, line, ops, proc);
	#elif defined(DEADLOCK_BREIF)
		fprintf (stream, "\t%s:%d\n", file, line);
		fprintf (stream, "\t\tPROC %s () instruction %s\n", proc, ops);
	#else
		fprintf (stream, "\n  Instruction \"%s\"\n", ops);
		fprintf (stream, "    in PROC \"%s\"\n", proc);
		fprintf (stream, "    in file \"%s\"\n", file);
		fprintf (stream, "    near line %d\n", line);
	#endif
	return;
}


/*
 * void segv_handler (int signum)
 * SIGSEGV handler
 */
static void segv_handler (int signum)
{
	got_segv = 1;
	return;
}


/*
 * void kernel_dead (int status)
 * Called when the kernel deadlocks
 */
void kernel_dead (int status)
{
	word *search;
	int found;

	if (debug_dead) {
		if (deadlock_debug()) {
			/* It failed... */
			debug_dead = 0;
		}
	}
	if (!debug_dead) {
		dead (status);
	} else {
		dead_quiet (status);
	}
	return;
}


/*
 * void dump_process_workspace (word *lwptr)
 * Dumps the workspace values
 */
static void dump_process_workspace (word *lwptr)
{
	fprintf( stderr, "process: wptr=%p\n", lwptr );
	fprintf( stderr, "0\tTemp\t\t0x%8.8x\n", lwptr[0] );
	fprintf( stderr, "-4\tIptr\t\t0x%8.8x\n", lwptr[-1] );
	fprintf( stderr, "-8\tLink\t\t0x%8.8x\n", lwptr[-2] );
	fprintf( stderr, "-12\tAlt state\t0x%8.8x\n", lwptr[-3] );
	fprintf( stderr, "-16\tValid time\t0x%8.8x\n", lwptr[-4] );
	fprintf( stderr, "-20\tTime\t\t0x%8.8x\n", lwptr[-5] );
	fflush( stderr );
	return;
}


/*
 * int fpe_debug (void)
 * Called upon SIGFPE -- when a floating point error occurs
 */
int fpe_debug (void)
{

}


