/*
 *	dynproc.c - Dynamic process management
 *	Copyright (C) 2000 Fred Barnes <frmb2@ukc.ac.uk>
 *
 *	This program is free software; you can redistribute it and/or modify
 *	it under the terms of the GNU General Public License as published by
 *	the Free Software Foundation; either version 2 of the License, or
 *	(at your option) any later version.
 *
 *	This program is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU General Public License for more details.
 *
 *	You should have received a copy of the GNU General Public License
 *	along with this program; if not, write to the Free Software
 *	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <stdio.h>
#include <string.h>
#include <sys/types.h>
#include <unistd.h>

#ifdef DYNAMIC_PROCS

#include <dlfcn.h>

#define __DYNPROC_C

#include "sched_types.h"
#include "rts.h"
#include "dynproc.h"


/*
 *	void do_ccsp_openlib (char *filename, int namelen, int *handle)
 *	loads a dynamic library
 */
void do_ccsp_openlib (char *filename, int namelen, int *handle)
{
	char fname[FILENAME_MAX];

	if (namelen >= FILENAME_MAX) {
		namelen = (FILENAME_MAX - 1);
	}
	memcpy (fname, filename, namelen);
	fname[namelen] = '\0';
	*handle = (int)dlopen (fname, RTLD_LAZY|RTLD_GLOBAL);
	if (!(*handle)) {
		fprintf (stderr, "KRoC: failed to ccsp_openlib [%s] because: %s\n", fname, dlerror());
	}
}

/*
 *	void do_ccsp_closelib (void *handle)
 *	closes a dynamic library
 */
void do_ccsp_closelib (void *handle)
{
	if (handle != (void *)0) {
		dlclose (handle);
	}
}

/*
 *	occam -> C stubs
 */
void _ccsp_openlib (int *wsarg) { do_ccsp_openlib ((char *)(wsarg[0]), (int)(wsarg[1]), (int *)(wsarg[2])); }
void _ccsp_closelib (int *wsarg) { do_ccsp_closelib ((void *)(wsarg[0])); }


/*
 *	word *load_dynamic_process (int *wsargs, word **irfunc)
 *	loads a dynamic process and adds it to the run queue (after setting it up, of course)
 */
word *load_dynamic_process (int *wsargs, word **rfunc)
{
	char *new_ws, *new_vs;
	int *new_wsbytes, *new_vsbytes;
	void *d_handle;
	char *f_name, *error;
	int f_len;
	char symname[128];
	void *func;
	int i;
	word *XWptr;

	f_name = (char *)(wsargs[0]);
	f_len = (int)(wsargs[1]);
	d_handle = (void *)(wsargs[2]);
	if (f_len >= 112) {
		fprintf (stderr, "KRoC: load_dynamic_process: name too long!\n");
		return NULL;
	}
	memcpy (symname, "O_", 2);
	memcpy (symname+2, f_name, f_len);
	symname[f_len+2] = '\0';
	if (!d_handle) {
		fprintf (stderr, "KRoC: load_dynamic_process: NULL handle!\n");
		return NULL;
	}
	func = dlsym (d_handle, symname);
	error = dlerror ();
	if (error) {
		fprintf (stderr, "KRoC: load_dynamic_process: symbol %s not found\n", symname);
		return NULL;
	}
	symname[0] = '_';
	memcpy (symname + 1, f_name, f_len);
	memcpy (symname + 1 + f_len, "_wsbytes", 8);
	symname[f_len + 9] = '\0';
	new_wsbytes = (int *)dlsym (d_handle, symname);
	error = dlerror ();
	if (error) {
		fprintf (stderr, "KRoC: load_dynamic_process: symbol %s not found\n", symname);
		return NULL;
	}
	symname[f_len + 2] = 'v';
	new_vsbytes = (int *)dlsym (d_handle, symname);
	error = dlerror ();
	if (error) {
		fprintf (stderr, "KRoC: load_dynamic_process: symbol %s not found\n", symname);
		return NULL;
	}
	/* create new workspace */
	new_ws = (char *)malloc (*new_wsbytes + 48 + *new_vsbytes + 256);
	if (!new_ws) {
		fprintf (stderr, "KRoC: load_dynamic_process: out of memory!\n");
		return NULL;
	}
	XWptr = (word *)(new_ws + (int)(*new_wsbytes));
	/* copy args into new workspace */
	for (i=1; i<6; i++) {
		XWptr[i] = wsargs[i+1];
	}
	new_vs = new_ws + (*new_wsbytes + 48 + 128);
	XWptr[6] = (word)new_vs;
	XWptr[7] = (word)new_ws;		/* so we can free it! */
	#ifdef DEBUG_DYNAMIC_PROCS
		fprintf (stderr, "KRoC: load_dynamic_process: almost there! (wsbytes=%d, vsbytes=%d, func=%p, XWptr=%p)\n", *new_wsbytes, *new_vsbytes, func, XWptr);
	#endif
	*rfunc = func;
	return XWptr;
}


#else	/* !DYNAMIC_PROCS */


/*
 *	void no_dynamic_process_support (void)
 *	called if somebody tries to execute a dynamic process (ccsp.openlib should fail during compilation though..)
 */
void no_dynamic_process_support (void)
{
	fprintf (stderr, "KRoC: dynamic process support is not enabled!\n");
	user_bad_exit ();
}


#endif	/* !DYNAMIC_PROCS */

