/*
 *	Queue manipulation macros
 *	Allows the encapsulation and isolation of the implementation of priorities.
 *	Copyright (C) 1998 Jim Moores
 *
 *	This program is free software; you can redistribute it and/or modify
 *	it under the terms of the GNU General Public License as published by
 *	the Free Software Foundation; either version 2 of the License, or
 *	(at your option) any later version.
 *
 *	This program is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU General Public License for more details.
 *
 *	You should have received a copy of the GNU General Public License
 *	along with this program; if not, write to the Free Software
 *	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*
 * Copyright (C) Jim Moores 1998
 * University of Kent at Canterbury.
 */

#ifndef QUEUE_MACROS_H
#define QUEUE_MACROS_H
#ifdef PRIORITIES

#include "debug_defs.h"

#if 0
word* Fptr[NUM_PRIORITIES];
word* Bptr[NUM_PRIORITIES];
priority_t priority_table;
#else
prioratised_queue runq;
#endif

#ifdef USE_ASM_BITOPS
#else
#define IsRunQEmpty(NUM) (!(runq.priority_table & 1<<(NUM)))
#define IsQEmpty(NUM,Q) (!((Q)->priority_table & 1<<(NUM)))

#define SetRunQStatus(NUM) runq.priority_table |= (1<<(NUM))
#define SetQStatus(NUM, Q) (Q)->priority_table |= (1<<(NUM))
#define ClearRunQStatus(NUM) runq.priority_table &= (~(1<<NUM))
#define ClearQStatus(NUM, Q) (Q)->priority_table &= (~(1<<NUM))
#define GetPriority(P) ((P)[Priority_f])
#endif
#if 0
#define QUEUE(P) \
  if (IsRunQEmpty(GetPriority(P))) { \
    runq.Fptr[GetPriority(P)] = P; \
  } else { \
    (runq.Bptr[GetPriority(P)])[Link] = (word)P; \
  } \
  Bptr[GetPriority(P)] = P; \
  /* tag that its on a run queue for abitrary priority set */ \
  (P)[SuspendQ] = (word)&runq; \
  SetRunQStatus(GetPriority(P)) \
  /* now update the current maximum priority, so scheduler knows to \
   * schedule _this_ process if it is of higher priority than the current \
   */ \
  if (current_max_priority < GetPriority(P)) { \
    current_max_priority = GetPriority(P); \
  }
#else
inline void QUEUE(word *P) {
  DBGMSG("QUEUE: priority_table = 0x%x.  priority level %d.\n", \
                  runq.priority_table, GetPriority(P));
  if (IsRunQEmpty(GetPriority(P))) {
    runq.Fptr[GetPriority(P)] = P;
    #ifdef GSL_SMP
    DBGMSG("QUEUE: releasing lock, current value (b4) = %d", wait_lock);
    RELEASE(wait_lock); /* adding a process to an empty queue, wake any
                         * sleeping processes (cheaper than checking if
                         * _really_ necessary.
                         */
    #endif
  } else {
    (runq.Bptr[GetPriority(P)])[Link] = (word)P;
  }
  runq.Bptr[GetPriority(P)] = P;
  /* tag that its on a run queue for abitrary priority set */
  (P)[SuspendQ] = (word)&runq;
  SetRunQStatus(GetPriority(P));
  /* now update the current maximum priority, so scheduler knows to
   * schedule _this_ process if it is of higher priority than the current
   */
  if (current_max_priority < GetPriority(P)) {
    current_max_priority = GetPriority(P);
  }
}
#endif

void EnQueue(word *P, prioratised_queue *q) {
  DBGMSG("EnQueue: priority_table = 0x%x.  priority level %d.\n", \
                  q->priority_table, GetPriority(P));
  if (IsQEmpty(GetPriority(P),q)) {
    q->Fptr[GetPriority(P)] = P;
  } else {
    (q->Bptr[GetPriority(P)])[Link] = (word)P;
  }
  q->Bptr[GetPriority(P)] = P;
  /* tag that its on a run queue for abitrary priority set */
  (P)[SuspendQ] = (word)q;
  SetQStatus(GetPriority(P), q);
  /* now update the current maximum priority, so scheduler knows to
   * schedule _this_ process if it is of higher priority than the current
   */
}

/* Remove a process from a queue - this algorithm assumes that we know that
 * the process is on the queue.
 */
void RemoveFromQ(word *Wptr, prioratised_queue* q, int priority) {
  word *Fptr = q->Fptr[priority];
  word *Bptr = q->Bptr[priority];
  word *previous;
  word *current;
  if (Fptr == Wptr) {
    /* We've found it on the front of the queue */
    if (Fptr==Bptr) {
      /* only item on the queue */
      Fptr = Bptr = NULL;

    } else {
      /* more than one item on queue */
      Fptr = (word *)Fptr[Link];
    }
  } else {
    /* Its somewhere else on the queue */
    previous = Fptr;
    current = (word *)Fptr[Link];
    while (current != Wptr) {
      previous = current;
      current = (word *)current[Link];
    }
    /* current must = Wptr */
    if (current != Bptr) {
      /* not at the end of the queue */
      previous[Link] = current[Link];
      /* link over removed process */
    } else {
      /* at the end of the queue, move Bptr */
      Bptr = previous;
    }
  }
  /* update pointers passed in */
  q->Fptr[priority] = Fptr;
  q->Bptr[priority] = Bptr;
  /* tag that its no longer on a suspend queue (possibly a run queue) */
  Wptr[SuspendQ] = NotProcess_p;
}

/* Remove a process from a run queue - also updates the bitfield used for
 * fast search for full queues.
 */
void RemoveFromRunQ(int priority, word *Wptr) {
  RemoveFromQ(Wptr, &runq, priority);
  /* RACE HAZARD HERE IN SMP POSSIBLE (queue empty and bit not cleared
   * immediately
   */
  if (runq.Fptr == NULL) {
    /* queue has been emptied, so we need to reset appropriate bit */
    ClearRunQStatus(priority);
  }
}
/* was inline */
word* GetNextProcess(int priority_level) {
  int i;
  word *P;
  DBGMSG("GetNextProcess: priority_level=%d.\n",priority_level);
  DBGMSG("GetNextProcess: priority_table=0x%x.\n",runq.priority_table);
  DBGMSG("GetNextProcess: IsRunQEmpty(i)=%d",IsRunQEmpty(i));
  for (i=priority_level; (IsRunQEmpty(i)); i--) {
    DBGMSG("GetNextProcess: i=%d, priority_to_search_from=%d.\n",i,priority_level);
  }
  DBGMSG("GetNextProcess: Called. \n");
  /* loop ended because a full queue was found */
  /* we know this because this function is only called after checking */

  P = runq.Fptr[i];
  if (runq.Fptr[i] == runq.Bptr[i]) {
    runq.Fptr[i] = runq.Bptr[i] = NotProcess_p;
    ClearRunQStatus(i);
    DBGMSG("GetNextProcess: clearing bit for queue %d.\n",i);
  } else {
    runq.Fptr[i] = (word *)((runq.Fptr[i])[Link]);
  }
  /* tag that its not on a queue for abitrary priority setting and sema4s */
  P[SuspendQ] = NotProcess_p;
  return (P);
}

/* this function gets the next process from a specified set of queues */
/* for this we do _not_ assume that we will find a process - it will */
/* return NotProcess_p if none is found. */
word* GetNextProcessFromASuspendQ(prioratised_queue* q) {
  int i;
  word *P;
  //DBGMSG("GetNextProcess: priority_level=%d.\n",priority_level);
  for (i=NUM_PRIORITIES; (IsQEmpty(i,q) && (i>=0)); i--) {
    //DBGMSG("GetNextProcess: i=%d, priority_to_search_from=%d.\n",i,priority_level);
  }
  //DBGMSG("GetNextProcess: Called. \n");
  /* loop ended because a full queue was found */
  /* we know this because this function is only called after checking */

  P = q->Fptr[i];
  if (q->Fptr[i] == q->Bptr[i]) {
    q->Fptr[i] = q->Bptr[i] = NotProcess_p;
    ClearQStatus(i, q);
    DBGMSG("GetNextProcess: clearing bit for queue %d.\n",i);
  } else {
    q->Fptr[i] = (word *)((q->Fptr[i])[Link]);
  }
  /* tag that its not on a queue for abitrary priority setting and sema4s */
  if (P != NotProcess_p) {
    P[SuspendQ] = NotProcess_p;
  }
  return (P);
}

word* GetNextProcessFrom(int priority_level) {
  word *P;
  P = runq.Fptr[priority_level]; /* get process from front of run queue */
  /* clean up the queue */
  if (runq.Fptr[priority_level] == runq.Bptr[priority_level]) {
    runq.Fptr[priority_level] = NotProcess_p;
    runq.Bptr[priority_level] = NotProcess_p;
    ClearRunQStatus(priority_level);
    DBGMSG("GetNextProcessFrom: clearing but for queue %d.\n", priority_level);
  } else {
    runq.Fptr[priority_level] = (word *)((runq.Fptr[priority_level])[Link]);
  }
  /* tag that its not on a queue any more for abitrary priority set.
   * and semaphores with timeouts  */
  P[SuspendQ] = NotProcess_p;
  return (P);
}

void AppendAllFromQtoQ(prioratised_queue* from, prioratised_queue* to) {
  /* we used to keep a record of the highest priority process currently on
   * the queue, but updating this turned out to be too high an overhead, so
   * we need to do a linear search of the queues for each queue joining
   */
  int i;
  for (i=NUM_PRIORITIES; i>=0; i--) {
    if (!IsQEmpty(i,from)) {
      /* if source queue has something on it */
      if (IsQEmpty(i,to)) {
        /* if target queue is empty then just replace the pointers */
        to->Fptr[i] = from->Fptr[i];
        to->Bptr[i] = from->Bptr[i];
        SetQStatus(i, to); /* label it full */
      } else {
        /* target queue not empty, append instead */
        /* start by linking the front of the `from' to the back of the `to' */
        (to->Bptr[i])[Link] = (word)from->Fptr[i];
        /* move to's back pointer to point to the end of the new queue */
        to->Bptr[i] = from->Bptr[i];
      }
      if (current_max_priority<i) {
        current_max_priority = i;
      }
      from->Fptr[i] = NotProcess_p;
      from->Bptr[i] = NotProcess_p;
      ClearQStatus(i, from);
    }
  }
}

#define NoProcesses() (runq.priority_table == 0)
/* need to do the full work if multiple priorities */
#define GetNextProcess_NonEmptyQ() GetNextProcessFrom(GetPriority(Wptr))

#else
word* Fptr;
word* Bptr;
/* put the process P on the run queue */
#define QUEUE(P) \
  if (Fptr == NotProcess_p) { \
    Fptr = P; \
  } else {\
    Bptr[Link] = (word)P; \
  } \
  Bptr = P
#define QUEUE_Q(F,B) \
  if (Fptr == NotProcess_p) { \
    Fptr = F; \
  } else { \
    Bptr[Link] = (word)F;\
  } \
  Bptr = B

#define QUEUE_VQ(F,B) \
  if (Fptr == NotProcess_p) { \
    (volatile word *)Fptr = F; \
  } else { \
    (volatile word)(Bptr[Link]) = (volatile word)F; \
  } \
  (volatile word *)Bptr = B


/* returns the first process on the run queue.  This assumes that the
 * run queue is non-empty because this function has the contraint that
 * processes must have been added before calling it (for efficiency
 * reasons).
 */
#define GetNextProcess_NonEmptyQ() Fptr;Fptr = (word *)Wptr[Link]

/* This returns the first process on the run queue, but does not assume
 * that the queue is non-empty.
 */
#define GetNextProcess() \
    Fptr; \
    if (Fptr == Bptr) { \
      Fptr = NotProcess_p; \
      Bptr = NotProcess_p; \
    } else \
      Fptr = (word *)Fptr[Link]

/* returns true if the run queue is empty */
#define NoProcesses() (Fptr==NotProcess_p)
#endif
#endif /* queue_macros.h */
