/*
 *	KROC user process
 *	Copyright (C) 1995, 1996, 1997  D.C. Wood, P.H. Welch, D.J. Beckett
 *
 *	This program is free software; you can redistribute it and/or modify
 *	it under the terms of the GNU General Public License as published by
 *	the Free Software Foundation; either version 2 of the License, or
 *	(at your option) any later version.
 *
 *	This program is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU General Public License for more details.
 *
 *	You should have received a copy of the GNU General Public License
 *	along with this program; if not, write to the Free Software
 *	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*
 * This is the UNIX process that runs the occam code and traps its
 * termination cleanly or via signals.  The co-process handles blocking
 * on fd 0 (the keyboard) and uses a UNIX signal to deliver notification
 * of a new keypress.
 *
 * $Source: /proj/ofa/rts/RCS/userproc.c,v $
 *
 * $Id: userproc.c,v 1.11 1997/08/12 14:38:28 djb1 Exp $
 *
 * (C) Copyright 1995,1996,1997
 *    D.C. Wood <D.C.Wood@ukc.ac.uk>
 *    P.H. Welch <P.H.Welch@ukc.ac.uk>
 *    Dave Beckett <D.J.Beckett@ukc.ac.uk>
 * University of Kent at Canterbury
 *
 * Some modifications Copyright (C) 2000 Fred Barnes
 */

#include <unistd.h>
#include <fcntl.h>

#include "rts.h"
#include "386/sched_types.h"

#if defined(USEFUL_SEGFAULT) || defined(USEFUL_FPEDEBUG)
#include <signal.h>
#endif

#ifdef BLOCKING_SYSCALLS
extern void bsyscalls_destroy_clones (void);

static struct sigaction bsigact;
#endif	/* BLOCKING_SYSCALLS */

extern sync_type sf;

extern char rts_version_message[];     /* In version.c */

/* 'local' variables */

/* These are used in sched.s */
FILE *kroc_in;         /* not used */
FILE *kroc_out;
FILE *kroc_err;

int kbd_char_signalled = 0;
sigset_t sig_mask, zero_mask;
int trace_count = 0;

char TTin_buf;

static int kbdproc_pid, to_kbdproc, from_kbdproc;
static int stdin_is_tty;
static volatile int faulted;

/* Floating point linkup */
extern int fpe_debug (void);

static void user_stop_signal (void);
static void user_cont_signal (void);
static void userproc_exit (int exit_status, bool dump_core);
static void set_user_process_signals (void);
#ifdef USEFUL_SEGFAULT
static struct sigaction segv_action;
static void user_segv_handler (int sig, siginfo_t *siginfo, void *sptr);
#endif
#ifdef USEFUL_FPEDEBUG
static struct sigaction fpe_action;
static void user_fpe_handler (int sig, siginfo_t *siginfo, void *sptr);
#endif


/* Called from assembler kernel */
void occam_exit(int exit_status, int occam_iptr)
{
	#ifndef ALONE
		if (occam_iptr != 0) {
			fprintf (stderr, "OCCAM STOPped at %08x\n", occam_iptr);
		}
	#endif
	userproc_exit(exit_status, 0);
	return;
}


static void userproc_exit(int exit_status, bool dump_core) 
{
	#ifdef DEBUG_RTS
		if (trace) {
			fprintf (stderr, "USERPROC: Exit %d\n", exit_status); fflush (stderr);
		}
	#endif

	#ifdef DEBUG_RTS
	if (trace) {
		fputs ("USERPROC: About to kill KBDPROC\n", stderr); fflush (stderr);
	}
	#endif

	/* Terminate keyboard process, if still around */
	kill (kbdproc_pid, SIGTERM);

	#ifdef BLOCKING_SYSCALLS
		/* Kill any clones */
		bsyscalls_destroy_clones ();
	#endif	/* BLOCKING_SYSCALLS */

	/* reset terminal */
	if (stdin_is_tty) {
		restore_tty_state ();
	}
	#ifdef GENERATE_CORES
		if (dump_core) {
			fprintf (stderr, "KROC: Fatal error code %d, core dumped\n", exit_status); fflush (stderr);
			abort ();
		}
	#endif /* GENERATE_CORES */

	exit (exit_status);
}  /* userproc_exit */



void user_bad_exit (void)
{
	#ifdef DEBUG_RTS
		if (trace) {
			fputs ("USERPROC: I've been murdered!\n", stderr); fflush (stderr);
		}
	#endif
	userproc_exit (-8, 1);
} /* user_bad_exit */


void user_good_exit (void)
{
	userproc_exit (0, 0);
} /* user_good_exit */


/* SIGNAL handlers */

void user_tim_handler (void)      /* SIGALRM  handler - only set during hibernate */
{
	#ifdef DEBUG_RTS
		if (trace) {
			fputs ("USERPROC: Alarm ringing (SIGALRM).\n", stderr); fflush (stderr);
		}
	#endif
	sf.s.tim_sync = 1;
	#if defined(SIGNAL_TYPE_SYSV)
		signal (SIGALRM, (SigParam)user_tim_handler);
	#endif /* SIGNAL_TYPE */
}  /* user_tim_handler */


#if 0
void user_panic_handler (void)    /* SIGPANIC  handler - only set by panicking user */
{
	#ifdef DEBUG_RTS
		if (trace) {
			fputs ("USERPROC: Panic from user (SIGPANIC).\n", stderr); fflush (stderr);
		}
	#endif
	panic_sync = 1;
}  /* user_panic_handler */
#endif

#ifdef BLOCKING_SYSCALLS
void user_blocking_handler (void)
{
	#ifdef DEBUG_RTS
		if (trace) {
			fprintf (stderr, "USERPROC: Blocking system call finished (SIGBLOCKING = (SIGUSR1))\n");
		}
	#endif
	sf.s.block_sync = 1;
	#ifdef DEBUG_RTS
		if (trace) {
			fprintf (stderr, "USERPROC: Set block_sync\n");
		}
	#endif
} /* user_blocking_handler */
#endif	/* BLOCKING_SYSCALLS */

void user_kbd_handler (void)      /* SIGKBD handler - available any time */
{
	#ifdef DEBUG_RTS
		if (trace) {
			fprintf (stderr, "USERPROC: Character #%d arriving.\n", trace_count);
			++trace_count;
		}
	#endif
	kbd_char_signalled = 1;
	sf.s.kbd_sync = 1;
	#if defined(SIGNAL_TYPE_SYSV)
		signal (SIGKBD, (SigParam)user_kbd_handler);
	#endif /* SIGNAL_TYPE */
	#ifdef DEBUG_RTS
		if (trace) {
			fprintf (stderr, "USERPROC: set keyboard sync.");
			++trace_count;
		}
	#endif
}  /* user_kbd_handler */


void user_overflow (void)
{
	userproc_exit (1, 0);
} /* user_overflow */

#ifdef USEFUL_SEGFAULT
static void user_segv_handler (int sig, siginfo_t *siginfo, void *sptr)
{
	if (siginfo->si_signo != SIGSEGV) {
		fprintf (stderr, "KRoC: in user_segv_handler (), but not segv ?\n");
	} else {
		fprintf (stderr, "KRoC: segmentation fault:\n");
		fprintf (stderr, "    errno = %d\n", siginfo->si_errno);
		fprintf (stderr, "    origin = %d (", siginfo->si_code);
		switch (siginfo->si_code) {
		case SI_USER:
			fprintf (stderr, "user)\n");
			break;
		case SI_QUEUE:
			fprintf (stderr, "queue)\n");
			break;
		case SI_TIMER:
			fprintf (stderr, "timer)\n");
			break;
		case SI_MESGQ:
			fprintf (stderr, "mesgq)\n");
			break;
		case SI_ASYNCIO:
			fprintf (stderr, "asyncio)\n");
			break;
		case SI_SIGIO:
			fprintf (stderr, "sigio)\n");
			break;
		default:
			if (siginfo->si_code > 0) {
				fprintf (stderr, "kernel)\n");
			} else {
				fprintf (stderr, "[dunno])\n");
			}
			break;
		}
		fprintf (stderr, "    fault at 0x%x\n", (unsigned int)siginfo->si_addr);
		fprintf (stderr, "    sptr at  0x%x\n", (unsigned int)sptr);
	}
	if (!faulted) {
		faulted = 1;
		userproc_exit (1, 1);
	} else {
		kill (getpid (), SIGKILL);
	}
	return;
}
#endif	/* USEFUL_SEGFAULT */

#ifdef USEFUL_FPEDEBUG
static void user_fpe_handler (int sig, siginfo_t *siginfo, void *sptr)
{
	if (siginfo->si_signo != SIGFPE) {
		fprintf (stderr, "KRoC: in user_fpe_handler (), but not FPE ?\n");
	} else {
		fprintf (stderr, "KRoC: floating point exception:\n");
		switch (siginfo->si_code) {
		case FPE_INTDIV:
			fprintf (stderr, "    integer division by zero\n");
			break;
		case FPE_INTOVF:
			fprintf (stderr, "    integer overflow\n");
			break;
		case FPE_FLTDIV:
			fprintf (stderr, "    floating point divide by zero\n");
			break;
		case FPE_FLTOVF:
			fprintf (stderr, "    floating point overflow\n");
			break;
		case FPE_FLTUND:
			fprintf (stderr, "    floating point underflow\n");
			break;
		case FPE_FLTRES:
			fprintf (stderr, "    inexact result\n");
			break;
		case FPE_FLTINV:
			fprintf (stderr, "    invalid operation\n");
			break;
		case FPE_FLTSUB:
			fprintf (stderr, "    subscript out of range\n");
			break;
		default:
			fprintf (stderr, "    unspecified\n");
			break;
		}
		fprintf (stderr, "    fault at 0x%x\n", (unsigned int)siginfo->si_addr);
		fprintf (stderr, "    sptr at  0x%x\n", (unsigned int)sptr);
	}
	userproc_exit (1, 0);
	return;
}
#endif	/* USEFUL_FPEDEBUG */

#if defined(SIGNAL_TYPE_SYSV)
void user_trap_handler (int sig)
{
	if (sig == SIGSEGV) {
		fprintf (stderr, "KROC: Segmentation fault.\n");
	} else {
		fprintf (stderr, "KROC: Range error / STOP executed (signal %d)\n", sig);
	}
	if (!faulted) {
		faulted = 1;
		userproc_exit (1, 1);
	} else {
		/* Faulted during fault.. */
		kill (getpid(), SIGKILL);
	}
} /* user_trap_handler */
#else
void user_trap_handler (int sig, int code, struct sigcontext *scp, char *addr)
{
	/*      fprintf (stderr, "sig = 0x%x, code = 0x%x\n", sig, code);       */
	fputs ("KROC: ", stderr);
	switch (code & 0x1f) {           /* why? */
	case _SETERR:
		fputs ("Set error", stderr);
		break;
	case _OFLOW:
		fputs ("Integer overflow", stderr);
		break;
	case _RANGE:
		fputs ("Range error", stderr);
		break;
	case _FLOAT:
		fputs ("Floating-point error", stderr);
		break;
	case _UNREC:
		fputs ("Unrecognized instruction", stderr);
		break;
	case _UNIMP:
		fputs ("Unimplemented instruction", stderr);
		break;
	default:
		fprintf (stderr, "Unrecognized trap number 0x%x", code);
		break;
	}
	fputs ("\n", stderr);
	fflush (stderr);
	userproc_exit (code, 1);
} /* user_trap_handler */
#endif /* SIGNAL_TYPE */


#if defined(SIGNAL_TYPE_SYSV)
void user_fp_handler (int sig)
{
	if (fpe_debug ()) {
		fprintf (stderr, "KROC: Floating-point error\n");
		userproc_exit (42, 1);
	} else {
		userproc_exit (41, 0);
	}
}
#else
void user_fp_handler (int sig, int code, struct sigcontext *scp, char *addr)
{
	switch(code) {
	#ifdef FPE_INTOVF_TRAP
	case FPE_INTOVF_TRAP:
		fprintf (stderr, "KROC: FP error - integer overflow\n");
		break;
	#endif      
	#ifdef FPE_STARTSIG_TRAP
	case FPE_STARTSIG_TRAP:
		fprintf (stderr, "KROC: FP error - process using fp\n");
		break;
	#endif      
	#ifdef FPE_INTDIV_TRAP
	case FPE_INTDIV_TRAP:
		fprintf (stderr, "KROC: Floating Point error - integer divide by zero\n");
		break;
	#endif      
	#ifdef FPE_FLTINEX_TRAP
	case FPE_FLTINEX_TRAP:
		fprintf (stderr, "KROC: Floating Point error - floating inexact result\n");
		break;
	#endif      
	#ifdef FPE_FLTDIV_TRAP
	case FPE_FLTDIV_TRAP:
		fprintf (stderr, "KROC: Floating Point error - floating divide by zero\n");
		break;
	#endif      
	#ifdef FPE_FLTUND_TRAP
	case FPE_FLTUND_TRAP:
		fprintf (stderr, "KROC: Floating Point error - floating underflow\n");
		break;
	#endif      
	#ifdef FPE_FLTOPERR_TRAP
	case FPE_FLTOPERR_TRAP:
		fprintf (stderr, "KROC: Floating Point error - floating operand error\n");
		break;
	#endif      
	#ifdef FPE_FLTOVF_TRAP
	case FPE_FLTOVF_TRAP:
		fprintf (stderr, "KROC: Floating Point error - floating overflow\n");
		break;
	#endif      
	default:
		fprintf (stderr, "KROC: Unknown floating point error number 0x%x\n", code);
		break;
	}
	userproc_exit (42, 1);
}
#endif /* SIGNAL_TYPE */
/* user_fp_handler */


char read_the_kbd (void)             /* called from assembler */
{
	if (read (from_kbdproc, &TTin_buf, 1) != 1) {
		fputs ("USERPROC: Can't read tty char from KBDPROC\n", stderr);
		fflush (stderr);
		userproc_exit (3, 0);
	}
	#ifdef DEBUG_RTS
		if (trace) {
			fprintf (stderr, "USERPROC: '%c' (0x%x) was read from KBDPROC\n", Tin_buf, TTin_buf);
			fflush (stderr);
		}
	#endif
	return (TTin_buf);            /* dcw */
} /* read_the_kbd */


void acknowledge (void)                      /* called from assembler */
{       
	sf.s.kbd_sync = 0;
	#ifdef DEBUG_RTS
		if (trace) {
			fputs ("USERPROC: Acknowledge KBDPROC\n", stderr); fflush (stderr);
		}
	#endif

	if (write (to_kbdproc, &TTin_buf, 1) != 1) {
		#ifdef DEBUG_RTS
			if (trace) {
				fputs ("USERPROC: Can't (pipe) acknowledge keyboard\n", stderr);
				userproc_exit (3, 0);
			}
		#endif
	}
}  /* acknowledge */


void user_to_kbd_sync (void)
{
	char go_char;

	/* give go-ahead to keyboard process */
	#ifdef DEBUG_RTS
		if (trace) {
			fputs ("USERPROC: Signalling to KBDPROC\n", stderr); fflush (stderr);
		}
	#endif
	if (write (to_kbdproc, "c", 1) != 1) {
		fputs ("USERPROC: Can't write go-ahead\n", stderr); fflush (stderr);
		userproc_exit (3, 0);
	}

	/* wait for keyboard process's go-ahead */
	#ifdef DEBUG_RTS
		if (trace) {
			fputs ("USERPROC: Waiting for KBDPROC\n", stderr); fflush (stderr);
		}
	#endif

	if ((read (from_kbdproc, &go_char, 1) != 1) || (go_char != 'p')) {
		fprintf (stderr, "USERPROC: can't read KBDPROC's go-ahead '%c'\n", go_char);
		userproc_exit (3, 0);
	}
	#ifdef DEBUG_RTS
		if (trace) {
			fputs ("USERPROC: Got KBDPROC go-ahead\n", stderr); fflush (stderr);
		}
	#endif
} /* user_to_kbd_sync */


static void user_stop_signal (void)
{
	/* Advanced Programming in the UNIX Environment, Stevens; P321 */
	#ifdef DEBUG_RTS
		if (trace) {
			fputs("USERPROC: Stopped\n", stderr); fflush(stderr);
		}
	#endif

	/* set continue handler for after stop */
	signal (SIGCONT, (SigParam)user_cont_signal);

	#ifdef DEBUG_RTS
		if (trace) {
			fputs("USERPROC: Before kill()\n",stderr); fflush(stderr);
		}
	#endif

	if (stdin_is_tty) {
		restore_tty_state ();
		#ifdef DEBUG_RTS
			if (trace) {
				fputs("USERPROC: Restored tty state\n", stderr); fflush(stderr);
			}
		#endif
	}

	/* Stop the KBDPROC */
	kill (kbdproc_pid, SIGSTOP);

	/* Do the default SIGTSTP action - stop this process */
	kill (getpid(), SIGSTOP);

	#ifdef DEBUG_RTS
		if (trace) {
			fputs("USERPROC: After kill()\n", stderr); fflush(stderr);
		}
	#endif

	#if defined(SIGNAL_TYPE_SYSV)
		/* Re-establish signal handler */
		signal (SIGTSTP, (SigParam)user_stop_signal);
	#endif
}


static void user_cont_signal (void)
{
	/* Advanced Programming in the UNIX Environment, Stevens; P321 */
	sigset_t mask;

	#ifdef DEBUG_RTS
		if (trace) {
			fprintf(stderr, "USERPROC: In user_cont_signal()\n");  fflush(stderr);
		}
	#endif

	/* unblock SIGCONT since it is blocked while we're handling it */
	sigemptyset (&mask);
	sigaddset (&mask, SIGCONT);
	sigprocmask (SIG_UNBLOCK, &mask, NULL);

	if (stdin_is_tty) {
		/* Save terminal state */
		#ifdef DEBUG_RTS
			if (trace) {
				fputs("USERPROC: Saving tty state\n", stderr); fflush(stderr);
			}
		#endif
		save_tty_state ();
	}

	/* Do default SIGCONT action - continue the process */
	signal (SIGCONT, SIG_DFL);
	kill (getpid(), SIGCONT);
}


static void set_user_process_signals (void)
{
	/* Signals used by KROC kernel */

	#ifdef BLOCKING_SYSCALLS
		/* Modified kernel block_sync variable */
		bsigact.sa_handler = (SigParam)user_blocking_handler;
		bsigact.sa_flags = SA_RESTART;
		sigemptyset (&bsigact.sa_mask);
		if (sigaction (SIGBLOCKING, &bsigact, NULL)) {
			fprintf (stderr, "KRoC: unable to set handler for signal SIGBLOCKING (%d)\n", SIGBLOCKING);
			userproc_exit (1, 0);
		}
	#endif
	/* Modifies kernel kbd_sync variable */
	signal (SIGKBD, (SigParam)user_kbd_handler);
	/* Modifies kernel tim_sync variable */
	signal (SIGALRM, (SigParam)user_tim_handler);
	#if 0
		/* Called by user if they panic */
		signal (SIGPANIC, (SigParam)user_panic_handler);
	#endif

	/* Process termination signals */
	signal (SIGHUP, (SigParam)user_good_exit);  /* hangup */
	#if 0
		signal (SIGEMT, (SigParam)user_bad_exit);  /* emulator trap */
	#endif
	signal (SIGTERM, (SigParam)user_good_exit); /* software termination */

	/* Only catch if not being ignored */
	if (signal (SIGINT, SIG_IGN) != SIG_IGN) {
		signal (SIGINT, (SigParam)user_good_exit);    /* interrupt (^C) */
	}
	if (signal (SIGQUIT, SIG_IGN) != SIG_IGN) {
		signal (SIGQUIT, (SigParam)user_good_exit);   /* quit */
	}


	/* 'Error' signals */
	signal (SIGILL, (SigParam)user_trap_handler);  /* illegal instruction */
	signal (SIGBUS, (SigParam)user_trap_handler);  /* bus error */
	faulted = 0;
	#ifdef USEFUL_SEGFAULT
		segv_action.sa_sigaction = user_segv_handler;
		segv_action.sa_flags = SA_SIGINFO;
		sigemptyset (&segv_action.sa_mask);
		if (sigaction (SIGSEGV, &segv_action, NULL) < 0) {
			fprintf (stderr, "KRoC: unable to sigaction SEGV, using normal.\n");
			signal (SIGSEGV, (SigParam)user_trap_handler);
		}
	#else	/* !USEFUL_SEGFAULT */
		#ifndef DEFAULT_CORES
			signal (SIGSEGV, (SigParam)user_trap_handler); /* segmentation violation */
		#endif
	#endif	/* !USEFUL_SEGFAULT */
	#if defined(target_cpu_alpha)
		signal (SIGTRAP, (SigParam)user_trap_handler);
	#endif
	#ifdef USEFUL_FPEDEBUG
		fpe_action.sa_sigaction = user_fpe_handler;
		fpe_action.sa_flags = SA_SIGINFO;
		sigemptyset (&fpe_action.sa_mask);
		if (sigaction (SIGFPE, &fpe_action, NULL) < 0) {
			fprintf (stderr, "KRoC: unable to sigaction FPE, using normal.\n");
			signal (SIGFPE, (SigParam)user_fp_handler);
		}
	#else	/* !USEFUL_FPEDEBUG */
		signal (SIGFPE, (SigParam)user_fp_handler);
	#endif	/* !USEFUL_FPEDEBUG */

	sigemptyset (&zero_mask);
	sigemptyset (&sig_mask);
	#ifdef BLOCKING_SYSCALLS
		sigaddset (&sig_mask, SIGBLOCKING);
	#endif
	sigaddset (&sig_mask, SIGKBD);
	sigaddset (&sig_mask, SIGALRM);

	/* stop signal generated from keyboard (^Z) */
	/* Only catch this if running with job-control */
	if (signal (SIGTSTP, SIG_IGN) == SIG_DFL) {
		signal (SIGTSTP, (SigParam)user_stop_signal);
	}
}


void impossible (int n, int occam_iptr)  /* (probably not) called from scheduler */
{       
	fprintf (stderr, "\nKROC: *** Impossible thing number %d happened at occam address %08X***\n", n, occam_iptr);
	fputs ("KROC: Please report thing number with your bug report\n", stderr);
	userproc_exit (4, 0);
} /* impossible */


void status (int erfl)                   /* state of error flag */
{       
	if (erfl) {
		fputs ("KROC: Error flag set\n", stderr);
		fflush (stderr);
	}
} /* status */


void dead (int erfl)                     /* called by scheduler */
{       
	fputs ("\nKROC: Program deadlocked (no processes to run)\n", stderr);
	fflush (stderr);              /* still in raw mode */
	status (erfl);
	userproc_exit (2, 0);
} /* dead */

void dead_quiet (int erfl)
{
	fflush (stderr);
	status (erfl);
	userproc_exit (2, 0);
} /* dead_quiet */


unsigned int utime (void)                            /* user (processor) time */
{
	#if defined(HAVE_GETRUSAGE)
		struct rusage usage;          /* Resolution is 10000 usec */

		(void) getrusage (RUSAGE_SELF, &usage);
		return usage.ru_utime.tv_sec*1000000 + usage.ru_utime.tv_usec;
	#elif defined(HAVE_TIMES)
		#include <sys/times.h>

		struct tms buffer;

		(void)times(&buffer);  /* Counts in 1/HZ seconds (where HZ is 60, for SunOS) */
		return buffer.tms_utime * (1000000 / HZ);
	#else
		return rtime(); /* No processor time available */
	#endif
} /* utime */


#if defined(target_cpu_sparc)
unsigned int rtime (void)                    /* real (clock) time    */
{
	struct timeval tp;            /* resolution 1 usec?   */
	(void) gettimeofday (&tp, 0);
	return tp.tv_sec*1000000 + tp.tv_usec;
} /* rtime */
#elif defined(target_cpu_386)
unsigned int rtime (void)                    /* real (clock) time    */
{
	struct timeval tp;            /* resolution 1 usec?   */
	(void) gettimeofday (&tp, 0);
	#ifdef DEBUG_RTS
		fprintf(stderr,"rtime: time returning is %d\n",tp.tv_sec*1000000 + tp.tv_usec);
	#endif
	return (tp.tv_sec*1000000) + tp.tv_usec;
} /* rtime */
#elif defined(target_cpu_alpha)
unsigned int rtime (void)                    /* real (clock) time    */
{
	struct timespec ts;
	unsigned long msec;
	unsigned int occamtime;

	if (getclock(TIMEOFDAY, &ts)) {
		fprintf(stderr, "getclock(TIMEOFDAY) failed - ");
		perror(NULL);
		userproc_exit(1, 0);
	};
	/* Do not change this calculation - merging the lines will not work
	   due to C's handling of int / unsigned int arithmetic */
	msec = ((unsigned int)ts.tv_nsec)/1000;
	msec += ts.tv_sec * 1000000;
	occamtime = (unsigned int)(msec & 0xffffffff);
	return occamtime;
} /* rtime */
#endif


/*
  This fixed constant of 10ms is added to each setitimer call to
  kludge (get round) a Solaris 2.3 (or later?) bug in returning too early.
  */
#ifdef SOLARIS_TIMER_BUG
	#include <sys/param.h>
	static int quantum; 
#endif


#ifndef HAVE_UALARM
	#include <sys/time.h>
int ualarm(int delay, int interval)
{
	struct itimerval itv;
	#ifdef DEBUG_RTS
		fprintf (stderr, "ualarm: delay = %d, interval = %d\n", delay, interval);
	#endif
	#ifdef SOLARIS_TIMER_BUG
		delay += quantum; 
	#endif
	itv.it_value.tv_sec =delay / 1000000;
	itv.it_value.tv_usec = delay % 1000000;

	itv.it_interval.tv_sec=0;
	itv.it_interval.tv_usec=interval;
	return setitimer(ITIMER_REAL, &itv, (struct itimerval*)NULL);
}
#endif /* HAVE_UALARM */


/* Called from assembler kernel */
int safe_pause (void)
{       
	sigset_t old_mask;
	#ifdef DEBUG_RTS
		fprintf(stderr, "USERPROC: safepause() entered\n");
	#endif
	if (sigprocmask (SIG_BLOCK, &sig_mask, &old_mask) < 0) { /* critical region */
		fputs ("USERPROC: sigprocmask (SIG_BLOCK, ... , ...) failed.\n", stderr);
		userproc_exit (3, 0);
	}

	#if defined(target_cpu_alpha)
		while ((sf.s.kbd_sync & 1) == 0 || (sf.s.tim_sync & 1) == 0)
	#else
		while (sf.combined_sync == 0)  /* all in one word now */
	#endif
	{ /* signal hasn't happened yet */
		if (sigsuspend (&zero_mask) != -1) {                /* wait for it */
			fputs ("USERPROC: sigsuspend (...) failed.\n", stderr);
			userproc_exit (3, 0);
		}

	#if defined(target_cpu_alpha)
		if ((sf.s.kbd_sync & 1) == 1 || (sf.s.tim_sync & 1) == 1)
	#else
		if (sf.combined_sync != 0)  /* all in one word now */
	#endif
			break;
	} /* while */


	if (sigprocmask (SIG_SETMASK, &old_mask, NULL) < 0) {   /* end of critical */
		fputs ("USERPROC: sigprocmask (SIG_SETMASK, ..., ...) failed.\n", stderr);
		userproc_exit (3, 0);
	}
	#ifdef DEBUG_RTS
		fprintf(stderr, "USERPROC: safepause about to exit (return 0)\n"); /* Jim */
	#endif
	return 0;
} /* safe_pause */


void user_process (int kbd_pid, int to_kbd_fd, int from_kbd_fd, bool is_a_tty)
{
	int erfl;

	/* kroc channels */
	kroc_in = stdin;
	kroc_out = stdout;
	kroc_err = stderr;

	/* user_process - run occam */
	kbdproc_pid = kbd_pid;
	to_kbdproc = to_kbd_fd;
	from_kbdproc = from_kbd_fd;
	stdin_is_tty = is_a_tty;

	#ifdef SOLARIS_TIMER_BUG
		quantum=1000000/HZ;
	#endif

	set_user_process_signals ();

	kbd_char_signalled = 0;
	#ifdef DEBUG_RTS
		if (trace) {
			fputs ("USERPROC: About to synchronise with KBDPROC\n", stderr);
			fflush (stderr);
		}
	#endif

	user_to_kbd_sync ();

	#ifdef DEBUG_RTS
		if (trace) {
			fputs ("USERPROC: Synchronised ok\n", stderr);
			fflush (stderr);
		}
	#endif

	/* really do the occam thing */
	/* Jim Added underscore to fix assembler clashing with 386 enter instruction */
	#ifdef CSPLIB
		erfl = c_enter();              /* run the C program */
	#else
		#ifdef DEBUG_KERNEL
			fprintf( stderr, "USERPROC: about to call _enter() (%p)\n", _enter );
		#endif
		fflush( stderr );
		erfl = _enter ();              /* run the occam program */ 
		#ifdef BLOCKING_DEBUG
			fprintf (stderr, "KRoC: OUT OF _enter!\n");
		#endif
	#endif

	fflush (stdout);              /* clear buffer */
	status (erfl);
	userproc_exit (erfl, 0);

} /* user_process */



