/*
 *	CSPlib types
 *	Copyright (C) 1998 Jim Moores
 *
 *	This program is free software; you can redistribute it and/or modify
 *	it under the terms of the GNU General Public License as published by
 *	the Free Software Foundation; either version 2 of the License, or
 *	(at your option) any later version.
 *
 *	This program is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU General Public License for more details.
 *
 *	You should have received a copy of the GNU General Public License
 *	along with this program; if not, write to the Free Software
 *	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*
 * Copyright (C) Jim Moores 1998
 * This file defines the main data structures used by UKCThreads (specifically
 * CSPlib).  These are the process count and the process descriptor.  Some
 * offsets into the structure are also defined as macros.
 */

#ifndef CSPLIB_TYPES_H
#define CSPLIB_TYPES_H

struct _ProcessCount {
  word *cont_address;
  unsigned int count;
  word *caller_stackptr;
  word *caller_frameptr;
};

typedef struct _ProcessCount ProcessCount;

/* Naming conventions.
 * -------------------
 * Macros with names starting MacroIndirect are used in wrapper functions
 * Reference to functions with names prefixed `M' are functions that need
 * to be passed the process pointer.  A separate set of macros is then
 * defined without the M prefix that include an extra paramter.  This is
 * why the process pointer must always be called `p'.  For example:
 *   int MProcAlt(Process *p, Channel *c1, ...)
 * has an interface macro:
 *   #define ProcAlt(args...) MProcAlt(p, ## args)
 * so that:
 *   ProcAlt(a,b,c,d);
 * becomes (after preprocessing):
 *   MProcAlt(p,a,b,c,d);
 * so the programmer needn't remember to put in the process pointer.
 *
 * It has been suggested to me that it would be better to get rid of the
 * process pointer all together and use a global variable to hold it
 * between kernel calls.  This would certainly be a neater solution from
 * the user's point of view because the mysterious first paramter of
 * processes dissapears.  The problem is, however, that it would be
 * difficult to scale this to a full SMP version.  There would need to
 * be a separate global variable used by each processor, and there would
 * be an overhead in finding out which processor the code was currently
 * running on, and this would need to be done all the time, on every
 * kernel call.  I therefore it was better not to have to change back to
 * the current system later on (which would require user source changes).
 */

struct _Process {
  ProcessCount *usage_info;
  word workspace[22];

  /*
      STRUCTURE OF PROCESS DESCRIPTOR
      The bottom 6 words are used by the kernel as if it were an occam
      processes workspace (equivalent to a stack frame).  The various
      extra pieces of information needed for C processes are stored in
      space above these 6 words.
                                                           OFFSETS
                                                     Asm Wptr   C Array
      +------------------------------+<----^-----
 +21  | FP Context (108 bytes)       |    /|\           84         26
      +------------------------------+     |
 +20  | EDI                          |     |            80         25
      +------------------------------+     |
 +19  | ESI                          |     |            76         24
      +------------------------------+ Only used if
 +18  | EDX                          | full context     72         23
      +------------------------------+ save required
 +17  | ECX                          |     |            68         22
      +------------------------------+     |
 +16  | EBX                          |     |            64         21
      +------------------------------+     |
 +15  | EAX                          |    \|/           60         20
      +------------------------------+<----v------
 +14  | RESERVED                     |                  56         19
      +------------------------------+
 +13  | RESERVED                     |                  52         18
      +------------------------------+
 +12  | EBP (frame pointer)          |                  48         17
      +------------------------------+
 +11  | ESP (stack pointer)          |                  44         16
      +------------------------------+
 +10  | ProcessCount structure       |                  40         15
      + for if the process is        +
  +9  | ProcRun'd without being      |                  36         14
      + ProcJoined (keeping it here  +
  +8  | prevents mallocing it).      |                  32         13
      +                              +
  +7  |                              | <- ptr here      28         12
      +------------------------------+
  +6  | Ptr to suspend q (if on one) |                  24         11
      +------------------------------+
  +5  | process priority (not used)  |                  20         10
      +------------------------------+
  +4  | Parameter stack depth        |                  16          9
      +------------------------------+
  +3  | pointer to function to start |                  12          8
      +------------------------------+
  +2  | pointer to this process desc |                   8          7
      +------------------------------+
  +1  | stack pointer for function   |                   4          6
      +------------------------------+
   0  | Temp/Data to transfer/offset |  <-- WPtr         0          5
      +------------------------------+      passed to kernel
  -1  | instruction pointer          |                  -4          4
      +------------------------------+
  -2  | ptr to next process in queue |                  -8          3
      +------------------------------+
  -3  | ALT status or valid address  |                 -12          2
      +------------------------------+
  -4  | Validity of time at -5       |                 -16          1
      +------------------------------+
  -5  | A time to wait during an ALT |                 -20          0
      +------------------------------+

  */
  word *stack;  /* needed to dealloc space */
  word *wptr;
  int paramwords; /* the number of works in the paramter list - */
                  /* used by ProcParam(). */
  int stacksize; /* size of stack _in words_ */
};

typedef struct _Process Process;

#ifdef CSPLIB
struct _Semaphore {
  word value;
  prioratised_queue q;
};

typedef struct _Semaphore Semaphore;
#endif

/* Some offsets into the above table to allow things to be moved about */
/* by not hardwiring the values in. */
#ifdef CPU_TIMERS
#define base_offset 1
#else
#define base_offset 0
#endif

#define CWptrOffset 5+base_offset
#define CStackPtr 6+base_offset
#define CProcessDescPointer 7+base_offset
#define CStartFunction 8+base_offset
#define CPriority 10+base_offset
#define COnRunQ 11+base_offset
#define CProcCount 12+base_offset
#define StackPtrRegSave "44"
#define FramePtrRegSave "48"

#endif /* csplib_types.h */
