/*
 *	cfile.c - C parts of the occam file library
 *	Copyright (C) 2000 Fred Barnes <frmb2@ukc.ac.uk>
 *
 *	This program is free software; you can redistribute it and/or modify
 *	it under the terms of the GNU General Public License as published by
 *	the Free Software Foundation; either version 2 of the License, or
 *	(at your option) any later version.
 *
 *	This program is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU General Public License for more details.
 *
 *	You should have received a copy of the GNU General Public License
 *	along with this program; if not, write to the Free Software
 *	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <stdio.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <fcntl.h>


/*
 *	void r_check_access (char *fname, int flen, int what, int *result)
 *	Checks for file access "rights"
 */
static __inline__ void r_check_access (char *fname, int flen, int what, int *result)
{
	char pbuffer[FILENAME_MAX];
	int x;

	if (flen >= (FILENAME_MAX - 1)) {
		x = FILENAME_MAX - 1;
	} else {
		x = flen;
	}
	memcpy (pbuffer, fname, x);
	pbuffer[x] = '\0';
	*result = access (pbuffer, what);
}


/*
 *	void r_size (char *fname, int flen, int *result)
 *	gets the size of a file
 */
static __inline__ void r_size (char *fname, int flen, int *result)
{
	char pbuffer[FILENAME_MAX];
	int x;
	struct stat st_buf;

	if (flen >= (FILENAME_MAX - 1)) {
		x = FILENAME_MAX - 1;
	} else {
		x = flen;
	}
	memcpy (pbuffer, fname, x);
	pbuffer[x] = '\0';
	x = stat (pbuffer, &st_buf);
	if (x < 0) {
		*result = -1;
	} else {
		*result = (int)st_buf.st_size;
	}
}


/*
 *	void r_open (char *fname, int flen, int mode, int *fd)
 *	opens a file
 */
static __inline__ void r_open (char *fname, int flen, int mode, int *fd)
{
	char pbuffer[FILENAME_MAX];
	int x;

	if (flen >= (FILENAME_MAX - 1)) {
		x = FILENAME_MAX - 1;
	} else {
		x = flen;
	}
	memcpy (pbuffer, fname, x);
	pbuffer[x] = '\0';
	*fd = open (pbuffer, mode);
}


/*
 *	void r_open3 (char *fname, int flen, int mode, int perm, int *fd)
 *	opens a file with permission if creating
 */
static __inline__ void r_open3 (char *fname, int flen, int mode, int perm, int *fd)
{
	char pbuffer[FILENAME_MAX];
	int x;

	if (flen >= (FILENAME_MAX - 1)) {
		x = FILENAME_MAX - 1;
	} else {
		x = flen;
	}
	memcpy (pbuffer, fname, x);
	pbuffer[x] = '\0';
	*fd = open (pbuffer, mode, perm);
}


/*
 *	void r_pipe (int *fd_0, int *fd_1, int *result)
 *	Creates a pair of pipe descriptors
 */
static __inline__ void r_pipe (int *fd_0, int *fd_1, int *result)
{
	int p[2];

	*result = pipe (p);
	if (!*result) {
		*fd_0 = p[0];
		*fd_1 = p[1];
	}
}


/*
 *	void r_dup2 (int old_fd, int new_fd, int *result)
 *	Duplicates a file descriptor
 */
static __inline__ void r_dup2 (int old_fd, int new_fd, int *result)
{
	*result = dup2 (old_fd, new_fd);
}


/*
 *	r_read (int fd, char *buffer, int bufsize, int *result)
 *	reads from a file
 */
static __inline__ void r_read (int fd, char *buffer, int bufsize, int *result)
{
	*result = read (fd, buffer, bufsize);
}


/*
 *	void r_write (int fd, char *buffer, int bufsize, int *result)
 *	writes to a file
 */
static __inline__ void r_write (int fd, char *buffer, int bufsize, int *result)
{
	*result = write (fd, buffer, bufsize);
}


/*
 *	void r_seek (int fd, int offset, int whence, int *result)
 *	seeks to a specified point in a file
 */
static __inline__ void r_seek (int fd, int offset, int whence, int *result)
{
	*result = (int)lseek (fd, offset, whence);
}


/*
 *	void r_close (int fd)
 *	closes a file
 */
static __inline__ void r_close (int fd)
{
	close (fd);
}


/*
 *	void r_mkdir (char *path, int path_len, int perm, int *result)
 *	creates a directory
 */
static __inline__ void r_mkdir (char *path, int path_len, int perm, int *result)
{
	char fname[FILENAME_MAX];

	if (FILENAME_MAX <= path_len) {
		path_len = FILENAME_MAX - 1;
	}
	memcpy (fname, path, path_len);
	fname[path_len] = '\0';
	*result = mkdir (fname, perm);
}


/*
 *	void r_rmdir (char *path, int path_len, int *result)
 *	removes a directory
 */
static __inline__ void r_rmdir (char *path, int path_len, int *result)
{
	char fname[FILENAME_MAX];

	if (FILENAME_MAX <= path_len) {
		path_len = FILENAME_MAX - 1;
	}
	memcpy (fname, path, path_len);
	fname[path_len] = '\0';
	*result = rmdir (fname);
}


/*
 *	void r_unlink (char *file, int file_len, int *result)
 *	removes a file
 */
static __inline__ void r_unlink (char *file, int file_len, int *result)
{
	char fname[FILENAME_MAX];

	if (FILENAME_MAX <= file_len) {
		file_len = FILENAME_MAX - 1;
	}
	memcpy (fname, file, file_len);
	fname[file_len] = '\0';
	*result = unlink (fname);
}


/*
 *	void r_fd_fd_copy (int src_fd, int dst_fd, int count, int *result)
 *	coppies `count' bytes from `src_fd' to `dst_fd'.  If `count' is 0, data is
 *	coppied until the source EOFs, or an error occurs
 *	`result' gets:	0 if done OK, -1 if src error, -2 if dest error, or -3 if
 *			`count' > 0 and `src_fd' EOFs before it's all been read.
 */
static __inline__ void r_fd_fd_copy (int src_fd, int dst_fd, int count, int *result)
{
	int x, y, z, gone_out, tmp;
	char buffer[1024];

	for (gone_out = 0;;) {
		if (!count) {
			x = 1024;
		} else {
			x = ((count - gone_out) <= 1024) ? (count - gone_out) : 1024;
		}
		y = read (src_fd, buffer, x);
		if (y < 0) {
			/* got error */
			*result = -1;
			return;
		} else if (!y) {
			/* got EOF */
			if (count) {
				*result = -3;
			} else {
				*result = 0;
			}
			return;
		}
		for (z=0; z<y;) {
			tmp = write (dst_fd, buffer, (y - z));
			if (tmp < 0) {
				/* got error */
				*result = -1;
				return;
			}
			z += tmp;
		}
		gone_out += y;
		if (gone_out == count) {
			*result = 0;
			return;
		}
	}
}


void _fl_check_access (int *w)				{ r_check_access ((char *)(w[0]), (int)(w[1]), (int)(w[2]), (int *)(w[3])); }
void _fl_size (int *w)					{ r_size ((char *)(w[0]), (int)(w[1]), (int *)(w[2])); }
void _fl_open (int *w)					{ r_open ((char *)(w[0]), (int)(w[1]), (int)(w[2]), (int *)(w[3])); }
void _fl_open3 (int *w)					{ r_open3 ((char *)(w[0]), (int)(w[1]), (int)(w[2]), (int)(w[3]), (int *)(w[4])); }
void _fl_pipe (int *w)					{ r_pipe ((int *)(w[0]), (int *)(w[1]), (int *)(w[2])); }
void _fl_dup2 (int *w)					{ r_dup2 ((int)(w[0]), (int)(w[1]), (int *)(w[2])); }
void _fl_read (int *w)					{ r_read ((int)(w[0]), (char *)(w[1]), (int)(w[2]), (int *)(w[3])); }
void _fl_write (int *w)					{ r_write ((int)(w[0]), (char *)(w[1]), (int)(w[2]), (int *)(w[3])); }
void _fl_seek (int *w)					{ r_seek ((int)(w[0]), (int)(w[1]), (int)(w[2]), (int *)(w[3])); }
void _fl_close (int *w)					{ r_close ((int)(w[0])); }
void _fl_mkdir (int *w)					{ r_mkdir ((char *)(w[0]), (int)(w[1]), (int)(w[2]), (int *)(w[3])); }
void _fl_rmdir (int *w)					{ r_rmdir ((char *)(w[0]), (int)(w[1]), (int *)(w[2])); }
void _fl_unlink (int *w)				{ r_unlink ((char *)(w[0]), (int)(w[1]), (int *)(w[2])); }
void _fl_fd_fd_copy (int *w)				{ r_fd_fd_copy ((int)(w[0]), (int)(w[1]), (int)(w[2]), (int *)(w[3])); }

