/* Portable timer implementation functions.
 * ----------------------------------------
 * only include these definitions if not using the faster 
 * macro definitions in the header file.
 * There are some extra functions defined at the end of this
 * block - these are the functions used by the kernel, but 
 * are derived from the CERN/UKC spec so that only those need
 * be provided.
 */

#include "timer.h"
#ifdef CPU_TIMERS
/* BEGIN NATIVE TIMER IMPLEMENTATION */
long long Time_hz = 0LL;
long long Time_ticks_per_hz = 1000020LL;
long long Time_ticks_per_sec = 0LL;
long long Time_ticks_per_us = 0LL;
double Time_ticks_per_us_d = 0.0;

Time timeout_alarm;

#define atlevel(X) if (X)
#define MESSAGE 1
#define print(args...) fprintf(stderr, ## args)

void Time_Print(Time t)
{
  printf("Time = %Ld = %Ld secs %Ld us\n",
	 t,
	 t/Time_ticks_per_sec,
	 (t - (t/Time_ticks_per_sec)*Time_ticks_per_sec)/Time_ticks_per_us
	 );
}

int Time_CPUSpeed()
{
  Time t1;
  Time t2;
  int cpu_speed;
  int new_cpu_speed = -1;
  t1 = Time_GetTime();
  usleep(1000*1000);
  t2 = Time_GetTime();
  cpu_speed = ((t2 - t1)/Time_ticks_per_hz);
  fflush(stdout);
  if (abs(cpu_speed-100)<=5) { new_cpu_speed = 100; }
  if (abs(cpu_speed-133)<=5) { new_cpu_speed = 133; }
  if (abs(cpu_speed-166)<=5) { new_cpu_speed = 166; }
  if (abs(cpu_speed-200)<=5) { new_cpu_speed = 200; }
  if (abs(cpu_speed-233)<=5) { new_cpu_speed = 233; }
  if (abs(cpu_speed-266)<=5) { new_cpu_speed = 266; }
  if (abs(cpu_speed-300)<=5) { new_cpu_speed = 300; }
  if (abs(cpu_speed-333)<=5) { new_cpu_speed = 333; }
  if (abs(cpu_speed-366)<=5) { new_cpu_speed = 366; }
  if (abs(cpu_speed-400)<=5) { new_cpu_speed = 400; }
  if (abs(cpu_speed-433)<=5) { new_cpu_speed = 433; }
  if (abs(cpu_speed-450)<=5) { new_cpu_speed = 450; }
  if (abs(cpu_speed-466)<=5) { new_cpu_speed = 466; }
  if (abs(cpu_speed-500)<=5) { new_cpu_speed = 500; }
  return new_cpu_speed;
}

void Time_SetCPUSpeed(int cpu_speed)
{
  atlevel(MESSAGE) print("CPU speed = %d MHz\n",cpu_speed);
  Time_hz = cpu_speed;
  Time_ticks_per_sec = Time_ticks_per_hz * Time_hz;
  Time_ticks_per_us = Time_ticks_per_sec / 1000000LL;
  Time_ticks_per_us_d = Time_ticks_per_sec / 1000000.0;
}

void Time_Init()
{
  int cpu_speed = 0;
  int n=4;
  atlevel(MESSAGE) print("Initialising real-time clock.\n");
  atlevel(MESSAGE) print("Measuring CPU speed: ");
  fflush(stdout);
  do {
    int new_cpu_speed = Time_CPUSpeed();
    print("%d:%d MHz  ",n,new_cpu_speed);
    fflush(stdout);
    if ((new_cpu_speed != cpu_speed)||(new_cpu_speed == -1)) {
      n=3;
      cpu_speed = new_cpu_speed;
    } else {
      n--;
    }
  } while(n>0);
  print("-> %d MHz\n",cpu_speed);
  Time_SetCPUSpeed(cpu_speed);
  Time_ResetAlarm();
}

Time Time_GetTime(void)
{
  Time now;
  __asm__(".byte 0x0f,0x31"
          :"=a" (lo(now)),
           "=d" (hi(now)));
  return now;
}
#ifndef INLINE_TIMERS
#warning Not inlining timer code.

bool Time_PollAlarm()
{
  Time now = Time_GetTime();
  #if 0
    if (Time_AFTER(now, timeout_alarm)) 
      fprintf(stderr, "Time_PollAlarm(): returning true.\n");
  #endif
  return Time_AFTER(now, timeout_alarm);
}
void Time_SetAlarm(Time t)
{
  #if 0
    fprintf(stderr, "Time_SetAlarm: called with time = %Ld", t);
  #endif
  timeout_alarm = t + Time_GetTime();
}

void Time_ResetAlarm(void) {
  timeout_alarm = 0x7FFFFFFFFFFFFFFFLL;
}

Time Time_PLUS(Time t1, Time t2) {
  return (t1+t2);
}

Time Time_MINUS(Time t1, Time t2) {
  return (t1-t2);
}

bool Time_AFTER(Time t1, Time t2) {
  return (t1-t2)>0;
}
#endif
/* END OF NATIVE TIMER IMPLEMENTATION */
#else
/* BEGIN UNIX TIMER IMPLEMENTATION */
long long Time_hz = 1000000LL;
long long Time_ticks_per_hz = 1LL;
long long Time_ticks_per_sec = 1000000LL;
long long Time_ticks_per_us = 1LL;
double Time_ticks_per_us_d = 1.0;
#ifndef INLINE_TIMERS
#warning Not inlining timer code
Time Time_GetTime(void) {
  return rtime();
}

Time Time_PLUS(Time t1, Time t2) {
  return (t1+t2);
}

Time Time_MINUS(Time t1, Time t2) {
  return (t1-t2);
}

bool Time_AFTER(Time t1, Time t2) {
  return (t1-t2)>0LL;
}

bool Time_PollAlarm(void) {
  return (sf.s.tim_sync);
}

void Time_SetAlarm(Time t) {
  ualarm(t, 0);
}

void Time_ResetAlarm(void) {
  ualarm(0,0);
}
#endif
/* END UNIX TIMER IMPLEMENTATION */
#endif

