/*
 *	DEC alpha code disassembly
 *	Copyright (C) 1995 Michael Poole
 *
 *	This program is free software; you can redistribute it and/or modify
 *	it under the terms of the GNU General Public License as published by
 *	the Free Software Foundation; either version 2 of the License, or
 *	(at your option) any later version.
 *
 *	This program is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU General Public License for more details.
 *
 *	You should have received a copy of the GNU General Public License
 *	along with this program; if not, write to the Free Software
 *	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*{{{  includes*/
#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif
#include "includes.h"
#include "suplib.h"
#include "gendeca.h"
#include "code1def.h"
#include "disalpha.h"
/*}}}*/
/*{{{  public variables*/
PUBLIC BIT32 * code_ptr;
/*}}}*/
/*{{{  local variables*/
PRIVATE char operand_string [32];
PRIVATE char instr_string [32];
/*}}}*/
/*{{{  DEC alpha code disassembly*/
/*{{{  PUBLIC look_at_alpha_instruction(ilword, instr, operands, asm_only, special)*/
/***************************************************************************
 *
 *  look_at_alpha_instruction disassembles the instruction in ilword
 *                    into an instruction name and an operands string.
 *                    The instruction len is returned.
 *                    special: 1=RET, 2=reljump, 0=other
 *
 **************************************************************************/
PUBLIC void look_at_alpha_instruction (const BIT32 ilword,
        const char **instr, const char **operands,
        const INT32 disaddr,
        const BOOL asm_only, int * special)
{
  enum {IA_mem, IA_fmem, IA_memfunc, IA_membranch, IA_branch, IA_operate, IA_fpop, IA_pal};
  char *instname = &instr_string[0];
  int iaformat = -1;
  BOOL fanotused = FALSE;
  BOOL fbnotused = FALSE;
  int opcode = (ilword >> 26);
  operand_string [0] = '\0';
  *special = 0;
  /*{{{  extract opcode, operands from ilword*/
  switch (opcode)
  {
    /*{{{  0x*/
    case 0x00:
    {
      strcpy (instname, "PAL");
      iaformat = IA_pal;
      break;
    }
    case 0x08:
    {
      strcpy (instname, "LDA");
      iaformat = IA_mem;
      break;
    }
    case 0x09:
    {
      strcpy (instname, "LDAH");
      iaformat = IA_mem;
      break;
    }
    case 0x0B:
    {
      strcpy (instname, "LDQ_U");
      iaformat = IA_mem;
      break;
    }
    case 0x0F:
    {
      strcpy (instname, "STQ_U");
      iaformat = IA_mem;
      break;
    }
    /*}}}*/
    /*{{{  1x*/
    case 0x10:
    {
      BYTE intfunc = (ilword >> 5) & 0x7F;
      iaformat = IA_operate;
      /*{{{  integer arithmetic*/
      switch (intfunc)
      {
        case 0x00:
        {
          strcpy (instname, "ADDL");
          break;
        }
        case 0x09:
        {
          strcpy (instname, "SUBL");
          break;
        }
        case 0x20:
        {
          strcpy (instname, "ADDQ");
          break;
        }
        case 0x29:
        {
          strcpy (instname, "SUBQ");
          break;
        }
        case 0x40:
        {
          strcpy (instname, "ADDLV");
          break;
        }
        case 0x49:
        {
          strcpy (instname, "SUBLV");
          break;
        }
        case 0x60:
        {
          strcpy (instname, "ADDQV");
          break;
        }
        case 0x69:
        {
          strcpy (instname, "SUBQV");
          break;
        }
        case 0x0F:
        {
          strcpy (instname, "CMPBGE");
          break;
        }
        case 0x1D:
        {
          strcpy (instname, "CMPULT");
          break;
        }
        case 0x2D:
        {
          strcpy (instname, "CMPEQ");
          break;
        }
        case 0x3D:
        {
          strcpy (instname, "CMPULE");
          break;
        }
        case 0x4D:
        {
          strcpy (instname, "CMPLT");
          break;
        }
        case 0x6D:
        {
          strcpy (instname, "CMPLE");
          break;
        }
        case 0x02:
        {
          strcpy (instname, "S4ADDL");
          break;
        }
        case 0x0B:
        {
          strcpy (instname, "S4SUBL");
          break;
        }
        case 0x12:
        {
          strcpy (instname, "S8ADDL");
          break;
        }
        case 0x1B:
        {
          strcpy (instname, "S8SUBL");
          break;
        }
        case 0x22:
        {
          strcpy (instname, "S4ADDQ");
          break;
        }
        case 0x2B:
        {
          strcpy (instname, "S4SUBQ");
          break;
        }
        case 0x32:
        {
          strcpy (instname, "S8ADDQ");
          break;
        }
        case 0x3B:
        {
          strcpy (instname, "S8SUBQ");
          break;
        }
      }
      /*}}}*/
      break;
    }
    case 0x11:
    {
      BYTE logfunc = (ilword >> 5) & 0x7F;
      iaformat = IA_operate;
      /*{{{  logical operations*/
      switch (logfunc)
      {
        case 0x00:
        {
          strcpy (instname, "AND");
          break;
        }
        case 0x08:
        {
          strcpy (instname, "BIC");
          break;
        }
        case 0x20:
        {
          strcpy (instname, "BIS");
          break;
        }
        case 0x28:
        {
          strcpy (instname, "ORNOT");
          break;
        }
        case 0x40:
        {
          strcpy (instname, "XOR");
          break;
        }
        case 0x48:
        {
          strcpy (instname, "EQV");
          break;
        }
        case 0x14:
        {
          strcpy (instname, "CMOVLBS");
          break;
        }
        case 0x16:
        {
          strcpy (instname, "CMOVLBC");
          break;
        }
        case 0x24:
        {
          strcpy (instname, "CMOVEQ");
          break;
        }
        case 0x26:
        {
          strcpy (instname, "CMOVNE");
          break;
        }
        case 0x44:
        {
          strcpy (instname, "CMOVLT");
          break;
        }
        case 0x46:
        {
          strcpy (instname, "CMOVGE");
          break;
        }
        case 0x64:
        {
          strcpy (instname, "CMOVLE");
          break;
        }
        case 0x66:
        {
          strcpy (instname, "CMOVGT");
          break;
        }
      }
      /*}}}*/
      break;
    }
    case 0x12:
    {
      BYTE shiftfunc = (ilword >> 5) & 0x7F;
      iaformat = IA_operate;
      /*{{{  shift etc operations*/
      switch (shiftfunc)
      {
        /*{{{  shifts*/
        case 0x34:
        {
          strcpy (instname, "SRL");
          break;
        }
        case 0x39:
        {
          strcpy (instname, "SLL");
          break;
        }
        case 0x3C:
        {
          strcpy (instname, "SRA");
          break;
        }
        /*}}}*/
        /*{{{  byte extracts*/
        case 0x06:
        {
          strcpy (instname, "EXTBL");
          break;
        }
        case 0x16:
        {
          strcpy (instname, "EXTWL");
          break;
        }
        case 0x26:
        {
          strcpy (instname, "EXTLL");
          break;
        }
        case 0x36:
        {
          strcpy (instname, "EXTQL");
          break;
        }
        case 0x5A:
        {
          strcpy (instname, "EXTWH");
          break;
        }
        case 0x6A:
        {
          strcpy (instname, "EXTLH");
          break;
        }
        case 0x7A:
        {
          strcpy (instname, "EXTQH");
          break;
        }
        /*}}}*/
        /*{{{  byte inserts*/
        case 0x0B:
        {
          strcpy (instname, "INSBL");
          break;
        }
        case 0x1B:
        {
          strcpy (instname, "INSWL");
          break;
        }
        case 0x2B:
        {
          strcpy (instname, "INSLL");
          break;
        }
        case 0x3B:
        {
          strcpy (instname, "INSQL");
          break;
        }
        case 0x57:
        {
          strcpy (instname, "INSWH");
          break;
        }
        case 0x67:
        {
          strcpy (instname, "INSLH");
          break;
        }
        case 0x77:
        {
          strcpy (instname, "INSQH");
          break;
        }
        /*}}}*/
        /*{{{  byte masking*/
        case 0x02:
        {
          strcpy (instname, "MSKBL");
          break;
        }
        case 0x12:
        {
          strcpy (instname, "MSKWL");
          break;
        }
        case 0x22:
        {
          strcpy (instname, "MSKLL");
          break;
        }
        case 0x32:
        {
          strcpy (instname, "MSKQL");
          break;
        }
        case 0x52:
        {
          strcpy (instname, "MSKWH");
          break;
        }
        case 0x62:
        {
          strcpy (instname, "MSKLH");
          break;
        }
        case 0x72:
        {
          strcpy (instname, "MSKQH");
          break;
        }
        /*}}}*/
        /*{{{  zeroing*/
        case 0x30:
        {
          strcpy (instname, "ZAP");
          break;
        }
        case 0x31:
        {
          strcpy (instname, "ZAPNOT");
          break;
        }
        /*}}}*/
      }
      /*}}}*/
      break;
    }
    case 0x13:
    {
      BYTE multfunc = (ilword >> 5) & 0x7F;
      iaformat = IA_operate;
      /*{{{  multiply operations*/
      switch (multfunc)
      {
        case 0x00:
        {
          strcpy (instname, "MULL");
          break;
        }
        case 0x20:
        {
          strcpy (instname, "MULQ");
          break;
        }
        case 0x30:
        {
          strcpy (instname, "UMULH");
          break;
        }
        case 0x40:
        {
          strcpy (instname, "MULLV");
          break;
        }
        case 0x60:
        {
          strcpy (instname, "MULQV");
          break;
        }

      }
      /*}}}*/
      break;
    }
    case 0x15:
    {
      /* BYTE ieeefunc = (ilword >> 5) & 0x7F; */
      iaformat = IA_fpop;
      /*{{{  vax fp operations*/
      /*}}}*/
      break;
    }
    case 0x16:
    {
      BYTE fpfunc = (ilword >> 5) & 0x7FF;
      iaformat = IA_fpop;
      fanotused = FALSE;
      /*{{{  ieee fp operations*/
      switch (fpfunc)
      {
        default:
        {
          int fpu_mode = (((fpfunc ^ 0x80) >> 6) & 3);
          switch ((fpfunc & 0x73F) ^ 0x80)
          {
            /*{{{  arithmetic etc with one of 4 rounding modes*/

            case 0x080:
            {
              strcpy (instname, "ADDS");
              break;
            }
            case 0x0A0:
            {
              strcpy (instname, "ADDT");
              break;
            }
            case 0x081:
            {
              strcpy (instname, "SUBS");
              break;
            }
            case 0x0A1:
            {
              strcpy (instname, "SUBT");
              break;
            }
            case 0x082:
            {
              strcpy (instname, "MULS");
              break;
            }
            case 0x0A2:
            {
              strcpy (instname, "MULT");
              break;
            }
            case 0x083:
            {
              strcpy (instname, "DIVS");
              break;
            }
            case 0x0A3:
            {
              strcpy (instname, "DIVT");
              break;
            }
            case 0x0BC:
            {
              strcpy (instname, "CVTQS");
              fanotused = TRUE;
              break;
            }
            case 0x0BE:
            {
              strcpy (instname, "CVTQT");
              fanotused = TRUE;
              break;
            }
            case 0x0AC:
            {
              strcpy (instname, "CVTTS");
              fanotused = TRUE;
              break;
            }
            case 0x0AF:
            {
              strcpy (instname, "CVTTQ");
              fanotused = TRUE;
              break;
            }
            /*}}}*/
          }
          switch (fpu_mode)
          {
            case FPU_Z:  strcat (instname, "C"); break;
            case FPU_M:  strcat (instname, "M"); break;
            case FPU_P:  strcat (instname, "D"); break;
          }
          break;
        }
        case 0x0A4:
        {
          strcpy (instname, "CMPTUN");
          break;
        }
        case 0x0A5:
        {
          strcpy (instname, "CMPTEQ");
          break;
        }
        case 0x0A6:
        {
          strcpy (instname, "CMPTLT");
          break;
        }
        case 0x0A7:
        {
          strcpy (instname, "CMPTLE");
          break;
        }
      }
      /*}}}*/
      break;
    }
    case 0x17:
    {
      BYTE fpfunc = (ilword >> 5) & 0x7FF;
      iaformat = IA_fpop;
      fanotused = FALSE;
      /*{{{  fp operate instructions*/
      switch (fpfunc)
      {
        case 0x020:
        {
          strcpy (instname, "CPYS");
          break;
        }
        case 0x021:
        {
          strcpy (instname, "CPYSN");
          break;
        }
        case 0x022:
        {
          strcpy (instname, "CPYSE");
          break;
        }
        case 0x010:
        {
          strcpy (instname, "CVTLQ");
          fanotused = TRUE;
          break;
        }
        case 0x030:
        {
          strcpy (instname, "CVTQL");
          fanotused = TRUE;
          break;
        }
        case 0x024:
        {
          strcpy (instname, "MT_FPCR");
          fanotused = TRUE;
          fbnotused = TRUE;
          break;
        }
        case 0x025:
        {
          strcpy (instname, "MF_FPCR");
          fanotused = TRUE;
          fbnotused = TRUE;
          break;
        }
      }
      /*}}}*/
      break;
    }
    case 0x18:
    {
      INT16 mfunc = ilword & 0xFFFF;
      iaformat = IA_memfunc;
      /*{{{  memfunc operations*/
      switch (mfunc)
      {
      case 0:
          strcpy (instname, "TRAPB"); break;
      }
      /*}}}*/
      break;
    }
    case 0x1A:
    {
      int jtype = (ilword >> 14) & 3;
      iaformat = IA_membranch;
      /*{{{  computed jump operations*/
      switch (jtype)
      {
        case 0:
        {
          strcpy (instname, "JMP");
          break;
        }
        case 1:
        {
          strcpy (instname, "JSR");
          break;
        }
        case 3:
        {
          strcpy (instname, "JSR_COROUTINE");
          break;
        }
        case 2:
        {
          strcpy (instname, "RET");
          *special = 1;
          break;
        }
      }
      /*}}}*/
      break;
    }
    /*}}}*/
    /*{{{  2x*/
    case 0x20:
    {
      strcpy (instname, "LDF");
      iaformat = IA_fmem;
      break;
    }
    case 0x21:
    {
      strcpy (instname, "LDG");
      iaformat = IA_fmem;
      break;
    }
    case 0x22:
    {
      strcpy (instname, "LDS");
      iaformat = IA_fmem;
      break;
    }
    case 0x23:
    {
      strcpy (instname, "LDT");
      iaformat = IA_fmem;
      break;
    }
    case 0x24:
    {
      strcpy (instname, "STF");
      iaformat = IA_fmem;
      break;
    }
    case 0x25:
    {
      strcpy (instname, "STG");
      iaformat = IA_fmem;
      break;
    }
    case 0x26:
    {
      strcpy (instname, "STS");
      iaformat = IA_fmem;
      break;
    }
    case 0x27:
    {
      strcpy (instname, "STT");
      iaformat = IA_fmem;
      break;
    }
    case 0x28:
    {
      strcpy (instname, "LDL");
      iaformat = IA_mem;
      break;
    }
    case 0x29:
    {
      strcpy (instname, "LDQ");
      iaformat = IA_mem;
      break;
    }
    case 0x2A:
    {
      strcpy (instname, "LDL_L");
      iaformat = IA_mem;
      break;
    }
    case 0x2B:
    {
      strcpy (instname, "LDQ_L");
      iaformat = IA_mem;
      break;
    }
    case 0x2C:
    {
      strcpy (instname, "STL");
      iaformat = IA_mem;
      break;
    }
    case 0x2D:
    {
      strcpy (instname, "STQ");
      iaformat = IA_mem;
      break;
    }
    case 0x2E:
    {
      strcpy (instname, "STL_C");
      iaformat = IA_mem;
      break;
    }
    case 0x2F:
    {
      strcpy (instname, "STQ_C");
      iaformat = IA_mem;
      break;
    }
    /*}}}*/
    /*{{{  3x*/
    case 0x30:
    {
      strcpy (instname, "BR");
      iaformat = IA_branch;
      break;
    }
    case 0x31:
    {
      strcpy (instname, "FBEQ");
      iaformat = IA_branch;
      break;
    }
    case 0x32:
    {
      strcpy (instname, "FBLT");
      iaformat = IA_branch;
      break;
    }
    case 0x33:
    {
      strcpy (instname, "FBLE");
      iaformat = IA_branch;
      break;
    }
    case 0x34:
    {
      strcpy (instname, "BSR");
      iaformat = IA_branch;
      break;
    }
    case 0x35:
    {
      strcpy (instname, "FBNE");
      iaformat = IA_branch;
      break;
    }
    case 0x36:
    {
      strcpy (instname, "FBGE");
      iaformat = IA_branch;
      break;
    }
    case 0x37:
    {
      strcpy (instname, "FBGT");
      iaformat = IA_branch;
      break;
    }
    case 0x38:
    {
      strcpy (instname, "BLBC");
      iaformat = IA_branch;
      break;
    }
    case 0x39:
    {
      strcpy (instname, "BEQ");
      iaformat = IA_branch;
      break;
    }
    case 0x3A:
    {
      strcpy (instname, "BLT");
      iaformat = IA_branch;
      break;
    }
    case 0x3B:
    {
      strcpy (instname, "BLE");
      iaformat = IA_branch;
      break;
    }
    case 0x3C:
    {
      strcpy (instname, "BLBS");
      iaformat = IA_branch;
      break;
    }
    case 0x3D:
    {
      strcpy (instname, "BNE");
      iaformat = IA_branch;
      break;
    }
    case 0x3E:
    {
      strcpy (instname, "BGE");
      iaformat = IA_branch;
      break;
    }
    case 0x3F:
    {
      strcpy (instname, "BGT");
      iaformat = IA_branch;
      break;
    }
    /*}}}*/
    default:
    /*{{{  show a data word*/
    {
    }
    /*}}}*/
  }
  switch (iaformat)
  {
    /*{{{  mem*/
    case IA_mem:
    {
      BYTE Ra, Rb;
      INT32 disp;
      disp = ilword & 0xFFFF;
      if (disp > 0x7FFF)
        disp -= 0x10000;
      Ra = (ilword >> 21) & 0x1F;
      Rb = (ilword >> 16) & 0x1F;
      sprintf (operand_string, "$%d,%ld($%d)", Ra, (long)disp, Rb);
      break;
    }
    /*}}}*/
    /*{{{  fmem*/
    case IA_fmem:
    {
      BYTE Fa, Rb;
      INT32 disp;
      disp = ilword & 0xFFFF;
      if (disp > 0x7FFF)
        disp -= 0x10000;
      Fa = (ilword >> 21) & 0x1F;
      Rb = (ilword >> 16) & 0x1F;
      sprintf (operand_string, "$f%d,%ld($%d)", Fa, (long)disp, Rb);
      break;
    }
    /*}}}*/
    /*{{{  memfunc*/
    case IA_memfunc:
    {
      int func = ilword & 0xFFFF;
      switch (func)
      {
        case 0x0000:
        {
          strcpy (instname, "TRAPB");
          break;
        }
        case 0x4000:
        {
          strcpy (instname, "MB");
          break;
        }
        case 0x8000:
        {
          strcpy (instname, "FETCH");
          break;
        }
        case 0xA000:
        {
          strcpy (instname, "FETCH_M");
          break;
        }
        case 0xC000:
        {
          strcpy (instname, "RPCC");
          break;
        }
        case 0xE000:
        {
          strcpy (instname, "RC");
          break;
        }
        case 0xF000:
        {
          strcpy (instname, "RS");
          break;
        }
      }
      break;
    }
    /*}}}*/
    /*{{{  membranch*/
    case IA_membranch:
    {
      INT32 jhint = ilword & 0x3FFF;
      BYTE Ra, Rb;
      Ra = (ilword >> 21) & 0x1F;
      Rb = (ilword >> 16) & 0x1F;
      if (jhint > 0x1FFF)
        jhint -= 0x4000;
      sprintf (operand_string, "$%d,($%d),%ld", Ra, Rb, (long)jhint);
      break;
    }
    /*}}}*/
    /*{{{  branch*/
    case IA_branch:
    {
      BYTE Ra = (ilword >> 21) & 0x1F;
      INT32 disp = ilword &0x1FFFFF;
      if (disp > 0xFFFFF)
        disp -= 0x200000;
      sprintf (operand_string, "$%d,.+%ld", Ra, (long)((disp+1)*4));
      break;
    }
    /*}}}*/
    /*{{{  operate*/
    case IA_operate:
    {
      BYTE Ra = (ilword >> 21) & 0x1F;
      BYTE Rc = ilword & 0x1F;
      if (((ilword >> 12) & 1) == 0)
      {
        BYTE Rb = (ilword >> 16) & 0x1F;
        sprintf (operand_string, "$%d,$%d,$%d", Ra, Rb, Rc);
      }
      else
      {
        int lit = (ilword >> 13) & 0xFF;
        sprintf (operand_string, "$%d,%d,$%d", Ra, lit, Rc);
      }
      break;
    }
    /*}}}*/
    /*{{{  fpop*/
    case IA_fpop:
    {
      BYTE Fa = (ilword >> 21) & 0x1F;
      BYTE Fb = (ilword >> 16) & 0x1F;
      BYTE Fc = ilword & 0x1F;
      if (fanotused)
        if (fbnotused)                   /* M?_FPCR FR,FR,FR */
        {
          assert (Fa == Fb);
          assert (Fc == Fb);
          sprintf (operand_string, "$f%d", Fc);   /* gas dislikes 3 regs */
          fbnotused = FALSE;
        }
        else
          sprintf (operand_string, "$f%d,$f%d", Fb, Fc);
      else
        sprintf (operand_string, "$f%d,$f%d,$f%d", Fa, Fb, Fc);
      break;
    }
    /*}}}*/
    /*{{{  pal*/
    case IA_pal:
    {
      break;
    }
    /*}}}*/
    default: break;
  }
  /*}}}*/
  *operands = sup_strlwr(operand_string);
  *instr = sup_strlwr(instr_string);
}
/*}}}*/
/*{{{  PUBLIC print_instruction_word (df, iword, start, address, special, asm_only)*/
/***************************************************************************
 *
 *  print_instruction disassembles the instruction iword
 *                    and writes it to stdout.
 *  If asm_only is TRUE ensures the text will be acceptable to assembler
 *                    special: 1=RET, 2=reljump, 0=other
 *
 **************************************************************************/
PUBLIC void print_instruction_word (FILE *df,
                               const BIT32 iword,
                               const INT32 start,
                               const INT32 address,
                               int *special, const BOOL asm_only)
{
  const char *ialpha_name, *operands;

  look_at_alpha_instruction (iword, &ialpha_name, &operands, address,
                       asm_only, special);
  if (asm_only)
    fprintf (df, " ");
  else
    fprintf(df, "%8lX %8ld: %08lX ", (long)(start + address), (long)address, (long)iword);
  fprintf (df, "%8s %s\n", ialpha_name, operands);
}
/*}}}*/
/*{{{  PUBLIC print_const_word(df, dword, start, address, asm_only)*/
/***************************************************************************
 *
 *  print_const_block disassembles a word of bytes
 *  and writes it to stdout
 *  If asm_only is TRUE ensures the text will be acceptable to assembler
 *
 **************************************************************************/
PUBLIC void print_const_word (FILE * df,
                   const BIT32 dword,
                   const INT32 start,
                   const INT32 address,
                   const BOOL asm_only)
{
  int i;
  char conststring [5];
  BIT32 dw = dword;
  char c1;

  if (asm_only)
    fprintf(df, " .byte ");
  else
    fprintf(df, "%8lX %8ld: ", (long)(start + address), (long)address);
  c1 = ' ';
  for (i = 0; i < 4; i++)
  {
    BYTE c;
    c = dw & 0xff;
    dw >>=8;
    if (asm_only)
    {
      fprintf (df, "%c0x%.2X", c1, c);
      c1 = ',';
    }
    else
      fprintf (df, "%c%.2X", c1, c);
    conststring [i] = (isprint (c) && (c != '\'') && (c != '\"')) ? (char)c : '.';
  }
  conststring [4] = '\0';
  fprintf (df, "    #%s\n", conststring);
}
/*}}}*/
/*{{{  PUBLIC disassemble_code_block (df, start, get_lword, code_offset, code_length, indent, asm_only)*/
PUBLIC void disassemble_code_block (FILE * df, const INT32 start,
                             BIT32 (* get_lword)(),
                             const int code_offset,
                             const int code_length, const int indent,
                             const BOOL asm_only)
 /*  If asm_only is TRUE ensures the text will be acceptable to assembler */
 /*  NOTE: the test for the end of a constant block is VERY dependent on
     details of what code the compiler generates at the start of a PROC, etc */
  {
    #define IA_BR      0x30
    BOOL in_const_table;
    int special_inst = 0;
    int i = 0;
    in_const_table = FALSE;
    while (i < code_length)
    {
      BIT32 iword = (*get_lword) ();
      if (in_const_table)
        if ((iword >> 16) == ((IA_STL << 10) | (REGRETURN << 5) | REGW) ||
            ((iword >> 21) == ((IA_BR << 5) | REGZERO)))
          /* stl ra,0(s4) or br zero,0 */
          in_const_table = FALSE;
      fprintf (df, "%*s", indent, "");
      if (in_const_table)
      {
        print_const_word (df, iword, start, code_offset + i, asm_only);
      }
      else
      {
        print_instruction_word (df, iword, start, code_offset + i,
                              &special_inst, asm_only);
        in_const_table = in_const_table || (special_inst == 1);
      }
      i += 4;
    }
  }
/*}}}*/
/*}}}*/
