/* $Id: tran1.c,v 1.2 1997/03/25 14:52:40 djb1 Exp $ */

/*
 *	tree transformer
 *	Copyright (C) 1990 Inmos Limited
 *
 *	This program is free software; you can redistribute it and/or modify
 *	it under the terms of the GNU General Public License as published by
 *	the Free Software Foundation; either version 2 of the License, or
 *	(at your option) any later version.
 *
 *	This program is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU General Public License for more details.
 *
 *	You should have received a copy of the GNU General Public License
 *	along with this program; if not, write to the Free Software
 *	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */


/*#define DEBUG*/

/*{{{  include files */
#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif
#include <stdio.h>
#include "midinc.h"
#include "occompfe.h"
#include "harndef.h"
#include "compharn.h" /* IMPORTED */ /* testflag */
#include "extlib.h" /* IMPORTED */

#include "trandef.h"
#include "tran1def.h"
#include "predefhd.h"

/*}}}  */

/*{{{  constants */
/* This is used to optimise VALOFs which expand into constants. */
#define OPTIMISE_VALOFS TRUE

/* any number will do, but a high one makes them look like temporaries */
#define ANON_VAR_SCOPE 0xFFF
/*}}}  */

/*{{{  PUBLIC  variables */
/* just public to the other 'trans' modules */
PUBLIC const trans_params_t *trans_params;
/*}}}  */
/*{{{  PRIVATE variables */
PRIVATE BOOL inline_expansions;
PRIVATE BOOL optimise_valof = TRUE;
PRIVATE int  overlap_checks;

typedef struct /* bug 1324 24/10/91 */
  {
    treenode **insertpoint;
    BOOL       insertvalof;
  } trans_data_t;
PRIVATE trans_data_t trans_data;

PRIVATE BOOL no_anonymous_vars = FALSE; /* bug 1455 5/11/91 */

PRIVATE int trans_lexlevel;
PRIVATE SOURCEPOSN trans_locn;

/* This flag is set while inside ASM, and inside the rhs of a RETYPE.
   If a channel is encountered while it is set, it is known that we can't
   perform fancy optimisations on that channel.
*/
PRIVATE BOOL trans_flag_scalar_chans;
/*}}}  */

/*{{{  forward declarations */
PRIVATE void trans        (treenode **tptr);
PRIVATE void transexp     (treenode **tptr, BOOL is_sub_expression);
PRIVATE void sub_transexp (treenode **tptr, BOOL is_sub_expression);
/*}}}  */

/*{{{  PRIVATE treenode *buildoverlapcheck (treenode *tptr) */
/*****************************************************************************
 *
 *  buildoverlapcheck takes an overlapcheck tree 'tptr' and returns an
 *                    expression tree which will perform the overlap check.
 *
 *****************************************************************************/
PRIVATE treenode *buildoverlapcheck ( treenode *tptr )
{
  treenode *lh_subscripts = LeftOpOf(tptr);
  treenode *rh_subscripts = RightOpOf(tptr);
  SOURCEPOSN locn = LocnOf(tptr);
  treenode *result = NULL;
  while (!EndOfList(lh_subscripts) && !EndOfList(rh_subscripts))
    /*{{{  add this check to result */
    {
      treenode *lhs = ThisItem(lh_subscripts);
      treenode *rhs = ThisItem(rh_subscripts);
      treenode *checktree;
      if (TagOf(lhs) == S_FOR || TagOf(rhs) == S_FOR)
	/*{{{  checktree =  full check */
	/*{{{  COMMENT what the full check tree looks like */
	/**********************  Start comment out ****************************
	The full check is

			    BITAND
			 /         \
		      GR             GR
		     /  \           /  \
		   ADD  base2     ADD  base1
		 /     \        /     \
	      base1  count1  base2  count2
	If this is TRUE, we have an overlap.
	 **********************   End comment out  ****************************/
	/*}}}  */
	{
	  treenode *base1, *count1, *base2, *count2;
	  treenode *c1, *c2;
	  /*{{{  set up base1, count1, base2, count2 */
	  if (TagOf(lhs) == S_FOR)
	    { base1 = LeftOpOf(lhs); count1 = RightOpOf(lhs); }
	  else
	    { base1 = lhs;           count1 = newconstant(1); }
	  if (TagOf(rhs) == S_FOR)
	    { base2 = LeftOpOf(rhs); count2 = RightOpOf(rhs); }
	  else
	    { base2 = rhs;           count2 = newconstant(1); }
	  /*}}}  */
	  c1 = newdopnode(S_ADD, locn, base1, count1,                    S_INT);
	  c1 = newdopnode(S_GR,  locn, c1,    copytree(base2, trans_lexlevel), S_INT);
	  c2 = newdopnode(S_ADD, locn, base2, count2,                    S_INT);
	  c2 = newdopnode(S_GR,  locn, c2,    copytree(base1, trans_lexlevel), S_INT);

	  /* bug 1117 29/1/91 - one half of these may be constant TRUE.
	     if so, we must constant fold it out:
	     (if it were constant FALSE, it would not have been generated
	      by the alias checker)
	     (similarly, they cannot both be constant TRUE)
	  */
	  c1 = foldexp(c1); c2 = foldexp(c2);
	  if      (isconst(c1)) /* must be constant TRUE */
	    checktree = c2;
	  else if (isconst(c2)) /* must be constant TRUE */
	    checktree = c1;
	  else
	    checktree = newdopnode(S_BITAND, locn, c1, c2, S_INT);
	}
	/*}}}  */
      /*{{{  else if both are constant */
      #if 1  /* Constant FALSE possibilities are now removed in the frontend */
	  /* No - the configurer can generate stuff which ends up looking
	     constant, so we allow for that here. bug TS/1495 31/01/92.
	  */
      else if (isconst(lhs) && isconst(rhs) && (LoValOf(lhs) != LoValOf(rhs)))
	/* don't need a check here */
	/* If they are different, we don't need a check at all */
	checktree = NULL;
      #endif
      /*}}}  */
      else
	/*{{{  checktree = simple check */
	checktree = newdopnode(S_EQ, locn, lhs, rhs, S_INT);
	/*}}}  */
      if (checktree != NULL)
	{
	/* 1/11/90 Added for bug 1032 to prevent an alias check of
	   a[i][0] and a[j][0] inserting an AND TRUE for the 0 against 0 check.
	   Theoretically we should also check that the result of folding the
	   checktree is TRUE, but we've already removed all FALSE checks
	   in the frontend.
	*/
	  checktree = foldexp(checktree);
	  if ((TagOf(checktree) == S_CONSTEXP) /*&& (LoValOf(checktree) != 0)*/)
	    checktree = NULL; /* remove this check */
	}
      if (checktree != NULL)
	/*{{{  AND checktree into result */
	{
	  if (result == NULL)
	    result = checktree;
	  else
	    result = newdopnode(S_BITAND, locn, result, checktree, S_INT);
	}
	/*}}}  */
      lh_subscripts = NextItem(lh_subscripts);
      rh_subscripts = NextItem(rh_subscripts);
    }
    /*}}}  */
  if (result == NULL)
    result = dummyexp_p;
  else
    result = newdopnode(S_CSUB0, locn, result, newconstant(1), S_INT);

  /* Now we don't need to bother folding the whole thing, because we've
     folded each check individually - 1/11/90 bug 1032 */
  /*result = foldexp(result);*/
  /* fprintf(outfile, "\nOverlap tree generated :\n");
     printtree(0, result); */
  return result;
}
/*}}}  */
/*{{{  PRIVATE void      overlapwarnings */
PRIVATE void overlapwarnings (const SOURCEPOSN locn)
{
  if ((overlap_checks > 0) && (warning_flags & WARNING_OVERLAPS))
    {
      if (overlap_checks > 1)
	msg_out_i(SEV_WARN, TRANS, TRANS_RUN_TIME_OVERLAP_CHECKS, locn, overlap_checks);
      else
	msg_out(SEV_WARN, TRANS, TRANS_RUN_TIME_OVERLAP_CHECK, locn);
    }
}
/*}}}  */
/*{{{  PRIVATE void      set_abbreviation_mode */
PRIVATE void set_abbreviation_mode(treenode *const t)
{
  /* bug 1156 - 15/2/91 
     We have to set the access mode of the variable */
  treenode *const nptr = DNameOf(t);
  switch(trans_params->abbrevmode_fn(t))
    {
      case AM_CONST:
	/* Though it isn't really, just give it a valid mode */
	SetNMode(nptr, NM_WORKSPACE);
	break;
      case AM_ISRHS:
	SetNMode(nptr, NModeOf(nameof(DValOf(t))));
	break;
      case AM_PTR:
	SetNMode(nptr, NM_POINTER);
	break;
      case AM_VAL: case AM_COPYINOUT:
	(void)isinvectorspace(nptr); /* does SetNMode */
	break;
      case AM_NOTHING:
	SetNMode(nptr, NM_DEFAULT);
	break;
    }
}
/*}}}  */
/*{{{  PUBLIC  treenode *trans_create_declaration */
PUBLIC treenode *trans_create_declaration(treenode ***const insertpoint,
       BOOL *const insertvalof,
       const int lexlevel, const char *name,
       const int decltag, const int nametag,
       treenode *const typetree,
       treenode *const expression)

/*{{{  comment */
/*
What this does is to create a new variable at some suitable scoping level,
so that it can be accessed by transformed code.
Thus we can insert an abbreviation of some complicated expression:

  x := y + complicated.expression
becomes:
  VAL anon IS complicated.expression :
  x := y + anon

This has uses when dealing with complicated bodies such as arrays whose sizes
aren't known at compile-time. Also used for constructors.

If the expression is 'inside' a short-circuiting expression,
we have to insert a VALOF around it:

  x := y OR (complicated.expression > 88)
becomes:
  x := y OR (VAL anon IS complicated.expression :
	     VALOF
	       SKIP
	       RESULT anon > 88
	    )

We have to be careful about where we insert these VALOFs, so that we don't
create a VALOF returning an array, unless the backend can handle them.
The example of AND/OR is OK, because their right-hand-sides are always
BOOL scalars.

The problem comes with counted arrays...

Conor O'Neill, 4/11/91
*/
/*}}}  */
{
  const SOURCEPOSN locn = LocnOf(**insertpoint);
  treenode *new_name;
  treenode *new_decl;
  char name_str[20];

  if (name == NULL)
    {
      static int anon_count = 0; /* yes - static to create unique names */
      sprintf(name_str, "$anon%d", anon_count++);
      name = &name_str[0];
    }

  /* we create this as NM_WORKSPACE so that issimplelocal works on the
     result; if it is placed into vectorspace then this is not so.
  */
  new_name = newnamenode(nametag, locn, lookupword(name, strlen(name)),
			 typetree, NULL, lexlevel, ANON_VAR_SCOPE,
			 /*NM_DEFAULT*/ NM_WORKSPACE);
  new_decl = newdeclnode(decltag, locn, new_name, expression,
			 **insertpoint);
  SetNDecl(new_name, new_decl);
  **insertpoint = new_decl;
  *insertpoint = DBodyAddr(new_decl);

  if (insertvalof != NULL && *insertvalof) /* we must insert a dummy VALOF too */
    {
      treenode *const valof = newvalofnode(S_VALOF, locn,
				     newleafnode(S_SKIP, locn),
				     newlistnode(S_LIST, locn, DBodyOf(new_decl), NULL));
      **insertpoint = valof;
      *insertvalof = FALSE;
    }
  return new_name;
}
/*}}}  */
/*{{{  PRIVATE treenode *create_declaration */
PRIVATE treenode *create_declaration(const int decltag, const int nametag,
				     treenode *const typetree,
				     treenode *const expression)
{
  return trans_create_declaration(&trans_data.insertpoint,
				  &trans_data.insertvalof,
				  trans_lexlevel, NULL,
				  decltag, nametag, typetree, expression);
}
/*}}}  */
/*{{{  PRIVATE treenode *fold_type_tree */
PRIVATE treenode *fold_type_tree(treenode *type)
{
  treenode *t;
  for (t = follow_user_type(type);
       TagOf(t) == S_ARRAY;
       t = follow_user_type(ARTypeOf(t)))
    SetARDimLength(t, foldexp(ARDimLengthOf(t)));
  return type;
}
/*}}}  */
/*{{{  PRIVATE treenode *copy_type_tree */
PRIVATE treenode *copy_type_tree(treenode *const type_tree, const int my_lexlevel)
{
#ifdef OCCAM2_5
  if (TagOf(type_tree) == S_RECORD)
    /*{{{  return the name of the record type */
    {
      if (ARTypeOf(type_tree) != NULL)
	return DNameOf(NDeclOf(DNameOf(ARTypeOf(type_tree))));
    }
    /*}}}  */
#endif
  return copytree(type_tree, my_lexlevel);
}
/*}}}  */
/*{{{  PRIVATE treenode *create_abbreviation */
PRIVATE treenode *create_abbreviation(treenode *const expression, const BOOL val)
{
  treenode *const exp_type = gettype(expression);
  treenode *const type = copy_type_tree(exp_type, trans_lexlevel); /* bug TS/1604 3/3/92 */
  treenode *const new_name =  create_declaration(val ? S_VALABBR : S_ABBR,
						val ? N_VALABBR : N_ABBR,
						 fold_type_tree(type), 
						 expression);
  set_abbreviation_mode(NDeclOf(new_name));
  return new_name;
}
/*}}}  */
/*{{{  PRIVATE treenode **create_local_abbreviation */
PRIVATE treenode **create_local_abbreviation(treenode **var, treenode **tptr)
{
  *var = create_abbreviation(*var, FALSE);
  while (isspecification(*tptr))
    tptr = DBodyAddr(*tptr);    /* skip over the new abbreviation */
  return tptr;
}
/*}}}  */
/*{{{  PRIVATE treenode *trans_getprotocol */
PRIVATE treenode *trans_getprotocol(treenode *type, BOOL *tagged_output)
{
  switch(TagOf(type))
    {
      case S_CHAN:
	type = ProtocolOf(type);
	if (TagOf(type) == N_SPROTDEF)
	  return NTypeOf(type);
	if (TagOf(type) == N_TPROTDEF)
	  {
	    const int old = switch_to_temp_workspace();
	    treenode *protocol = newleafnode(S_BYTE, NOPOSN);
	    switch_to_prev_workspace(old);

	    *tagged_output = TRUE;
	    return protocol;
	  }
	return type;
      case S_PORT:
      case S_TIMER:
	return NULL;
      default:
	badtag(NOPOSN, TagOf(type), "trans_getprotocol");
    }
  return type;
}
/*}}}  */
/*{{{  PRIVATE treenode *create_commlist */
PRIVATE treenode *create_commlist(const int tag, treenode *const chan,
  treenode *list, treenode *protocol, const BOOL tagged_output,
  const SOURCEPOSN locn)
{
  treenode *new = NULL;
  BOOL first = TRUE;
  for (; !EndOfList(list); list = NextItem(list))
    {
      treenode *newcomm;
      treenode *thisprot;
      if (tagged_output && first)
	{
	  thisprot = protocol;
	  protocol = NTypeOf(CExpOf(ThisItem(list)));
	  first = FALSE;
	}
      else if (TagOf(protocol) == S_LIST)
	{
	  thisprot = ThisItem(protocol);
	  protocol = NextItem(protocol);
	}
      else
	thisprot = protocol;

      newcomm = newactionnode(tag, locn, chan, addtofront(ThisItem(list), NULL));
      SetActionType(newcomm, copytree(thisprot, trans_lexlevel));

      new = addtofront(newcomm, new);
    }
  reverselist(&new);
  return new;
}
/*}}}  */
/*{{{  PRIVATE void      trans_constructor_assign */
PRIVATE void trans_constructor_assign(treenode **tptr)
/* we have
    element := [ e0, e1, e2, e3, ..., en]
   we turn this into
    SEQ
      element[0] := e0
      ...
      element[n] := en
*/
{
  treenode *ass = *tptr;
  const SOURCEPOSN locn = LocnOf(ass);
  treenode *lhs = LHSOf(ass);
  treenode *rhs = RHSOf(ass);
  treenode *list;
#ifdef OCCAM2_5
  treenode *field_list = NULL;
#endif
  int subscript;

  /*if (!issimplelocal(lhs, trans_lexlevel)) */
  if (!issimplelocal(lhs, trans_lexlevel) && !islocal(lhs, trans_lexlevel))
    tptr = create_local_abbreviation(&lhs, tptr);

#ifdef OCCAM2_5
  if (fe_typeof(rhs) == S_RECORD)
    {
      treenode *const record_type = follow_user_type(fe_gettype(rhs));
      field_list = ARTypeOf(record_type);
    }
#endif

  for (list = LitExpOf(rhs), subscript = 0;
       !EndOfList(list);
       list = NextItem(list), subscript++)
    {
      treenode *newlhs;
      treenode *subass;
    #ifdef OCCAM2_5
      if (field_list != NULL)
	/*{{{  record subscription */
	{
	  newlhs = newarraysubnode(S_RECORDSUB, locn, lhs, DNameOf(field_list));
	  field_list = DBodyOf(field_list);
	}
	/*}}}  */
      else
    #endif
	newlhs = newarraysubnode(S_ARRAYSUB, locn, lhs, newconstant(subscript));
      subass = newactionnode(S_ASS, locn, newlhs, ThisItem(list));

      /* any sub-constructors are dealt with by the call to trans which is
	 done on the whole structure.
      */

      NewItem(subass, list);
    }
  *tptr = newcnode(S_SEQ, locn, LitExpOf(rhs));

  freenode(&rhs);
  freenode(&ass);
}
/*}}}  */
/*{{{  PRIVATE BOOL      trans_bool_case */
PRIVATE BOOL trans_bool_case(treenode **const tptr)
/*{{{  comment about what this transformation is */
/* bug TS/1461 14/09/92 */
/*
  we turn
  CASE expression
    TRUE
      P1
    FALSE
      P2
  =>
  IF
    expression
      P1
    TRUE
      P2
and
  CASE expression
    FALSE
      P1
    TRUE
      P2
  =>
  IF
    expression
      P2
    TRUE
      P1
*/
/*}}}  */
{
  treenode *case_stmt = *tptr;
  treenode *lhs       = LHSOf(case_stmt);
  treenode *rhs       = RHSOf(case_stmt);
  treenode *first,       *second;
  treenode *first_items, *second_items;

  /*{{{  check that we have the correct 'pattern' */
  if (listitems(rhs) != 2)
    return FALSE;

  /* don't allow any leading specifications */ /* (cos I'm lazy) */
  first  = ThisItem(rhs);
  second = ThisItem(NextItem(rhs));
  if ((TagOf(first) != S_SELECTION) || (TagOf(second) != S_SELECTION))
    return FALSE;

  /* don't allow ELSE */
  first_items  = CondGuardOf(first);
  second_items = CondGuardOf(second);
  if ((TagOf(first_items) == S_ELSE) || (TagOf(second_items) == S_ELSE))
    return FALSE;

  /*}}}  */

  /* We now know that we have two cases; the frontend will have ensured
  that they are distinct; thus they must be TRUE and FALSE
  */

  /*printf("In correct bool case\n");*/

  /*{{{  overwrite it with an IF */
  {
    const SOURCEPOSN locn = LocnOf(case_stmt);
    INT32     first_val,    second_val;
    treenode *true_body, *false_body;
    treenode *first_guard, *second_guard, *if_stmt;

    first_val  = LoValOf(ThisItem(first_items));
    second_val = LoValOf(ThisItem(second_items));

    if (first_val != 0)
      {
	true_body  = CondBodyOf(first);
	false_body = CondBodyOf(second);
      }
    else
      {
	false_body = CondBodyOf(first);
	true_body  = CondBodyOf(second);
      }

    first_guard  = newcondnode(S_CHOICE, locn, lhs, true_body);
    second_guard = newcondnode(S_CHOICE, locn, newconstant(1), false_body);
    if_stmt = newcnode(S_IF, locn,
		       newlistnode(S_LIST, locn, first_guard,
		       newlistnode(S_LIST, locn, second_guard, NULL)));
    *tptr = if_stmt;
  }
  /*}}}  */

  /*{{{  clean up the CASE */
  /* clean up the mess */
  SetLHS(case_stmt, NULL);
  SetCondBody(first, NULL);
  SetCondBody(second, NULL);
  freetree(&case_stmt);
  /*}}}  */

  return TRUE;
}
/*}}}  */
/*{{{  PRIVATE void      create_initialised_constructor */
PRIVATE void create_initialised_constructor(treenode *const nptr)
{
  /* nptr is a VAL abbreviation, which we want to turn into an
     initialised variable, with an explicit process which
     does the initialisation.
  */
  treenode *const spec = NDeclOf(nptr);
  SetTag(spec, S_DECL);
  SetTag(nptr, N_DECL);
  SetDVal(spec, newactionnode(S_ASS, LocnOf(nptr), nptr, DValOf(spec)));
  trans_constructor_assign(DValAddr(spec));
}
/*}}}  */
/*{{{  PRIVATE BOOL      trans_process_sequential_io */
PRIVATE BOOL trans_process_sequential_io(treenode **tptr)
/* We turn sequential communications into a sequence of
   single communications
*/
/* returns TRUE if it has modified things substantially enough to need
   to recursively call trans on the result.
*/
{
  BOOL modified = FALSE;
  treenode *comm = *tptr;
  treenode *chan = LHSOf(comm);
  treenode *type = gettype(chan); /* will always be CHAN OF/PORT OF/TIMER */
  const int n = listitems(RHSOf(comm));
  BOOL tagged_output = FALSE;
  treenode *protocol = trans_getprotocol(type, &tagged_output);

  if ((n > 1) && !issimplelocal(chan, trans_lexlevel))
    /*{{{  abbreviate the channel expression */
    {
      tptr = create_local_abbreviation(&chan, tptr);
      modified = TRUE;
    }

    /*}}}  */

  if (n > 1) /* Can't be a PORT or TIMER */
    /*{{{  turn it into a sequence of communications */
    {
      const SOURCEPOSN locn = LocnOf(comm);
      treenode *rhs;
      treenode *new = NULL;

      /*{{{  turn it into a CHAN OF ANY */
      #if 0
      {
	treenode **t;
	/* worrying about the type being an array is unnecessary,
	   because the type of the channel must always be a scalar,
	   since we're just about to use it for a communication.
	*/
	t = ProtocolAddr(type);
	*t = newleafnode(S_ANY, locn);
      }
      #endif
      /*}}}  */

      new = create_commlist(TagOf(comm), chan, RHSOf(comm),
			    protocol, tagged_output, locn);
      *tptr = newcnode(S_SEQ, locn, new);
      modified = TRUE;

      rhs = RHSOf(comm);
      while (rhs != NULL)
	{
	  treenode *temp = NextItem(rhs);
	  freenode (&rhs);
	  rhs = temp;
	}
      freenode(&comm);
    }
    /*}}}  */
  else if (protocol != NULL) /* NULL means PORT or TIMER */
    /*{{{  single communication */
    {
      if (TagOf(protocol) == S_LIST)
	protocol = ThisItem(protocol);
      SetActionType(comm, copytree(protocol, trans_lexlevel));
    }
    /*}}}  */
  return modified;
}
/*}}}  */
/*{{{  PRIVATE void      trans_process_tagged_input */
PRIVATE void trans_process_tagged_input(treenode **tptr)
/*{{{  comment */
/* we turn:
  chan ? CASE tag ; exp1; exp2
into
  c IS chan :  -- optional
  BYTE b :
  SEQ
    c ? b
    IF
      b = tag
	SKIP
    c ? exp1
    c ? exp2
*/
/*}}}  */
{
  treenode *comm = *tptr;
  treenode *chan = LHSOf(comm);
  const SOURCEPOSN locn = LocnOf(comm);
  treenode *byte_var;

  /*{{{  channel and BYTE temp declaration */
  if ((listitems(RHSOf(comm)) > 1) && !issimplelocal(chan, trans_lexlevel))
    tptr = create_local_abbreviation(&chan, tptr);
  byte_var = create_declaration(S_DECL, N_DECL, newleafnode(S_BYTE, locn), NULL);
  while (isspecification(*tptr))
    tptr = DBodyAddr(*tptr);    /* skip over the new declaration */
  /*}}}  */

  {
    treenode *rhs = RHSOf(comm);
    treenode *const tag = ThisItem(rhs);
    treenode *const tagcomm  = newactionnode(S_INPUT, locn, chan, addtofront(byte_var, NULL));
    treenode *const commlist = create_commlist(S_INPUT, chan, NextItem(rhs),
					       NTypeOf(CExpOf(tag)), FALSE, locn);

    SetActionType(tagcomm, newleafnode(S_BYTE, locn));

    if (NEED_ERRORS &&
	(listitems(NTypeOf(ProtocolOf(basetype_tree(NTypeOf(nameof(chan)))))) > 1))
      {
      #if 0
	treenode *const eq       = newdopnode(S_EQ, locn, byte_var, tag, S_BYTE);
      #else
	treenode *const eq       = (LoValOf(tag) == 0)
		       ? newmopnode(S_NOT, locn, byte_var, S_BYTE)
		       : newdopnode(S_EQ, locn, byte_var, tag, S_BYTE);
      #endif
	treenode *const guard    = newcondnode(S_CHOICE, locn, eq, newleafnode(S_SKIP, locn));
	treenode *const if_test  = newcnode(S_IF, locn, addtofront(guard, NULL));
	*tptr = newcnode(S_SEQ, locn, addtofront(tagcomm, addtofront(if_test, commlist)));
      }
    else /* bug TS/2070 29/01/93 */
      *tptr = newcnode(S_SEQ, locn, addtofront(tagcomm, commlist));

    { /* now free up the old structure */
      treenode *temp = NextItem(rhs);
      while (!EndOfList(temp))
	{
	  treenode *temp2 = NextItem(temp);
	  freenode (&temp);
	  temp = temp2;
	}
      freenode (&rhs);
      freenode (&comm);
    }
  }
}
/*}}}  */
/*{{{  PRIVATE void      trans_process_case_input */
PRIVATE void trans_process_case_input(treenode **tptr)
/*{{{  comment */
/* we turn:
  chan ? CASE
    VAL zz IS 99 :
    tag0
      P
    tag1 ; x ; y
      Q
into
  c IS chan :  -- optional
  BYTE b :
  SEQ
    c ? b
    CASE b
      VAL zz IS 99 :
      tag0
	P
      tag1
	SEQ
	  c ? x
	  c ? y
	  Q
*/
/*}}}  */
{
  treenode *comm = *tptr;
  treenode *chan = LHSOf(comm);
  const SOURCEPOSN locn = LocnOf(comm);
  treenode *byte_var;

  /*{{{  channel and BYTE temp declaration */
  if (!issimplelocal(chan, trans_lexlevel))
    tptr = create_local_abbreviation(&chan, tptr);
  byte_var = create_declaration(S_DECL, N_DECL, newleafnode(S_BYTE, locn), NULL);
  while (isspecification(*tptr))
    tptr = DBodyAddr(*tptr);    /* skip over the new declaration */

  /*}}}  */
  /*{{{  create a CASE statement */
  {
    treenode *const tagcomm  = newactionnode(S_INPUT, locn, chan, addtofront(byte_var, NULL));
    SetActionType(tagcomm, newleafnode(S_BYTE, locn));

    if (nodetypeoftag(S_CASE_INPUT) != nodetypeoftag(S_CASE))
      abort();

    SetTag(comm, S_CASE);
    SetLHS(comm, byte_var);

    *tptr = newcnode(S_SEQ, locn, addtofront(tagcomm, addtofront(comm, NULL)));
  }
  /*}}}  */
  /*{{{  process each of the variants */
  {
    treenode *vlist;

    /* check that we can convert a S_VARIANT into an S_SELECTION safely */
    if (nodetypeoftag(S_VARIANT) != nodetypeoftag(S_SELECTION)) abort();

    for (vlist = RHSOf(comm); !EndOfList(vlist); vlist = NextItem(vlist))
      {
	treenode *const variant  = skipspecifications(ThisItem(vlist));
	treenode *const exp      = VRTaggedListOf(variant);
	const SOURCEPOSN comm_locn = LocnOf(variant); /* bug INSdi02088 05/04/93 */
	treenode *const commlist = create_commlist(S_INPUT, chan, NextItem(exp),
				   NTypeOf(CExpOf(ThisItem(exp))), FALSE, comm_locn);

	/*{{{  free up the old input list */
	{
	  treenode *t = NextItem(exp);
	  while (!EndOfList(t))
	    {
	      treenode *const temp = NextItem(t);
	      freenode (&t);
	      t = temp;
	    }
	}
	/*}}}  */

	NewNextItem(NULL, exp); /* now only one item in list */
	SetTag(variant, S_SELECTION);

	if (commlist != NULL)
	  SetCondBody(variant,
		      newcnode(S_SEQ, comm_locn,
			       appendlist(addtofront(CondBodyOf(variant), NULL),
				  commlist)));
      }
  }
  /*}}}  */

}
/*}}}  */
/*{{{  PRIVATE treenode *create_slice_io */
PRIVATE treenode *create_slice_io(const int tag, const SOURCEPOSN locn,
	treenode *const chan, treenode *const array, treenode *const length)
{
  treenode *array_slice;
  treenode *array_comm;

  array_slice = newsegmentnode(S_SEGMENT, locn, array, newconstant(0), length);
  if (tag == S_OUTPUT)
    array_slice = foldexp(array_slice);

  array_comm = newactionnode(tag, locn, chan, addtofront(array_slice, NULL));
  SetActionType(array_comm, copytree(gettype(array_slice), trans_lexlevel));
  return array_comm;
}
/*}}}  */
/*{{{  PRIVATE void      trans_process_counted_io */
PRIVATE void trans_process_counted_io(treenode **tptr)
/*{{{  comment */
/* we turn:
  chan ? len :: array
into
  c IS chan      :  -- optional
  type input.var :  -- optional
  INT slice.var  :  -- optional
  SEQ
    c ? input.var
    len := input.var           -- optional
    slice.var := INT input.var -- optional
    IF
      slice.var <> 0
	c ? [array FROM 0 FOR slice.var]
      TRUE
	SKIP
*/

/* we turn:
  chan ! len :: array
into
  c IS chan       :  -- optional
  type output.var :  -- optional
  INT slice.var   :  -- optional
  SEQ
    output.var := len           -- optional
    c ! output.var
    slice.var := INT output.var -- optional
    IF                          -- optional
      slice.var <> 0            -- optional
	c ! [array FROM 0 FOR slice.var]
      TRUE
	SKIP
*/

/*}}}  */
{
  treenode *comm = *tptr;
  treenode *chan = LHSOf(comm);
  const SOURCEPOSN locn = LocnOf(comm);
  treenode *const length_exp = LeftOpOf(ThisItem(RHSOf(comm)));
  treenode *array            = RightOpOf(ThisItem(RHSOf(comm)));
  const int length_type = ntypeof (length_exp);
  const BOOL nonsimple  = !issimplelocal(length_exp, trans_lexlevel)
		       || (nodetypeoftag(TagOf(length_exp)) != NAMENODE); /* bug INSdi01894 23/03/93 */
  const BOOL noninteger = (length_type != S_INT) && (length_type != targetintsize);
  treenode *p;

  if (!issimplelocal(chan, trans_lexlevel))
    tptr = create_local_abbreviation(&chan, tptr);

  if (TagOf(comm) == S_INPUT)
    /*{{{  counted input */
    {
      treenode *input_var;
      treenode *slice_var;
      treenode *length_comm;
      treenode *array_comm;
      treenode *guard;

      /*{{{  length temps */
      if (nonsimple)
	input_var = create_declaration(S_DECL, N_DECL, newleafnode(length_type, locn), NULL);
      else
	input_var = length_exp;

      if (noninteger)
	slice_var = create_declaration(S_DECL, N_DECL, newleafnode(S_INT, locn), NULL);
      else
	slice_var = input_var;

      if (RANGECHECKING && !issimple(array)) /* bug TS/1172 27/04/92 */
	array = create_abbreviation(array, FALSE);

      /*}}}  */

      length_comm = newactionnode(S_INPUT, locn, chan, addtofront(input_var, NULL));
      SetActionType(length_comm, newleafnode(length_type, locn));

      array_comm = create_slice_io(S_INPUT, locn, chan, array, slice_var);

      guard = newdopnode(S_NE, locn, slice_var, newconstant(0), S_INT);
      p = addtofront(newcnode(S_IF, locn,
	      addtofront(newcondnode(S_CHOICE, locn, guard, array_comm),
	      addtofront(newcondnode(S_CHOICE, locn, newconstant(1)/*newleafnode(S_TRUE, locn)*/,
						     newleafnode(S_SKIP, locn)),
			 NULL))),
		     NULL);

      if (noninteger)
	p = addtofront(newactionnode(S_ASS, locn, slice_var,
				     newmopnode(S_EXACT, locn, input_var, S_INT)),
		       p);

      if (nonsimple)
	p = addtofront(newactionnode(S_ASS, locn, length_exp, input_var), p);

      p = addtofront(length_comm, p);
    }
    /*}}}  */
  else
    /*{{{  counted output */
    {
      const BOOL constlen = isconst(length_exp);
      treenode *output_var;
      treenode *slice_var;
      treenode *length_comm;

      /*{{{  length temps */
      if (nonsimple && !constlen)
	output_var = create_declaration(S_DECL, N_DECL, newleafnode(length_type, locn), NULL);
      else
	output_var = length_exp;

      if (noninteger && !constlen)
	slice_var = create_declaration(S_DECL, N_DECL, newleafnode(S_INT, locn), NULL);
      else
	slice_var = output_var;

      if (RANGECHECKING && !isconst(array) && !issimple(array)) /* bug TS/1172 27/04/92 */
	array = create_abbreviation(array, TRUE);

      /*}}}  */

      length_comm = newactionnode(S_OUTPUT, locn, chan, addtofront(output_var, NULL));
      SetActionType(length_comm, newleafnode(length_type, locn));

      if (constlen                              && 
	  TagOf(length_exp) == S_CONSTEXP       &&
	  LoValOf(length_exp) == 0              &&
	  HiValOf(length_exp) == 0                     )
	p = NULL;
      else
	/*{{{  send a slice of the array */
	{
	  treenode *array_comm;
	  if (constlen)
	    {
	      slice_var = newmopnode(S_EXACT, locn, copytree(length_exp, trans_lexlevel), S_INT);
	      slice_var = foldexp(slice_var);
	    }

	  array_comm = create_slice_io(S_OUTPUT, locn, chan, array, slice_var);

	  if (constlen)
	    p = addtofront(array_comm, NULL);
	  else
	    /*{{{  put in a test for zero length */
	    {        
	      treenode *const guard = newdopnode(S_NE, locn, slice_var, newconstant(0), S_INT);
	      p = addtofront(newcnode(S_IF, locn,
		      addtofront(newcondnode(S_CHOICE, locn, guard, array_comm),
		      addtofront(newcondnode(S_CHOICE, locn, newconstant(1)/*newleafnode(S_TRUE, locn)*/,
							     newleafnode(S_SKIP, locn)),
				 NULL))),
			     NULL);

	      if (noninteger)
		p = addtofront(newactionnode(S_ASS, locn, slice_var,
					     newmopnode(S_EXACT, locn, output_var, S_INT)),
			       p);
	    }
	    /*}}}  */
	}
	/*}}}  */

      p = addtofront(length_comm, p);

      if (nonsimple && !constlen)
	p = addtofront(newactionnode(S_ASS, locn, output_var, length_exp), p);
    }
    /*}}}  */

  while (isspecification(*tptr))
    tptr = DBodyAddr(*tptr);    /* skip over the new declaration */
  *tptr = newcnode(S_SEQ, locn, p);

  freenode(ThisItemAddr(RHSOf(comm)));
  freenode(RHSAddr(comm));
  freetree(ActionTypeAddr(comm));
  freenode(&comm);

}
/*}}}  */
/*{{{  PRIVATE void      swap_names */
/*{{{  swap_names_t declaration */
typedef struct
  {
    treenode *old_name; /* name to be replaced */
    treenode *new_name; /* name to replace it with */
  } swap_names_t;
/*}}}  */
/*{{{  PRIVATE int do_swap_names */
PRIVATE int do_swap_names(treenode **tptr, void *const voidptr)
{
  const swap_names_t *const pair = voidptr;
  const int tag = TagOf(*tptr);

  if (*tptr == pair->old_name) /* ie must be the old namenode */
    *tptr = pair->new_name;
  else if (tag == S_PROCDEF || tag == S_LFUNCDEF || tag == S_SFUNCDEF)
    {
      /* bug TS/1780 24/07/92 */
      walk_free_vars(DNameOf(*tptr), do_swap_names, voidptr);
    }
  return CONTINUE_WALK;
}
/*}}}  */
/*{{{  PRIVATE void swap_names */
PRIVATE void swap_names(treenode **tptr, treenode *const old, treenode *const new)
{
  swap_names_t pair;
  pair.old_name = old;
  pair.new_name = new;
  modprewalktree(tptr, do_swap_names, &pair);
}
/*}}}  */
/*}}}  */
/*{{{  PRIVATE void      trans_process_replpar */
PRIVATE void trans_process_replpar(treenode **const tptr)
/*
   we transform

   PAR i = start FOR len
     ...  body (uses 'i')

into:

  PROC $anonrepl(VAL INT i)
    ...  body (uses new 'i')
  :
  PAR i = start FOR len
    $anonrepl(i)

Note that because lexlevels are incremented inside repl pars, we don't
need to adjust lex levels when doing this modification.
*/
{
  treenode *repl = *tptr;
  const SOURCEPOSN locn = LocnOf(repl);
  treenode *nptr;
  treenode *proc;
  treenode *formal;
  treenode *instance;

  /*{{{  create a proc */
  {
    treenode *const body = ReplCBodyOf(repl);
    char name_str[20];
    static int repl_anon_count = 0; /* yes - static to create unique names */

    sprintf(name_str, "_anonrepl%d", repl_anon_count++);

    nptr = newnamenode(N_PROCDEF, locn, lookupword(name_str, strlen(name_str)),
		       NULL, NULL, trans_lexlevel, ANON_VAR_SCOPE,
		       NM_DEFAULT);
    proc = newdeclnode(S_PROCDEF, locn, nptr, body, NULL);
    SetNDecl(nptr, proc);
    formal = newnamenode(N_VALPARAM, locn, NNameOf(ReplCNameOf(repl)),
			 newleafnode(S_INT, locn),
			 proc, trans_lexlevel+1, ANON_VAR_SCOPE,
			 NM_WORKSPACE);
    SetNType(nptr, newlistnode(S_LIST, locn, formal, NULL));
  }
  /*}}}  */
  /*{{{  modify the body to use the new formal */
  {
    swap_names(&proc, ReplCNameOf(repl), formal);

  #if 0
    printf("trans_process_replpar: tree is");
    printtree(stdout, 0, proc);
    printf("\n");
  #endif

    mark_free_vars(proc, FALSE);
  }
  /*}}}  */
  /*{{{  call trans on the result */
  trans (&proc);
  /*}}}  */
  /*{{{  create an instance */
  instance = newinstancenode(S_PINSTANCE, locn, nptr,
			     newlistnode(S_LIST, locn, ReplCNameOf(repl), NULL));
  SetNUsed(nptr, TRUE);
  trans_lexlevel++;
  trans(&instance);
  trans_lexlevel--;
  /*}}}  */
  /*{{{  bolt everything together */
  SetReplCBody(repl, instance);
  SetDBody(proc, repl);
  *tptr = proc;
  /*}}}  */
}
/*}}}  */
/*{{{  PRIVATE void      trans_process_repl_for_one */
PRIVATE void trans_process_repl_for_one(treenode **tptr)
{
  treenode *repl = *tptr;
  const SOURCEPOSN locn = LocnOf(repl);
  treenode *const repl_start = ReplCStartExpOf(repl);
  treenode *const repl_nptr  = ReplCNameOf(repl);
  treenode *repl_new;

  *tptr = ReplCBodyOf(repl); /* skip past the replicator */

  /*printf("trans_process_repl_for_one: #%8X (%d)\n", locn, locn);*/
  if (isconst(repl_start))
    {
      repl_new = create_abbreviation(repl_start, TRUE);
    }
  else
    {
      treenode *ass;
      repl_new = create_declaration(S_DECL, N_DECL, newleafnode(S_INT, locn), NULL);
      ass = newactionnode(S_ASS, locn, repl_new, repl_start);
      SetDVal(NDeclOf(repl_new), ass); /* initialised variable */
    }
  tptr = DBodyAddr(*tptr);    /* skip over the new declaration */

  swap_names(tptr, repl_nptr, repl_new);

  if (isconst(repl_start))
    foldtree(*tptr); /* fold the constants */

  freetree(ReplCLengthExpAddr(repl));
  freenode(&repl);
}
/*}}}  */
/*{{{  PUBLIC  void      transsubscripts */
/*****************************************************************************
 *
 *  transsubscripts performs trans on segments and subscripts.
 *                  For segments, temporaries are inserted where needed on
 *                  the start and length expressions.
 *                  No tree transformation is done, this is performed by
 *                  trans on the top-level subscript only.
 *
 * if create_temps is TRUE, it will insert temporaries for the start and
 * length if required.
 *
 *****************************************************************************/
PUBLIC void transsubscripts (const trans_params_t *const trans_params_in,
			     treenode **tptr, const BOOL create_temps)
{
  const trans_params_t *const saved_params = trans_params;
  BOOL going = TRUE;
  trans_params = trans_params_in;
  while (going)
    {
      treenode *const t = *tptr;
      switch(nodetypeoftag(TagOf(t)))
	{
	  /*{{{  SEGMENT SEGMENTITEM */
	  case SEGMENTNODE:
	    {
	      BOOL saved_optimise_valof;
	      /*DEBUG_MSG(("transsubscripts: SEGMENTNODE. Now calling trans on SStartExp\n"));*/
	      sub_transexp(SStartExpAddr(t), TRUE);
	      /*DEBUG_MSG(("transsubscripts: SEGMENTNODE. Now calling trans on SLengthExp\n"));*/
	      saved_optimise_valof = optimise_valof;
	      optimise_valof = FALSE; /* Fix for bug 634 */
	    #ifdef OCCAM2_5
	      if (SLengthExpOf(t) == NULL)
		/*{{{  insert the 'FOR' expression */
		{
		/*if (!issimplelocal(SNameOf(t)))
		    SetSName(t, create_abbreviation(SNameOf(t), TRUE));
		*/
		  if (!issimplelocal(SStartExpOf(t), trans_lexlevel) &&
		      !isconst(SStartExpOf(t)))
		    SetSStartExp(t, create_abbreviation(SStartExpOf(t), TRUE));
		  SetSLengthExp(t, newdopnode(S_SUBTRACT, LocnOf(t),
					      newmopnode(S_ELSIZE, LocnOf(t), SNameOf(t), S_INT),
					      SStartExpOf(t), S_INT));
		}
		/*}}}  */
	    #endif
	      sub_transexp(SLengthExpAddr(t), TRUE);
	      optimise_valof = saved_optimise_valof;
	      if (TagOf(t) == S_SEGMENTITEM)  /* Added by CO'N 6/4/90 */
		{
		  /*DEBUG_MSG(("transsubscripts: SEGMENTITEM. Now calling trans on SSubscriptExp and SCheckExp\n"));*/
		  sub_transexp(SSubscriptExpAddr(t), TRUE);
		  sub_transexp(SCheckExpAddr(t),     TRUE);
		}
	      /*DEBUG_MSG(("transsubscripts: SEGMENTNODE. Recursing on SNameOf\n"));*/
	      transsubscripts(trans_params_in, SNameAddr(t), create_temps);
	      if (create_temps && !no_anonymous_vars) /* bug TS/1455 5/11/91 */
		/*{{{  put temporaries on start and length expressions if necessary */
		{
		  treenode *lengthexp = SLengthExpOf(t);
		  treenode *startexp  = SStartExpOf(t);
		  treenode *s = skipevals(startexp);
		  if (!isconst(s) && !issimplelocal(s, trans_lexlevel))
		    /* we would evaluate it twice - once in the range check and once in the
		       base, otherwise */
		    {
		      /*DEBUG_MSG(("transsubscripts: creating temp for startexp\n"));*/
		      SetSStartExp(t, create_abbreviation(startexp, TRUE)); /* bug TS/1324 24/10/91 */
		    }
		  /* we nearly always convert to a temporary here, incase the
		     value is required as a range check later */
		  if (!isconst(lengthexp) && !issimplelocal(lengthexp, trans_lexlevel))
		    {
		      /*DEBUG_MSG(("transsubscripts: creating temp for lengthexp\n"));*/
		      SetSLengthExp(t, create_abbreviation(lengthexp, TRUE)); /* bug TS/1324 24/10/91 */
		    }
		}
		/*}}}  */
	    }
	    going = FALSE;
	    break;
	  /*}}}  */
	  /*{{{  ARRAYSUB */
	  case ARRAYSUBNODE:
	    /*DEBUG_MSG(("transsubscripts: ARRAYSUBNODE. recursing on ASExpOf\n"));*/
	    sub_transexp(ASExpAddr(t), TRUE);
	    /*DEBUG_MSG(("transsubscripts: ARRAYSUBNODE. recursing on ASIndexOf\n"));*/
	    sub_transexp(ASIndexAddr(t), TRUE);
	    /*DEBUG_MSG(("transsubscripts: ARRAYSUBNODE. looping down ASBaseOf\n"));*/
	    tptr = ASBaseAddr(t);
	    break;
	  /*}}}  */
	  /*{{{  CONSTRUCTOR */
	  case LITNODE:
	    if (TagOf(t) == S_CONSTRUCTOR)
	      sub_transexp(tptr, TRUE);
	    going = FALSE;
	    break;
	  /*}}}  */
	  /*{{{  FINSTANCE */
	  #ifdef OCCAM2_5
	  case INSTANCENODE:
	    DEBUG_MSG(("transsubscripts: INSTANCENODE. calling sub_transexp on fn\n"));
	    sub_transexp(tptr, TRUE);
	    DEBUG_MSG(("transsubscripts: INSTANCENODE. creating abbr\n"));
	    *tptr = create_abbreviation(*tptr, TRUE);
	    going = FALSE;
	    break;
	  #endif
	  /*}}}  */
	  default:
	    going = FALSE;
	    break;
	}
    }
  trans_params = saved_params;
}
/*}}}  */
/*{{{  PRIVATE void      trans_resultlist */
PRIVATE void trans_resultlist(treenode **const valof)
{
  treenode *list;
  const trans_data_t saved_skip_data = trans_data;
  const BOOL skip_valof = TagOf(VLBodyOf(*valof)) == S_SKIP;

  if (skip_valof)
    /*{{{  arrange for insertions infront of the valof */
    {
      trans_data.insertpoint = valof;
      trans_data.insertvalof = FALSE;
      /*DEBUG_MSG(("trans_resultlist: setting trans_data at %lX\n", trans_locn));*/
    }
    /*}}}  */

  for (list = VLResultListOf(*valof); !EndOfList(list); list = NextItem(list))
    {
      const trans_data_t saved_data = trans_data;

      if (!skip_valof)
	/*{{{  arrange for insertions inside the result list */
	{
	  trans_data.insertpoint = ThisItemAddr(list);
	  trans_data.insertvalof = TRUE; /* new valof will have same type as this one,
				   so we're not inserting array valofs unless
				   they're already permitted.
				   CON 4/11/91.
				*/
	  /*DEBUG_MSG(("trans_resultlist: setting trans_data at %lX\n", trans_locn));*/
	}
	/*}}}  */

      sub_transexp(ThisItemAddr(list), TRUE);

      if (!skip_valof)
	/*{{{  reset insertions */
	{
	  trans_data = saved_data;
	  /*DEBUG_MSG(("trans_resultlist: resetting trans_data\n"));*/
	}
	/*}}}  */
    }

  if (skip_valof)
    /*{{{  reset insertions */
    {
      trans_data = saved_skip_data;
      /*DEBUG_MSG(("trans_resultlist: resetting trans_data\n"));*/
    }
    /*}}}  */
}
/*}}}  */
/*{{{  PRIVATE treenode **transspecs */
PRIVATE treenode ** transspecs(treenode **tptr)
{
  const trans_data_t saved_data = trans_data;
  trans_data.insertvalof = FALSE;
  while (isspecification(*tptr))
    {
      treenode *const t = *tptr;
      trans_locn = LocnOf(t);
      trans_data.insertpoint = tptr;
    /*DEBUG_MSG(("transspecs: setting trans_data at %lX on %s\n", trans_locn, itagstring(TagOf(t))));*/
      switch (TagOf(t))
	{
	  /*{{{  default - ignore */
	  default:
	  #if 0
	  case S_TPROTDEF: case S_SPROTDEF:
	  case S_WSPLACE: case S_VSPLACE:
	  case S_LABELDEF:
	  case S_PRAGMA: /* bug 829 19/9/91 */
	  case S_TYPEDECL:
	  #endif
	    break;
	  /*}}}  */
	  /*{{{  routine specification */
	  case S_PROCDEF: case S_SFUNCDEF: case S_LFUNCDEF:
	    if (!isinline(DNameOf(t)))  /* Added 29/5/90 by CO'N for bug 315 */
	      {
		treenode *n = DNameOf(t);
		trans_lexlevel++;  /* Formal parameters are at a higher lexical level */
		augmentformals(n, trans_lexlevel);
		if (!separatelycompiled(n))
		  trans(DValAddr(t));
		trans_lexlevel--;
	      }
	    break;

	  /*}}}  */
	  /*{{{  DECL */
	  case S_DECL: /* we have to set the access mode */
	    {
	      /* moved from the mapper 15/2/91 for bug 1156 */
	      treenode *nptr = DNameOf(t);
	      if (TagOf(nptr) == S_LIST)
		for (; !EndOfList(nptr); nptr = NextItem(nptr))
		  {
		    isinvectorspace(ThisItem(nptr)); /* does SetNMode */
		    if (TagOf(NTypeOf(ThisItem(nptr))) == S_CHAN)
		      SetNChanMark(ThisItem(nptr), FALSE);
		  }
	      else
		  {
		    isinvectorspace(nptr); /* does SetNMode */
		    if (TagOf(NTypeOf(nptr)) == S_CHAN)
		      SetNChanMark(nptr, FALSE);
		  }
	      trans(DValAddr(t));
	    }
	    break;

	  /*}}}  */
	  /*{{{  Abbreviations / RETYPEs */
	  case S_VALABBR: case S_VALRETYPE: case S_ABBR: case S_RETYPE:
	    {
	      treenode **rhsaddr = DValAddr(t);
	      treenode *const nptr = DNameOf(t);
	      const int tag = TagOf(nptr);
	      const BOOL retype = (tag == N_VALRETYPE) || (tag == N_RETYPE);
	      SOURCEPOSN thislocn = trans_locn;
	      const BOOL wasconstant = isconst(*rhsaddr); /* bug 1040 13/11/90 */
	      const BOOL saved_flag_scalar_chans = trans_flag_scalar_chans;
	      trans_flag_scalar_chans = retype;
	      if (TagOf(NTypeOf(nptr)) == S_CHAN)
		SetNChanMark(nptr, FALSE);

	      if (!retype && TagOf(*rhsaddr) == S_CONSTRUCTOR &&
		  trans_params->abbr_constructors)
		{
		  create_initialised_constructor(nptr); /* bug TS/1263 11/05/92 */
		  (void)isinvectorspace(nptr);
		  trans(DValAddr(NDeclOf(nptr)));
		  break; /* break out of master switch statement */
		}

	      /* if we're retyping, we consider the rhs to be a 'sub expression'
		 so that constructors, stc, get separated out.
		 bug TS/1263 11/05/92
	      */
	      overlap_checks = 0;
	      transexp(rhsaddr, retype);
	      overlapwarnings(thislocn);

	      (void)isinvectorspace(nptr);
	      set_abbreviation_mode(t);

	      /*{{{  fill in the unknown array dimensions */
	      {
		treenode *type = NTypeOf(nptr);
		if (!retype)
		  /*{{{  copy dimensions from the type of the right-hand side */
		  {
		    int dimension = 0;
		    for(; TagOf(type) == S_ARRAY; dimension++ , type = ARTypeOf(type))
		      if (ARDimOf(type) == (-1))
		    #if 0
			SetARDimLength(type, dimexpof(*rhsaddr, dimension));
		    #else
			/* bug TS/1404 & TS/0149 30/09/91 28/10/91 */
			{
			  treenode *newdim = dimexpof(*rhsaddr, dimension);
			  if (!isconst(newdim) && !issimplelocal(newdim, trans_lexlevel))
			    {
			      DEBUG_MSG(("transspecs: abbrev dim of '%s' is complicated\n", WNameOf(NNameOf(nptr))));
			    /*newdim = newtempnode(T_TEMP, newdim, NM_WORKSPACE);*/
			      newdim = create_abbreviation(newdim, TRUE); /* bug TS/1404 28/10/91 */
			    }
			  SetARDimLength(type, newdim);
			  if (isconst(newdim))
			    SetARDim(type, LoValOf(newdim));
			}
		    #endif
		  }
		  /*}}}  */
		else
		  /*{{{  there is at most one hidden dimension in a retype */
		  {
		  #if 0 /* bug 1505 28/11/91 now superseeded */
		    /*{{{  insert a temp for the hidden dimension */
		    for (; TagOf(type) == S_ARRAY; type = ARTypeOf(type))
		      if (ARDimOf(type) == (-1))
			/* Make up a temporary for the open dimension */
			{
			  treenode *temp = newtempnode(T_TEMP/*T_PREEVALTEMP*/, dummyexp_p, NM_WORKSPACE); /* Bug TS/1505 28/11/91 */
			  DEBUG_MSG(("transspecs: creating temp dimension at trans_lexlevel %d\n", trans_lexlevel));
			  SetARDimLength(type, temp);
			}
		    /*}}}  */
		  #else
		    /* insert anonymous variables for hidden dimension */
		    const INT32 scalarsize = bytesin(basetype_tree(type));
		    INT32 lsize = scalarsize;
		    treenode **opendimension = NULL;
		  #ifdef OCCAM2_5
		    int vardimensions = 0;
		  #endif
		    treenode *const rhstype = gettype(*rhsaddr);
		    const INT32 rsize = known_bytesin(rhstype);
		    const int old = switch_to_temp_workspace();
		    treenode *dim_exp = unknowndimsof(rhstype);
		    switch_to_prev_workspace(old);

		    /*{{{  look for an open dimension, calculate product of known dimensions */
		    {
		      treenode *tt;
		      for (tt = type; TagOf(tt) == S_ARRAY; tt = ARTypeOf(tt))
			if (ARDimLengthOf(tt) == NULL)
			  opendimension = ARDimLengthAddr(tt);
		    #ifdef OCCAM2_5
			else if (ARDimOf(tt) == (-1))
			  vardimensions++;
		    #endif
			else
			  lsize *= ARDimOf(tt);
		    }
		    /*}}}  */

		  #ifdef OCCAM2_5
		    if (NVReshapesOf(nptr) && (vardimensions != 0))
		      /*{{{  process 'RESHAPES' with var dims */
		      {
			/*{{{  COMMENT */
			/* We have a RESHAPES such as:
			   [i+j][k+l]INT RESHAPES rhs :
			   or
			   [][k+l]INT RESHAPES rhs :

			  * we range check variable dimensions, to ensure non-negative
			  * we abbreviate all `complicated' variable dimensions,
			    to turn them all into local variables.
			  * we calculate known size of lhs type, in units of the type of the array
			  * we calculate known size of rhs type, in units of the type of the array
			  * if necessary, we calculate the open dimension by division.
					  and adjust the lhs to account for this
			  * we create an abbreviation to check that the two sizes
			    are equal.

			  IMPORTANT NOTE - I have not attempted to tweak all sorts
			  of optimisations out of this. I consider correctness
			  to be far more important. At least, when the algorithm is
			  this simple, there is a good chance that the result
			  is correct!

			  CON - 10/1/94.
			*/

			/*}}}  */
			/*{{{  variables */
			const INT32 known_lhsize = known_bytesin(type)    / bytesin(basetype_tree(type));
			const INT32 known_rhsize = known_bytesin(rhstype) / bytesin(basetype_tree(rhstype));
			treenode *unknown_lhsize = NULL;
			treenode *lhsize_exp;
			treenode *rhsize_exp;
			treenode *a;

			/*}}}  */
			/*{{{  abbreviate any unknown dims if necessary */
			{
			  treenode *tt;
			  for (tt = type; TagOf(tt) == S_ARRAY; tt = ARTypeOf(tt))
			    if ((ARDimLengthOf(tt) != NULL) &&
				(ARDimOf(tt) == (-1)))
			      /*{{{  variable dimension */
			      {
				transexp(ARDimLengthAddr(tt), retype);
				a = ARDimLengthOf(tt);
				if (RANGECHECKING)
				  a = newdopnode(S_CSUB0, trans_locn, a,
				      newconstant((bytesperword == 2) ? MOSTNEG_INT16 : MOSTNEG_INT32),
				       S_INT);
				if (!issimplelocal(a, trans_lexlevel))
				  a = create_abbreviation(a, TRUE);
				SetARDimLength(tt, a);
				unknown_lhsize = add_to(unknown_lhsize, S_TIMES, a);
			      }
			      /*}}}  */
			}
			/*}}}  */
			/*{{{  calculate lhsize_exp */
			if (known_lhsize == 1)
			  lhsize_exp = unknown_lhsize;
			else
			  lhsize_exp = add_to(unknown_lhsize, S_TIMES, newconstant(known_lhsize));

			if ((opendimension != NULL) && (RANGECHECKING) &&
			    !isconst(lhsize_exp) && !issimplelocal(lhsize_exp, trans_lexlevel))
			  lhsize_exp = create_abbreviation(lhsize_exp, TRUE);
			/*}}}  */
			/*{{{  calculate rhsize_exp */
			dim_exp = copytree(dim_exp, trans_lexlevel);
			if (known_rhsize == 1)
			  rhsize_exp = dim_exp;
			else
			  rhsize_exp = add_to(dim_exp, S_TIMES, newconstant(known_rhsize));

			if ((opendimension != NULL) && (RANGECHECKING) &&
			    !isconst(rhsize_exp) && !issimplelocal(rhsize_exp, trans_lexlevel))
			  rhsize_exp = create_abbreviation(rhsize_exp, TRUE);
			/*}}}  */
			/*{{{  calculate open dimension */
			if (opendimension != NULL)
			  {
			    a = newdopnode(S_DIV, trans_locn,
					   copytree(rhsize_exp, trans_lexlevel),
					   copytree(lhsize_exp, trans_lexlevel),
					   S_INT);
			    *opendimension = create_abbreviation(a, TRUE);
			    lhsize_exp = newdopnode(S_TIMES, trans_locn, *opendimension, lhsize_exp, S_INT);
			  }
			/*}}}  */
			/*{{{  check total size */
			if (RANGECHECKING)
			  {
			    a = newdopnode(S_EQ, trans_locn, lhsize_exp, rhsize_exp, S_INT);
			    a = newdopnode(S_CCNT1, trans_locn, a, newconstant(1), S_INT);
			    (void)create_abbreviation(a, TRUE);
			  }
			/*}}}  */
		      }
		      /*}}}  */
		    else
		  #endif
		    if ((opendimension != NULL) && (dim_exp == NULL))
		      /*{{{  expansion of INLINE routine, dimensions are now known */
		      {
			/* bug INSdi01929 */
			/* This is an expansion similar to:
			  [100]BYTE b :
			  []INT16 x16 RETYPES b:
			  ...
			  This has probably been created by expanding out an INLINE routine,
			  otherwise the front-end would have inserted the unknown type.
			*/

			if (rsize % lsize == 0)
			  {
			    const INT32 new_dim = rsize / lsize;
			    treenode *tt;
			    for (tt = type; TagOf(tt) == S_ARRAY; tt = ARTypeOf(tt))
			      if (ARDimOf(tt) == (-1))
				{
				  SetARDimLength(tt, newconstant(new_dim));
				  SetARDim(tt, new_dim);
				}
			  }
			else
			  msg_out_s(SEV_FATAL, TRANS, TRANS_RETYPE_TYPE_MISMATCH, thislocn,
				    WNameOf(NNameOf(nptr)));
		      }
		      /*}}}  */
		    else if (opendimension != NULL)
		      /*{{{  handle the open dimension */
		      {
			dim_exp = copytree(dim_exp, trans_lexlevel); /* copy into 'real' space */
			if (rsize % lsize == 0)
			  /*{{{  multiply dim_exp by rsize / lsize */
			  {
			    if (rsize != lsize)
			      dim_exp = newdopnode(S_TIMES, thislocn, dim_exp,
						   newconstant(rsize / lsize), S_INT);
			  }
			  /*}}}  */
			else
			  /*{{{  check and calculate open dimension */
			  {
			    /* lsize != 1, or we wouldn't have come into this branch
			       of the if */

			    /* if lsize is a power of 2, this sets lsize_bit to the bit number */
			    const int lsize_bit = whichpowerof2(lsize);

			    if (rsize != 1)
			      dim_exp = newdopnode(S_TIMES, thislocn, dim_exp,
						   newconstant(rsize), S_INT);
			    if (RANGECHECKING)
			      /*{{{  check rhs size and lhs size match */
			      {
				BOOL copy = FALSE;
				treenode *temp1;
				treenode *temp2;
				if (TagOf(dim_exp) == S_ELSIZE)
				  {
				    const int this_old = switch_to_temp_workspace();
				    treenode *const exp = dimexpof(OpOf(dim_exp), 0);
				    switch_to_prev_workspace(this_old);
				    copy = issimplelocal(exp, trans_lexlevel);
				  }

				if (issimplelocal(dim_exp, trans_lexlevel) || copy)
				  temp1 = dim_exp;
				else
				  temp1 = create_abbreviation(dim_exp, TRUE);

				temp2 = copy ? copytree(temp1, trans_lexlevel) : temp1;

				/* bug TS/1882 16/10/92 */
				/* This is definitely safe, because the expression of which we want the
				   remainder (temp1) is a size of an array, and therefore will always be
				   positive.
				*/
				if (lsize_bit >= 0)
				  dim_exp = newdopnode(S_BITAND, thislocn, temp1, newconstant(lsize - 1), S_INT);
				else
				  dim_exp = newdopnode(S_REM, thislocn, temp1, newconstant(lsize), S_INT);

				dim_exp = newdopnode(S_CSUB0, thislocn, dim_exp, newconstant(1), S_INT);
				dim_exp = newdopnode(S_EVAL,  thislocn, dim_exp, temp2, S_INT);
			      }
			      /*}}}  */

			    /* bug TS/1882 16/10/92 */
			    /* This is definitely safe, because the expression of which we want the
			       remainder (dim_exp) is either the size of an array,
			       or an eval node of temp2 which is a copy of temp1 which is a copy
			       of dim_exp, and therefore will always be positive.
			    */
			    if (lsize_bit >= 0)
			      dim_exp = newdopnode(S_RSHIFT, thislocn, dim_exp, newconstant(lsize_bit), S_INT);
			    else
			      dim_exp = newdopnode(S_DIV, thislocn, dim_exp, newconstant(lsize), S_INT);
			  }
			  /*}}}  */

			/*{{{  check if the open dimension could overflow */
			/*{{{  Comment about this */
			/*
			Comment about the rationale behind the check.

			Note that, according to bug TS/1359, we have arranged that:
			1) On a 32-bit processor, no array may be larger than 1/2 memory size.
			   Thus the size of any array can be expressed as a positive number of
			   bytes, with no need to consider unsignedness.
			2) on a 16-bit processor, we can have an array whose total size is
			   greater than half the memory size. Thus the total size in bytes
			   must be considered as an unsigned number, but in any case
			   will be less than 2^16.
			3) Each dimension in a multi dimensioned array must have a positive size.
			4) Therefore the only occasion where we need to consider inserting a
			   check that the calculated dimension is too large is:
			   a) on a 16-bit processor
			   b) where we're RETYPEing into a byte/bool array

			CON, 19/10/92.
			*/
			/*}}}  */

			/* bug TS/1359 19/10/92 */
			if (RANGECHECKING && (bytesperword == 2) && (scalarsize * lsize == 1))
			  {
			    dim_exp = newdopnode(S_CSUB0, thislocn, dim_exp,
						 newconstant(0xFFFF8000), S_INT);
			  }
			/*}}}  */

			/*{{{  create abbreviation if necessary */
			if (issimplelocal(dim_exp, trans_lexlevel))
			  {
			    /* can happen eg:
			       PROC p ([]INT32 i32)
				 []INT retype RETYPES i32 :
				 etc.
			    */
			    *opendimension = dim_exp;
			  }
			else
			  *opendimension = create_abbreviation(dim_exp, TRUE);
			/*}}}  */
		      }
		      /*}}}  */
		    else if (dim_exp != NULL && (RANGECHECKING))
		      /*{{{  check that the sizes match */
		      {
			/* bug TS/1980 21/01/93 - we must check that the size of the rhs
			   matches that specified in the lhs.
			*/
			/* we have an example such as:

			   PROC p ([][]INT array)
			     [100][2]BYTE b RETYPES array :

			   therefore we want to check that
			   lsize == rsize * (dim_exp)
			   or, more to the point, dim_exp == (lsize / rsize)
			   We create an anonymous variable with this expression
			   which ensures that the check is generated.
			*/
			treenode *check_exp;
			dim_exp = copytree(dim_exp, trans_lexlevel); /* copy into 'real' space */
			if (lsize == rsize)
			  check_exp = dim_exp;
			else
			  check_exp = newdopnode(S_EQ, thislocn, dim_exp,
						 newconstant(lsize / rsize), S_INT);
			check_exp = newdopnode(S_CCNT1, thislocn, check_exp,
					       newconstant(1), S_INT);
			(void) create_abbreviation(check_exp, TRUE);
		      }
		      /*}}}  */
		  #endif
		  }
		  /*}}}  */
	      }
	      /*}}}  */

	      if (!wasconstant && isconst(*rhsaddr)) /* bug 1040 13/11/90 */
		{
		  /* We keep on getting bugs because of non-constants in the
		     front-end being turned into constants in the back end
		     by INLINING functions.
		     This makes sure that we don't - bug 1040 13/11/90 */
		  *rhsaddr = newvalofnode(S_VALOF, thislocn,
					  newleafnode(S_SKIP, thislocn),
					  newlistnode(S_LIST, thislocn, *rhsaddr, NULL));
		}

	      trans_flag_scalar_chans = saved_flag_scalar_chans;
	    }
	    break;
	  /*}}}  */
	  /*{{{  PLACE */
	  case S_PLACE:
	    transexp(DValAddr(t), FALSE);
	    break;
	  /*}}}  */
	}
      tptr = DBodyAddr(t);
    }
  trans_data = saved_data;
  return tptr;
}
/*}}}  */
/*{{{  PRIVATE void      sub_transexp */
PRIVATE void sub_transexp (treenode **tptr, BOOL is_sub_expression)
/* This is trans for an expression. */
/* 'is_sub_expression' is FALSE if we're just about to store the expression:
   Bug TS/1324 24/10/91 */
{
  while (*tptr != NULL)
    {
      treenode *t;
      treenode **original_tptr = tptr; /* used to optimise valofs */
      tptr = transspecs(tptr);
      t = *tptr;
      if (t == NULL) return;
      trans_locn = LocnOf(t);
      switch(nodetypeoftag(TagOf(t)))
	/*{{{  cases */
	{
	  /*{{{  default - error */
	  default:
	    badtag(trans_locn, TagOf(t), "sub_transexp");
	    return;
	  /*}}}  */
	  /*{{{  LIST */
	  case LISTNODE:
	    sub_transexp(ThisItemAddr(t), is_sub_expression);
	    tptr = NextItemAddr(t);
	    break;
	  /*}}}  */
	  /*{{{  monadic and conversion */
	  case MOPNODE:
	    switch (TagOf(t))
	      {
		/*{{{  default */
		default:
		  tptr = OpAddr(t);
		  is_sub_expression = TRUE;
		  break;
		/*}}}  */
		/*{{{  checked -> unchecked */
		case S_NEG:
		  if (!NEED_ERRORS) /* INSdi02354 */
		    {
		      if (isshortint(MOpTypeOf(t)))
			SetTag(t, S_UMINUS);
		    }
		  tptr = OpAddr(t);
		  is_sub_expression = TRUE;
		  break;
		/*}}}  */
		/*{{{  ELSISE, SEGSTART */
		case S_ELSIZE: case S_SEGSTART:
		  return;
		/*}}}  */
		/*{{{  conversions */
		case S_EXACT: case S_ROUND: case S_TRUNC: /* Added 20/6/90 for bug 334 */
		  {
		    const int sourcetype = ntypeof (OpOf(t));
		    const int desttype   = MOpTypeOf(t);
		    if (((desttype   == S_INT) ? targetintsize : desttype  ) ==
			((sourcetype == S_INT) ? targetintsize : sourcetype))
		      *tptr = OpOf(t);
		    else
		      {
			tptr = OpAddr(t);
			is_sub_expression = TRUE;
		      }
		  }
		  break;
		/*}}}  */
		/*{{{  BYTESIN */
		#ifdef OCCAM2_5
		case S_BYTESIN:
		  /* it must be a data object who's size isn't known;
		     otherwise it would have been constant folded.
		  */
		  {
		    treenode *type = gettype(OpOf(t));
		    treenode *result = NULL;
		    const INT32 base_bytes = known_bytesin(type);

		    if (base_bytes != 1)
		      result = newconstant(base_bytes);

		    for (; TagOf(type) == S_ARRAY;
			 type = follow_user_type(ARTypeOf(type)))
		      if (ARDimOf(type) == (-1))
			result = add_to(result, S_MULT,
				 copytree(ARDimLengthOf(type), trans_lexlevel));
		    freetree(tptr);
		    *tptr = result;
		  }
		  break; /* re-iterate with the new tree */
		#endif
		/*}}}  */
	      }
	    break;
	  /*}}}  */
	  /*{{{  CONSTRUCTOR */
	  case LITNODE: /* S_CONSTRUCTOR */
	    if (is_sub_expression && trans_params->abbr_constructors &&
		!no_anonymous_vars) /* bug TS/1455 5/11/91 */
	      {
		treenode *name;
		/* we must run trans on the sub-expressions _after_
		   structuring the assignments. */
		/*sub_transexp(OpAddr(t), FALSE);*/ /* will be a major expression of the abbr */

		name = create_abbreviation(t, TRUE);
		create_initialised_constructor(name); /* bug TS/1263 11/05/92 */

	      #ifdef OCCAM2_5
		/* This can't happen - it is impossible for the abbreviation
		   to overwrite the node, because the abbreviation must precede the
		   whole process/expression.
		   NOT TRUE: it can now happen when a constructor is listed
		   as the result of a VALOF.
		*/
		if (*tptr == t) /* the abbreviation has not overwritten this node */
		  *tptr = name;
		else
		  {
		    /* we have transformed the current node from 'expression' into
		       VAL name IS expression :
		       VALOF
			 SKIP
			 RESULT expression
		       We now need to overwrite the 'expression' in result list
		       with the name of the abbreviation.
		    */
		    treenode *const resultlist = VLResultListOf(skipspecifications(*tptr));
		    /*fprintf(outfile, "sub_transexp: trans const -> VALOF at trans_locn:%d\n", trans_locn);*/
		    NewItem(name, resultlist);
		  }
	      #else
		*tptr = name;
	      #endif

		trans(DValAddr(NDeclOf(name))); /* run this _after_ restructuring */

		return;
	      }
	    tptr = LitExpAddr(t);
	    break;
	  /*}}}  */
	  /*{{{  dyadic */
	  case DOPNODE:
	    switch(TagOf(t))
	      {
		/*{{{  default - normal */
		default:
		  sub_transexp(LeftOpAddr(t), TRUE);
		  tptr = RightOpAddr(t);
		  is_sub_expression = TRUE;
		  break;
		/*}}}  */
		/*{{{  checked -> unchecked */
		case S_ADD: case S_SUBTRACT: case S_MULT:
		  if (!NEED_ERRORS) /* INSdi02354 */
		    {
		      const int type = DOpTypeOf(t);
		      if (!isreal(type) && (TagOf(t) == S_MULT))
			SetTag(t, S_TIMES);
		      else if (isshortint(type))
			SetTag(t, (TagOf(t) == S_ADD) ? S_PLUS : S_MINUS);
		    }
		  sub_transexp(LeftOpAddr(t), TRUE);
		  tptr = RightOpAddr(t);
		  is_sub_expression = TRUE;
		  break;
		/*}}}  */
		/*{{{  shifts */
		case S_LSHIFT: case S_RSHIFT:
		  if ((errormode & ERRORMODE_SHIFTCHECK) && !isconst(RightOpOf(t)))
		    /*{{{  add tree to range check the shift count */
		    {
		      const INT32 shiftrange = (8 * (INT32)bytesinscalar(DOpTypeOf(t))) + 1;
		      SetRightOp(*tptr, newdopnode(S_CSUB0, LocnOf(t), RightOpOf(t),
						   newconstant(shiftrange),
					S_INT));
		    }
		    /*}}}  */
		  sub_transexp(LeftOpAddr(t), TRUE);
		  tptr = RightOpAddr(t);
		  is_sub_expression = TRUE;
		  break;
		/*}}}  */
		/*{{{  eval */
		case S_EVAL:
		  if (TagOf(LeftOpOf(t)) == S_OVERLAPCHECK)
		    {
		      SetLeftOp(*tptr, buildoverlapcheck(LeftOpOf(t)));
		      overlap_checks += (TagOf(LeftOpOf(t)) != S_DUMMYEXP);
		    }
		  if (TagOf(LeftOpOf(t)) == S_DUMMYEXP)
		    {
		      *tptr = RightOpOf(t);
		      freenode(&t);
		    }
		  else
		    {
		      sub_transexp(LeftOpAddr(t), TRUE);
		      tptr = RightOpAddr(t);
		      is_sub_expression = TRUE;
		    }
		  break;
		/*}}}  */
		/*{{{  AFTER */
		/* bug TS/1568 16/01/92 turn (x AFTER y) into ((x MINUS y) > 0) */
		case S_AFTER:
		  SetTag(t, S_MINUS);
		#ifdef OCCAM2_5
		  if (DOpTypeOf(t) == S_BYTE)
		    {
		      /* x AFTER y === (signed)(x MINUS y) > 0                  */
		      /*           === ((x MINUS y) > 0) OR ((x MINUS y) < 128) */
		      /*           === ((x MINUS y) MINUS 1) < 127              */
		      /*           === (x MINUS (y PLUS 1)) < 127               */
		      /* NOTE that the latter is more efficient if y is constant */
		      if (isconst(RightOpOf(t)))
			SetRightOp(t, foldexp(newdopnode(S_PLUS, LocnOf(t), RightOpOf(t),
							 newintconstant(1, S_BYTE),
							 S_BYTE)));
		      else
			*tptr = newdopnode(S_MINUS, LocnOf(t), t,
					 newintconstant(1, S_BYTE),
					 S_BYTE);
		      *tptr = newdopnode(S_LS, LocnOf(t), *tptr,
					 newintconstant(127, S_BYTE),
					 S_BYTE);
		    }
		  else
		#endif
		    {
		      /* x AFTER y === (x MINUS y) > 0 */
		      *tptr = newdopnode(S_GR, LocnOf(t), t,
					 newintconstant(0, DOpTypeOf(t)),
					 DOpTypeOf(t));
		    }
		  /* Now loop through to properly run trans on this expression */
		  break;
		/*}}}  */
		/*{{{  AND/OR/COLON2 */
		case S_AND:
		case S_OR:
		/*case S_COLON2:*/ /* removed for bug TS/1455 4/11/91 */
		  sub_transexp(LeftOpAddr(t), TRUE);
		#if 0
		  /*{{{  attempt to optimise chains of ANDs and ORs */
		  /* I tried very simply to optimise chains of AND or OR, where each
		     item is simple, but this de-optimises other stuff done in the
		     code generator to such a degree that there is no point.
		     Better luck next time.
		     CON - 30/9/91
		  */
		  if (/*TagOf(t) != S_COLON2 &&*/ /* removed for bug TS/1455 */
		      issimplelocal(RightOpOf(t)))
		    {
		      /* we can possibly convert to a strict operation (faster) */
		      /* This converts b AND c AND d AND e into:
			  b BITAND c AND d BITAND e
			 which can be faster
		      */
		      const int new_tag = (TagOf(t) == S_AND) ? S_BITAND : S_BITOR;
		      if (TagOf(LeftOpOf(t)) != new_tag)
			{
			  DEBUG_MSG(("sub_transspec: converting short-circuit -> strict\n"));
			  SetTag(t, new_tag);
			}
		    }
		  /*}}}  */
		#endif
		#if 0 /* this doesn't allow for the fact that they are short-circuiting,
			 so we mustn't insert abbreviations 'above' them.
			 Bug TS1324 24/10/91 */
		  tptr = RightOpAddr(t);
		  is_sub_expression = TRUE;
		  break;
		#else
		  {
		    const trans_data_t saved_data = trans_data;
		    trans_data.insertvalof = TRUE; /* OK, cos it will always be a scalar BOOL */
		    trans_data.insertpoint = RightOpAddr(t);
		  /*sub_transexp(RightOpAddr(t), TagOf(t) != S_COLON2);*/ /* TS/1455 4/11/91 */
		    sub_transexp(RightOpAddr(t), TRUE);
		    trans_data = saved_data;
		    return;
		  }
		#endif
		/*}}}  */
	      }
	    break;
	  /*}}}  */
	  /*{{{  function instance */
	  case INSTANCENODE:
	    trans(tptr);  /* use the same code as for procedure instances */
	    return;
	  /*}}}  */
	  /*{{{  element */
	  case ARRAYSUBNODE:
	  case SEGMENTNODE:
	    transsubscripts(trans_params, tptr, TRUE);
	    /*{{{  look out for newly created abbreviations */
	    /* bug INSdi03483 08/02/94 */
	    while (isspecification(*tptr))
	      tptr = DBodyAddr(*tptr);
	    if (TagOf(*tptr) == S_VALOF)
	      tptr = ThisItemAddr(VLResultListOf(*tptr));
	    /*}}}  */
	    *tptr = transformelement(trans_params, *tptr, RANGECHECKING, trans_lexlevel);
	    return;
	  /*}}}  */
	  /*{{{  name */
	  case NAMENODE:
	    if (trans_flag_scalar_chans
		&& (NTypeOf(t) != NULL) /* bug INSdi02497 2/9/93 */
		&& (TagOf(NTypeOf(t)) == S_CHAN))
	      SetNChanMark(t, TRUE);
	    return;
	  /*}}}  */
	  /*{{{  constants and dummy expression */
	  case WORDNODE: case CONSTEXPNODE:
	  case CONSTTABLENODE: /*case HIDDENPARAMNODE:*/
	  case LEAFNODE: /* S_DUMMYEXP should be the only one */
	    return;
	  /*}}}  */
	  /*{{{  valof */
	  case VALOFNODE: /* this must be an inline valof */
	    trans(VLBodyAddr(t));
	    /*transexp(VLResultListAddr(t), TRUE);*/
	    trans_resultlist(tptr);
	  #if OPTIMISE_VALOFS
	  /*We optimise away specifications which have been left in front of
	    valofs which can be turned into constants:
	    (This can particularly occur with INLINE FUNCTIONs).
	    Eg:
	    z := increment(1)
	    which might be expanded into:
	    z := (VAL y IS 2 :
		  VALOF
		    SKIP
		    RESULT 2
		 )
	    which is now turned into:
	    z := 2
	  */
	    if (optimise_valof) /* we turn this off while inside segment lengths */
	    {
	      treenode *result = VLResultListOf(t);
	      /* This only works for single valued functions, so we
		 might aswell optimise it just for that case */
	    #if 0  /* 20/9/90 */
	      int constant = TRUE;
	      int results = 0;
	      while (constant && !EndOfList(result))
		{
		  constant = constant && isconst(ThisItem(result));
		  result = NextItem(result);
		  results++;
		}
	      if (constant && (results == 1))
		*original_tptr = ThisItem(VLResultListOf(t));
	    #else
	      if ((listitems(result) == 1) && isconst(ThisItem(result)))
		*original_tptr = ThisItem(result);
	    #endif
	    }
	  #endif
	    return;
	  /*}}}  */
	}
	/*}}}  */
    }
}
/*}}}  */
/*{{{  PRIVATE void      transexp */
PRIVATE void transexp (treenode **const tptr, const BOOL is_sub_expression)
/* This simply calls sub_transexp, and folds the resulting expression tree,
   incase any functions have been in-lined, which provide opportunities for
   constant propagation.
*/
{
  const BOOL temp = inline_expansions;
  inline_expansions = FALSE;
  sub_transexp(tptr, is_sub_expression);
  if (inline_expansions)
    {
      if (TagOf(*tptr) == S_LIST)
	*tptr = foldexplist(*tptr);
      else
	*tptr = foldexp(*tptr);
    }
  inline_expansions = temp;
}
/*}}}  */
/*{{{  PRIVATE void      trans */
/*{{{  comment */
/*****************************************************************************
 *
 *  trans takes a complete SC tree and performs the following transformations:
 *        1.  ARRAYSUB trees are expanded to ARRAYITEM trees
 *        2.  SEGMENT trees are expanded to SEGMENTITEM trees
 *        3.  Temporaries are inserted (but not scoped) on segment start and
 *            length trees if required
 *        4.  Hidden dimensions are inserted in all type trees.
 *        5.  Formal parameter lists are augmented with hidden dimensions
 *            and result pointers.
 *        6.  Checks are added to shift operators if required.
 *        7.  Inline routine calls are expanded out.
 *        8.  All variables have their SetNMode field set correctly.
 *        9.  Constructors are `abbreviated' out so that they only appear
 *            in abbreviations.
 *
 *****************************************************************************/
/*}}}  */
PRIVATE void trans ( treenode **tptr )
{
  BOOL going = TRUE;
  while (going && (*tptr != NULL))
    {
      /*{{{  save insertpoint */
      const trans_data_t saved_data = trans_data;
      /*}}}  */
      treenode *t;
      int tag;
      tptr = transspecs(tptr);
      t = *tptr; /* t is never modified after this */
      if (t == NULL) return;
      tag = TagOf(t);
      trans_locn = LocnOf(t);
      /*{{{  update insertpoint for anonymous variables */
      /* if the tag is FINSTANCE, we were just called from sub_transexp */
      if (tag != S_FINSTANCE) /* bug TS/1453 1/11/91 */
	{
	  trans_data.insertpoint = tptr;
	  trans_data.insertvalof = FALSE;
	}
      /*DEBUG_MSG(("trans: setting trans_data at %lX on %s\n", trans_locn, itagstring(tag)));*/

      /*}}}  */
      switch(nodetypeoftag(tag))
	/*{{{  cases */
	{
	  /*{{{  default - error */
	  default:
	    badtag(trans_locn, tag, "trans");
	    break;
	  /*}}}  */
	  /*{{{  SEQ IF ALT PAR PRIALT PRIPAR GUY ASM */
	  case CNODE:
	    tptr = CBodyAddr(t);
	    break;
	  /*}}}  */
	  /*{{{  REPLSEQ REPLIF REPLALT REPLPAR PRIREPLALT PRIREPLPAR */
	  case REPLCNODE:
	    transexp(ReplCStartExpAddr(t), FALSE);
	    transexp(ReplCLengthExpAddr(t), FALSE);
	    while (isspecification(*tptr))
	      tptr = DBodyAddr(*tptr);    /* skip over any new declarations */
	    if (parrepl(tag))
	      /*{{{  replicated PAR */
	      {
		if (trans_params->replpar_becomes_proc)
		  {
		    trans_process_replpar(tptr);
		  }
		else
		  {
		    trans_lexlevel++;
		    DEBUG_MSG(("trans: raising lex level as we enter a repl par\n"));
		    trans(ReplCBodyAddr(t));
		    trans_lexlevel--;
		    DEBUG_MSG(("trans: lowering lex level as we leave a repl par\n"));
		  }
		going = FALSE;
	      }
	      /*}}}  */
	    else if ((tag == S_REPLSEQ) &&
		     (TagOf(ReplCLengthExpOf(t)) == S_CONSTEXP)  &&
		     (LoValOf(ReplCLengthExpOf(t)) == 1))
	      /*{{{  'fold' replicator FROM ... FOR 1 */
	      {
		/* bug INSdi01228 */
		/* Don't do this for replicated PAR because that would need
		   adjusting all the lexlevels of locally declared variables.
		   Repl IF and repl ALT have the problem about what to do
		   when they appear as the outermost construct of their type.
		*/
		trans_process_repl_for_one(tptr);
	      }
	      /*}}}  */
	    else
	      {
		tptr = ReplCBodyAddr(t);
	      }
	    break;

	  /*}}}  */
	  #ifdef RESOURCES
	  /*{{{  CLAIM */
	  case CLAIMNODE:
	    transexp (ClaimVarAddr(t), FALSE);
	    tptr = ClaimBodyAddr (t);
	    break;
	  /*}}}  */
	  #endif
	  /*{{{  WHILE CHOICE */
	  case CONDNODE:
	    if (tag == S_VARIANT)
	      {
		const BOOL saved_anonymous_vars = no_anonymous_vars;
		/* None inside the input list of a case input - bug 1455 5/11/91 */
		/* OK even if a CASE rather than CASE input, because they will be
		   constant anyway. */
		no_anonymous_vars = TRUE;
		transexp(VRTaggedListAddr(t), FALSE);
		no_anonymous_vars = saved_anonymous_vars;
	      }
	    else
	      transexp(CondGuardAddr(t), FALSE);
	    tptr = CondBodyAddr(t);
	    break;
	  /*}}}  */
	  /*{{{  ALTERNATIVE */
	  case ALTNODE:
	    {
	      /* modified 2/11/90 to create a copy of the channel expression
		 so that it isn't mapped twice; once when enabling, and
		 once when the input is actually performed */
	      /* bug TS/1626 21/04/92; Now we move the input into the body of the ALT */
	      treenode *input = AltInputOf(t);

	      /*{{{  Set up Channel and Time expressions */
	      if ((input != NULL) && (AltChanExpOf(t) == NULL))
		{
		  switch(inputtypeof(input))
		    {
		      case INP_INPUT: case INP_CASE_INPUT: case INP_TAGGED_INPUT:
			SetAltChanExp(t, copytree(LHSOf(input), trans_lexlevel));
			break;
		      case INP_DELAYED_INPUT:
			SetAltChanExp(t, LHSOf(input));
			SetAltTimeExp(t, RHSOf(input));
			freenode(AltInputAddr(t));
			input = NULL;
			break;
		      default: /* various types of SKIP guard */
			break; /* leave the channel expression as NULL */
		    }
		}
	      /*}}}  */
	      /*{{{  Set up Guard */
	      if (AltGuardOf(t) == NULL)
		SetAltGuard(t, newconstant(1)); /* bug TS/1584 29/01/92 */
	      /*}}}  */
	      /*{{{  Move the input to the body */
	      /* bug TS/1626 21/04/92; Now we move the input into the body of the ALT */
	      if (input != NULL)
		{
		  treenode *newbody;
		  if (AltBodyOf(t) == NULL) /* case input, etc */
		    newbody = input;
		  else if (TagOf(AltBodyOf(t)) == S_SEQ)
		    {
		      newbody = AltBodyOf(t);
		      SetCBody(newbody, addtofront(input, CBodyOf(newbody)));
		    }
		  else
		    newbody = newcnode(S_SEQ, trans_locn,
				      addtofront(input, addtofront(AltBodyOf(t), NULL)));
		  SetAltInput(t, NULL);
		  SetAltBody(t, newbody);
		}
	      /*}}}  */

	      transexp(AltGuardAddr(t),   FALSE);
	      transexp(AltChanExpAddr(t), FALSE);
	      transexp(AltTimeExpAddr(t), FALSE);
	      /*trans(AltInputAddr(t));*/ /* This will now be NULL */
	      tptr = AltBodyAddr(t);
	    }
	    break;
	  /*}}}  */
	  /*{{{  LIST */
	  case LISTNODE:
	    trans(ThisItemAddr(t));
	    tptr = NextItemAddr(t);
	    break;
	  /*}}}  */
	  /*{{{  SKIP STOP END */
	  case LEAFNODE:
	    going = FALSE;
	    break;
	  /*}}}  */
	  /*{{{  instance */
	  case INSTANCENODE:
	    {
	      treenode *tt;
	      treenode *const name = INameOf(t);
	      const SOURCEPOSN locn = trans_locn;
	      const BOOL inside_fn = (tag == S_FINSTANCE);

	      /*{{{  call trans on actual parameters */
	      {
		/* we have to call trans on actual parameters first, so that
		   we correctly expand actual parameters which are themselves
		   instances of inlined functions - bug 1004 4/10/90 */
		const int saved_checks = overlap_checks;
		const BOOL saved_optimise_valof = optimise_valof;
		const BOOL dont_optimise_params = /* bug 1179 26/9/91 */
		    ((TagOf(name) == N_PREDEFPROC) && (NModeOf(name) == PD_ASSERT));
		overlap_checks = 0;
		optimise_valof = !dont_optimise_params; /* bug 1179 26/9/91 */

		transexp(IParamListAddr(t), TRUE); /* do params first */
					       /* added for bug TS/1530 11/12/91 */
		while (isspecification(*tptr)) /* incase anon vars have been added */
		  tptr = DBodyAddr(*tptr);
		if (TagOf(*tptr) == S_VALOF)
		  tptr = ThisItemAddr(VLResultListOf(*tptr));
		tt = *tptr;

		overlapwarnings(locn); /* bug 1118 25/1/91 */
		overlap_checks = saved_checks;
		optimise_valof = saved_optimise_valof;
	      }
	      /*}}}  */

	      if (isinline (name))
		/*{{{  expand out the inline routine definition */
		{
		  *tptr = expand_inline_instance(tt, locn, trans_lexlevel);
		  foldtree(*tptr); /* fold procedure instance */

		#if 0
		  printf("trans: expanded body of INLINE routine is\n");
		  printsrc(stdout, 0, *tptr);
		  printf("\n");
		#endif

		  inline_expansions = TRUE;
		  if (inside_fn)
		    sub_transexp(tptr, TRUE);
		  else
		    trans(tptr);
		}
		/*}}}  */
	      else
		{
		  /* transexp(IParamListAddr(tt)); */ /* done already */
		  if (inside_fn && (TagOf(name) == N_PREDEFFUNCTION))
		    /*{{{  special transformations for some predefines */
		    {
		      switch (NModeOf(name))
			{
			  /*{{{  SHIFT predefines */
			  /* TS/1979 30/11/92 - CON
			     Note that these are illegal in this implementation because the naive
			     instruction sequence would generate the incorrect result,
			     and we don't want to pessimise the 'quick' case to cope with the
			     nasty cases.

			     SHIFTLEFT and SHIFTRIGHT are permitted to have 'long' shift values, because
			     they still produce the correct results.
			  */
			  case PD_ASHIFTLEFT: case PD_ASHIFTRIGHT:
			  case PD_ROTATELEFT: case PD_ROTATERIGHT:
			  case PD_BITREVNBITS:
			    {
			      /* insert a range check on the 'places' (second) parameter */
			      treenode *const param2 = NextItem(IParamListOf(tt));
			      if ((errormode & ERRORMODE_SHIFTCHECK) && (!isconst(ThisItem(param2))))
				NewItem(newdopnode(S_CSUB0, locn, ThisItem(param2),
					newconstant((8 * (INT32)bytesinscalar(S_INT)) + 1), S_INT),
					param2);
			    }
			    break;
			  /*}}}  */
			  default:
			    break;
			}
		    }
		    /*}}}  */
		}
	    }
	    going = FALSE;
	    break;

	  /*}}}  */
	  /*{{{  action */
	  case ACTIONNODE:

	    /*{{{  Source transformations for i/o */
	    if ((ActionTypeOf(t) == NULL) &&
		trans_params->simplify_structured_io)
	      {
		BOOL break_out_of_switch;
		switch(tag)
		  {
		    /*{{{  sequential i/o */
		    case S_INPUT: case S_OUTPUT:
		      {
			const BOOL modified = trans_process_sequential_io(tptr);
			break_out_of_switch = modified;
		      }
		      break;
		    /*}}}  */
		    /*{{{  tagged input */
		    case S_TAGGED_INPUT:
		      trans_process_tagged_input(tptr);
		      break_out_of_switch = TRUE;
		      break;

		    /*}}}  */
		    /*{{{  case input */
		    case S_CASE_INPUT: 
		      trans_process_case_input(tptr);
		      break_out_of_switch = TRUE;
		      break;
		    /*}}}  */
		    default:
		      break_out_of_switch = FALSE;
		      break;
		  }
		if (break_out_of_switch)
		  break; /* break out of master switch statement */
	      }
	    /*}}}  */
	    /*{{{  Source transformations for counted array i/o */
	    if (!EndOfList(RHSOf(t)) &&
		(TagOf(RHSOf(t)) == S_LIST) && /* INSdi01982 */
		(TagOf(ThisItem(RHSOf(t))) == S_COLON2) &&
		trans_params->simplify_counted_io)
	      {
		if (tag == S_INPUT || tag == S_OUTPUT)
		  trans_process_counted_io(tptr);
		else
		  badtag(trans_locn, tag, "trans_counted_array");
		break; /* break out of master switch statement */
	      }
	    /*}}}  */
	    /*{{{  Source transformation for constructor assignment */
	    if ((tag == S_ASS) && (TagOf(RHSOf(t)) == S_CONSTRUCTOR) &&
		trans_params->abbr_constructors &&
		!usedin(LHSOf(t), LitExpOf(RHSOf(t)), trans_lexlevel))
	      {
		trans_constructor_assign(tptr);
		break; /* break out of master switch statement */
	      }
	    /*}}}  */
	    /*{{{  Source transformation for BOOL CASE */
	    if ((tag == S_CASE) && (ntypeof (LHSOf(t)) == S_BOOL))
	      {
		const BOOL break_out_of_switch = trans_bool_case(tptr);
		if (break_out_of_switch)
		  break; /* break out of master switch statement */
	      }
	    /*}}}  */

	    switch(tag)
	      {
		case S_CASE: case S_CASE_INPUT:
		  transexp(LHSAddr(t), FALSE);

		  if ((tag == S_CASE) /* added for bug 1052 - 23/11/90 */
		      && isshortint(ntypeof (LHSOf(t))))
		    {
		      SetLHS(*tptr, newmopnode(S_EXACT, LocnOf(t),
					       LHSOf(t), S_INT));
		      /* really we should constant fold here, but nobody uses
			 constants in CASE expressions! */
		      /* Note too that we don't need to change the type of
			 all the selection values - they fall out in the wash */
		    }
		  tptr = RHSAddr(t);
		  break;
		default:
		  {
		    const SOURCEPOSN thislocn = trans_locn;
		    const BOOL saved_anonymous_vars = no_anonymous_vars;
		    const BOOL abbreviate_constructors = (tag == S_OUTPUT) || /* bug TS/1238 11/12/91 */
			       ((tag == S_ASS) && ((TagOf(RHSOf(t)) == S_LIST) || /* bug TS/1263 11/05/92 */
						   (TagOf(RHSOf(t)) == S_CONSTRUCTOR))); /* INSdi03379 */

		    overlap_checks = 0;
		    transexp(LHSAddr(t), FALSE);

		    /* None inside the input list of an input - bug 1455 5/11/91 */
		    if (!trans_params->simplify_structured_io) /* bug TS/1626 means this is unnecessary! */
		      no_anonymous_vars = tag == S_INPUT ||
					  tag == S_TAGGED_INPUT;

		    /*transexp(RHSAddr(t), FALSE);*/
		    transexp(RHSAddr(t), abbreviate_constructors);
		    overlapwarnings(thislocn);
		    no_anonymous_vars = saved_anonymous_vars;
		  }
		  going = FALSE;
		  break;
	      }
	    break;
	  /*}}}  */
	  /*{{{  valof */
	  case VALOFNODE:  /* This must be a function body */
	    trans(VLBodyAddr(t));
	    /*transexp(VLResultListAddr(t), TRUE);*/
	    trans_resultlist(tptr);
	    going = FALSE;
	    break;
	  /*}}}  */
	  /*{{{  GUYCODE GUYSTEP */
	  case DOPNODE:
	    if (tag == S_GUYCODE)
	      {
		const BOOL saved_flag_scalar_chans = trans_flag_scalar_chans;
		trans_flag_scalar_chans = TRUE;
		transexp(RightOpAddr(t), FALSE);
		trans_flag_scalar_chans = saved_flag_scalar_chans;
	      }
	    going = FALSE;
	    break;
	  /*}}}  */
	}
	/*}}}  */
      /*{{{  reset insertpoint */
      trans_data = saved_data;
      /*DEBUG_MSG(("trans: resetting trans_data\n"));*/
      /*}}}  */
    }
}
/*}}}  */

/*{{{  PUBLIC void transmain */
PUBLIC void transmain (const trans_params_t *const trans_params_in, treenode **tptr )
{
  jmp_buf savedenv;
  const int savedlexlevel = trans_lexlevel;
  const trans_params_t *const saved_params = trans_params;
  const BOOL saved_flag_scalar_chans = trans_flag_scalar_chans;
  trans_lexlevel = 0;
  trans_params   = trans_params_in;
  trans_flag_scalar_chans = FALSE;

  memcpy((char *)savedenv, (char *)env, sizeof(env));
  if (setjmp(env) == 0)
    trans(tptr);
  else
    *tptr = NULL;
  memcpy((char *)env, (char *)savedenv, sizeof(env));

  trans_flag_scalar_chans = saved_flag_scalar_chans;
  trans_lexlevel = savedlexlevel;
  trans_params   = saved_params;
}
/*}}}  */
