/*
 * KROC Scheduler startup code for occam - version for systems without fork,signal
 *      Also contains extensions for external communications using 3L threads
 *
 * (C) Copyright 1995 D.C. Wood <D.C.Wood@ukc.ac.uk> and 
 *                    P.H. Welch <P.H.Welch@ukc.ac.uk>
 * University of Kent at Canterbury
 *
 * Based on work by: Peter Shephard @ QMC Sept. 1984 for
 * UNIX OPS 2.0 (Release 2.0.1 : 24th. July, 1986)
 */


#include "minrts.h"
#include <assert.h>

#ifdef _3L
#include <timer.h>
/*{{{  external link handling*/
#include <thread.h>
#include <chan.h>
/*{{{  struct linkdata*/
typedef struct
{
  int linkno;
  int flag;
  int wptr;
  CHAN * intl;
  CHAN * extl;
} linkdata;
/*}}}  */
int input_links = 0;  /* number of input links to task */
int output_links = 0; /* number of output links to task */
int links_pending = 0; /* number of links with pending transfers */
linkdata extlinks[8];
int linkdata_stride = sizeof (linkdata);
void* in_links = &extlinks[0];
CHAN thout[4], thin[4];  /* internal channels from occam rts to threads */
CHAN ** inports, ** outports; /* pointers to external channel arrays */
/*}}}  */
#endif

char * short_cmdline;
char * long_cmdline;
FILE * kroc_out = stdout;
FILE * kroc_err = stderr;
/*{{{  savecmdline*/
/* This function sets up the pointers long_cmdline, short_cmdline which are
   used in the implementation of hostio command line PROCs */
void savecmdline (int argc, char **argv)
{
  static char cmdline[512];
  int i = 0;
  cmdline[0] = '\0';
  long_cmdline = &cmdline[0];
  while (i < argc)
  {
    strcat (cmdline, argv[i]);
    i++;
    if (i < argc)
      strcat (cmdline, " ");
    if (i == 1)
      short_cmdline = &cmdline[strlen(cmdline)];
  }
}
/*}}}  */
/*{{{  rtime*/
/* This function should return an occam time in microseconds */

#ifdef _3L
unsigned int
rtime (void)                    /* real (clock) time    */
{
  unsigned int occamtime = 1000 * timer_now();
  return occamtime;
}
#endif
/*}}}  */
/*{{{  test_the_kbd*/
int
test_the_kbd (void)             /* called from assembler */
{
#ifdef _3L
#ifdef ALONE
   return 0;
#else
   return (kbhit());
#endif
#else
   return 0;
#endif
}
/*}}}  */
/*{{{  read_the_kbd*/
char
read_the_kbd (void)             /* called from assembler */
{
#ifdef _3L
#ifdef ALONE
   return 0;
#else
   /* this version allows a key to be read, but does not support ALT on keyboard */
   return ((char)getch());
#endif
#else
   return(0);
#endif
}
/*}}}  */
/*{{{  acknowledge (keyboard)*/
void
acknowledge (void)                      /* called from assembler */
/* not needed in the absence of asynchronous input, so does nothing */
{
}
/*}}}  */
/*{{{  safe_pause*/
int
safe_pause (void)
/* A proper implementation allows current process to hibernate until an
   interrupt routine sets tim_sync, kbd_sync or link_sync non-zero */
{       
#ifndef ALONE
#if 0
  putchar ('p');
  thread_deschedule();
  if (kbd_sync == 0)
  {
    timer_delay (5);
    if (kbhit())
      kbd_sync = 1;
  }
#endif
#endif
  return 0;
}
/*}}}  */
/*{{{  setalarm*/
/* This function need only do anything if it is possible to set (and reset) an
   alarm time, at the expiry of which a signal handler can set tim_sync to 1 */
/* Its effect should be null if the kernel does busy waiting when nothing else
   to do */
void setalarm (timethen, dummy, timenow)
  unsigned int timethen, dummy, timenow;
{
/* ualarm (timethen - timenow, 0); */  /* signals not available  */
}
/*}}}  */
/*{{{  dead*/
void
dead (erfl)                     /* called by scheduler */
int erfl;
{
#if (!defined(ALONE))
  fputs ("\nKROC: Program deadlocked (no processes to run)\n", stderr);
  fflush (stderr);              /* still in raw mode */
  occam_exit (2, 0);
#endif
} /* dead */
/*}}}  */
/*{{{  showqueue*/
#ifndef ALONE
#if DIAG
void showqueue (const int front, const int back)
{
  PTRINT qp, bp;
  int * link;
  int Iptr;
  qp = (PTRINT) front;
  bp = (PTRINT) back;
  fprintf (stderr, "\nQueue:");
  for (;;)
  {
    Iptr = *((int*)(qp-4));
    fprintf (stderr, "%p(%X);", (int*)(qp), Iptr);
    if (qp == bp) break;
    link = (int *)(qp - 8);       /* link is at W[-2] */
    qp = (PTRINT) (*link);
  }
  fprintf (stderr, "Endq\n");
}
#endif
#endif
/*}}}  */
/*{{{  impossible*/
void impossible (int reason, int occam_iptr)      /* called by scheduler */
{       
#if !defined(ALONE)
  fprintf (stderr,
   "\nKROC: Impossible situation: %d\n", reason);
  fflush (stderr);              /* still in raw mode */
  occam_exit (2, occam_iptr);
#endif
}
/*}}}  */
#ifdef _3L
/*{{{  C_inmessage*/
int C_inmessage (int len, void * msg, int linkno, int wptr)
{
  if (linkno >= input_links)
  /*{{{  error message*/
  {
  #ifndef ALONE
    fprintf (stderr, "Attempting input on undefined port\n");
  #endif
    occam_exit(2, *(((int *)wptr) - 1));
  }
  /*}}}  */
  else
  /*{{{  communicate with the corresponding thread*/
  /* this version relies on msg address already being stored
     at wptr+W_Pointer */
  {
    links_pending++;
  #ifndef ALONE
  #if DIAG
    printf ("Inputting %d byte message (%X) on link %d at %p (pending = %d)\n",
	     len, wptr+W_Pointer, linkno, msg, links_pending);
  #endif
  #endif
    chan_out_word (len, &thin[linkno]);
    chan_out_word (wptr, &thin[linkno]);
  }
  /*}}}  */
  #ifndef ALONE
  #if DIAG
    printf ("input message accepted\n");
  #endif
  #endif
  return 0; /* says deschedule this process now */
}
/*}}}  */
/*{{{  C_outmessage*/
int C_outmessage (int len, void * msg, int linkno, int wptr)
{
  if (linkno >= output_links)
  /*{{{  error message*/
  {
  #ifndef ALONE
    fprintf (stderr, "Attempting output on undefined port\n");
  #endif
    occam_exit(2, *(((int *)wptr) - 1));
  }
  /*}}}  */
  else
  {
    links_pending++;
  #ifndef ALONE
  #if DIAG
    printf ("Outputting %d byte message (%X) on link %d at %p (pending = %d)\n",
	   len, wptr+W_Pointer, linkno, msg, links_pending);
  #endif
  #endif
    chan_out_word (len, &thout[linkno]);
    chan_out_word (wptr, &thout[linkno]);
  }
  #ifndef ALONE
  #if DIAG
    printf ("output message accepted\n");
  #endif
  #endif
  return 0; /* says deschedule this process now */
}
/*}}}  */
/*{{{  link_thread*/
void link_thread (void *p)
{
  linkdata * port = (linkdata *)p;
  int size;
  int mwptr;
  PTRINT lpdata, pdata;
  void * addr;
  for (;;)
  {
    chan_in_word (&size, port->intl);
    if (size < 0) return;
    chan_in_word (&mwptr, port->intl);
    lpdata = (PTRINT)(mwptr + W_Pointer);
    pdata = *((int *)lpdata);
    addr = (void *)pdata;
    /* address of source stored in W[W_Pointer] */
    if (port->linkno & 1)
      chan_out_message (size, addr, port->extl);
    else
      chan_in_message (size, addr, port->extl);
    port->flag = 1;
    port->wptr = mwptr;
    link_sync = 1;
  }
}
/*}}}  */
/*{{{  init_links*/
void init_links (const int num_ins, const int num_outs)
{
    int i,j;
    void* thrws;

    input_links = num_ins;
    output_links = num_outs;

    for (i=0; i<8; i++)
    {
      extlinks[i].linkno = i;
      extlinks[i].flag = 0;
      extlinks[i].wptr = NOPROCESS;
    }
    for (j=0; j<num_ins; j++)
    {
      i = 2*j;
      extlinks[i].intl = &thin[j];
      extlinks[i].extl = inports[j];
      chan_init(extlinks[i].intl);
      thrws = malloc (THSTACKSIZE);
      thread_launch(link_thread, thrws, THSTACKSIZE, THREAD_URGENT,
		    (void *)&extlinks[i]);
      #ifndef ALONE
      #if DIAG
      fprintf (stderr, "Input thread launched for port at %p\n", inports[j]);
      #endif
      #endif
    }
    for (j=0; j<num_outs; j++)
    {
      i = (2*j)+1;
      extlinks[i].intl = &thout[j];
      extlinks[i].extl = outports[j];
      chan_init(extlinks[i].intl);
      thrws = malloc (THSTACKSIZE);
      thread_launch(link_thread, thrws, THSTACKSIZE, THREAD_URGENT,
		    (void *)&extlinks[i]);
      #ifndef ALONE
      #if DIAG
      fprintf (stderr, "Output thread launched for port at %p\n", outports[j]);
      #endif
      #endif
    }
}
/*}}}  */
/*{{{  shutdown_links*/
void shutdown_links ()
{
  int i;
  for (i=0; i<input_links; i++)
    chan_out_word (-1, &thin[i]);
  for (i=0; i<output_links; i++)
    chan_out_word (-1, &thout[i]);
}
/*}}}  */
#else
/*{{{  C_inmessage*/
int C_inmessage (int len, void * msg, int linkno, int wptr)
{
  return 1; /* says behave as if transfer happened */
}
/*}}}  */
/*{{{  C_outmessage*/
int C_outmessage (int len, void * msg, int linkno, int wptr)
{
  return 1; /* says behave as if transfer happened */
}
/*}}}  */
#endif
/*{{{  occam_exit*/
void occam_exit (int exit_status, int occam_iptr)
{
#ifndef ALONE
  if (occam_iptr != 0)
      fprintf (stderr, "OCCAM STOPped at %08X\n", occam_iptr);
#endif
  restore_tty_state();
#ifdef _3L
  shutdown_links();
#endif
#ifndef ALONE
  if (exit_status != 0)
    printf ("occam_exit (%d)\n", exit_status);
#endif
  exit (exit_status);
}  /* occam_exit */
/*}}}  */
/*{{{  main*/
#ifdef _3L
int main (int argc, char *argv[], char *envp[],
	  CHAN * in_ports[], int ins, CHAN * out_ports[], int outs)
#else
int main (int argc, char ** argv)
#endif
{
  int erfl;
#ifdef ALONE
#else
  savecmdline (argc, argv);
#endif
  /*{{{  save ports*/
  #ifdef _3L
    {
      inports = in_ports;
      outports = out_ports;
    }
  #endif
  /*}}}  */
  save_tty_state();
#ifdef _3L
  init_links (ins, outs);
#endif
  erfl = enter ();                       /* run the occam program */
  occam_exit (erfl, 0);                  /* should never get here */
}  /* main */
/*}}}  */
