/*
 * KROC translator: Function definitions and prototypes
 *
 * $Source: /extra/play/cvs/kroc/src/octran.h,v $
 *
 * $Id: octran.h,v 1.2 1999/09/08 17:00:29 rivimey Exp $
 *
 * asm.h
 * (C) Copyright 1999 R. Ivimey-Cook <ruthc@sharra.demon.co.uk>
 *
 */

#ifndef OCTRAN_ASM_H
#define OCTRAN_ASM_H


/*
 * Write instruction mnem, and up to 3 arguments, to the current line. If
 * the instruction has less than three args, then arg3, ... arg1 are
 * NULL. It is not valid for arg1, 2 to be NULL and arg3 not to be NULL.
 *
 * This function starts a new line before the mnemonic is written, unless
 * immediately preceded by a .
 */
typedef void OperandProc (char *mnem, char *arg1, char *arg2, char *arg3);

/*
 * Write a equate to the current line, and move to the next line.
 */
typedef void EquateProc (char *sym, char *val);

/*
 * Write an alignment pseudo-op to the current line, and move to the next line.
 */
typedef void AlignProc (int sz);

/*
 * Write a comment to the current line, and move to the next line.
 */
typedef void CommentProc (char *str);

/*
 * Introduce a global function (that is, a function that is exported
 * to the linker). Name is not quoted: if it needs to be, the assembler
 * interface must do this.
 */
typedef void GlobalFunctionProc (char *name);

/*
 * Introduce a local function (that is, a function that is not exported
 * to the linker). Name is not quoted: if it needs to be, the assembler
 * interface must do this.
 */
typedef void LocalFunctionProc (char *name);

/*
 * The last function (global/local) has now finished.
 * Write any end-of-function code here (e.g. a newline, an
 * ENDP pseudo-op, etc).
 */
typedef void FunctionExitProc (char *name);

/*
 * Do anything that must be done at the start of each
 * file. The file handle is already open; this is for
 * header comments, etc.
 */
typedef void StartOfFileProc (void);

/*
 * Do anything that must be done at the end of a file.
 * For example, some assemblers require an end-of-file
 * directive. You do not have to close the file here.
 */
typedef void EndOfFileProc (void);

/*
 * Import a symbol, that is, make symbol name available
 * to the assembler from this point on. 
 */
typedef void ImportSymbolProc (char *name);

/*
 * Export a symbol, that is, put symbol in the assembler object
 * file. Symbol must have been defined first.
 */
typedef void ExportSymbolProc (char *name);

/*
 * Write the value value into the current output. Size is 
 * 1, 2, 4 or 8 and measured in bytes. Value is in host-
 * endianness.
 */
typedef void ConstantDataProc (int size, long value);

/*
 * Write the string into the current output. The values
 * are written such that the first character of the string
 * is at the lowest address in memory. No alignment or
 * padding is enforced.
 */
typedef void ConstantStringProc (char *str);

/*
 * Notification functions: the translator is entering
 * or leaving the translation routine for one transputer
 * instruction.
 */
typedef void EnterProc (void);
typedef void LeaveProc (void);

/*
 * Initialise and Close the code in this module.
 */
typedef void InitialiseProc (void);
typedef void CloseProc (void);

/*
 * Emit a local code label with local id 'id', and emit
 * a local code label reference in the given direction with
 * the given id. 
 * For assemblers that do not understand local labels the
 * interface must emulate them, possibly using the EnterProc/
 * LeaveProc routines.
 * In no case can a label be referred to outside the translation
 * unit.
 */
enum ReferenceType { ForwardReference, BackwardReference };
typedef void EmitLabelProc(int id);
typedef char *RefLabelProc(enum ReferenceType reft, int id);

/*
 * Write the instructions that shifts register B into C and
 * register A into B, leaving C destroyed and A available.
 */
typedef void StackShiftProc(void);

/*
 * Format an operand as an addressing mode. The definition
 * of which addressing mode is which is specific to target
 * processors but not assemblers.
 * Note: these routines return a pointer to the formatted
 * string for inclusion in an operand. They do not affect the
 * output file.
 */
typedef char *FormatIntElementProc(int val);
typedef char *FormatStrElementProc(char *str);

typedef char *FormatStrIntElementProc(char *str, int val);
typedef char *FormatStrStrElementProc(char *str, char *str2);

typedef char *FormatStrStrIntElementProc(char *str, char *str2, int val);
typedef char *FormatStrStrStrElementProc(char *str, char *str2, char *str3);


typedef struct 
{
    InitialiseProc       *asmInit;
    CloseProc            *asmClose;

    EnterProc            *asmEnter;
    LeaveProc            *asmLeave;

    OperandProc          *asmOperand;
    CommentProc          *asmComment;
    EquateProc           *asmEquate;
    AlignProc            *asmAlign;

    GlobalFunctionProc   *asmGlobalFn;
    LocalFunctionProc    *asmLocalFn;
    FunctionExitProc     *asmFuncExit;

    StartOfFileProc      *asmStartFile;
    EndOfFileProc        *asmEndFile;

    ImportSymbolProc     *asmImport;
    ExportSymbolProc     *asmExport;

    ConstantDataProc     *asmConstantData;
    ConstantStringProc   *asmConstantString;

    StackShiftProc	 *asmStackShift;
    EmitLabelProc        *asmEmitLab;
    RefLabelProc         *asmRefLab;

    FormatIntElementProc *asmIntFormats[AMI_NUMITEMS];
    FormatStrElementProc  *asmStrFormats[AMS_NUMITEMS];
    FormatStrIntElementProc  *asmStrIntFormats[AMSI_NUMITEMS];
    FormatStrStrElementProc  *asmStrStrFormats[AMSS_NUMITEMS];
    FormatStrStrIntElementProc  *asmStrStrIntFormats[AMSSI_NUMITEMS];
    FormatStrStrStrElementProc  *asmStrStrStrFormats[AMSSS_NUMITEMS];

} AssemblerProcVec;

extern AssemblerProcVec *currentAsm;

#define op(nm)			currentAsm->asmOperand(nm, 0, 0, 0)
#define op1(nm,a1)		currentAsm->asmOperand(nm, (a1), 0, 0)
#define op2(nm,a1,a2)		currentAsm->asmOperand(nm, (a1), (a2), 0)
#define op3(nm,a1,a2,a3)	currentAsm->asmOperand(nm, (a1), (a2), (a3))
#define comment(str)		currentAsm->asmComment(str)
#define align(sz)		currentAsm->asmAlign(sz)
#define equate(sy,va)		currentAsm->asmEquate(sy, va)
#define import(sy)		currentAsm->asmImport(sy)
#define export(sy)		currentAsm->asmExport(sy)
#define emitlab(id)		currentAsm->asmEmitLab(id)
#define labF(id)		currentAsm->asmRefLab(ForwardReference, id)
#define labB(id)		currentAsm->asmRefLab(BackwardReference, id)

#endif /* OCTRAN_ASM_H */
