/*
 * KROC User process
 *
 * This is the UNIX process that runs the occam code and traps its
 * termination cleanly or via signals.  The co-process handles blocking
 * on fd 0 (the keyboard) and uses a UNIX signal to deliver notification
 * of a new keypress.
 *
 * $Source: /extra/play/cvs/kroc/src/target/unix/userproc.c,v $
 *
 * $Id: userproc.c,v 1.1 1999/09/27 21:10:13 rivimey Exp $
 *
 * (C) Copyright 1995,1996,1997
 *    D.C. Wood <D.C.Wood@ukc.ac.uk>
 *    P.H. Welch <P.H.Welch@ukc.ac.uk>
 *    Dave Beckett <D.J.Beckett@ukc.ac.uk>
 * University of Kent at Canterbury
 *
 */

static char userproc_c_rcsid[] = "$Id: userproc.c,v 1.1 1999/09/27 21:10:13 rivimey Exp $";

#include "rts.h"

extern char kroc_version_message[];     /* In version.c */

/* 'local' variables */

/* These are used in sched.s */
FILE *kroc_in  = stdin;         /* not used */
FILE *kroc_out = stdout;
FILE *kroc_err = stderr;

int kbd_char_signalled = 0;
sigset_t sig_mask, zero_mask;
int trace_count = 0;

char TTin_buf;

static int kbdproc_pid, to_kbdproc, from_kbdproc;
static int stdin_is_tty;


static void user_stop_signal (void);
static void user_cont_signal (void);
static void userproc_exit (int exit_status, bool dump_core);
static void set_user_process_signals (void);


/* Called from assembler kernel */
void
occam_exit(int exit_status, int occam_iptr)
{
#ifndef ALONE
  if (occam_iptr != 0)
    fprintf (stderr, "OCCAM STOPped at %08x\n", occam_iptr);
#endif
  userproc_exit(exit_status, 0);
}


static void
userproc_exit(int exit_status, bool dump_core) 
{

#ifdef DEBUG
  if (trace) {
    fprintf (stderr, "USERPROC: Exit %d\n", exit_status); fflush (stderr);
  }
#endif

#ifdef  DEBUG
  if (trace) {
    fputs ("USERPROC: About to kill KBDPROC\n", stderr); fflush (stderr);
  }
#endif

  /* Terminate keyboard process, if still around */
  kill (kbdproc_pid, SIGTERM);

  /* reset terminal */
  if (stdin_is_tty)
    restore_tty_state ();

  if (dump_core) {
    fprintf (stderr, "KROC: Fatal error code %d, core dumped\n", exit_status); fflush (stderr);
    /* Do core dump */
    abort();
  }

  exit (exit_status);
}  /* userproc_exit */



void
user_bad_exit (void)
{
#ifdef DEBUG
  if (trace) {
    fputs ("USERPROC: I've been murdered!\n", stderr); fflush (stderr);
  }
#endif
  userproc_exit (-8, 1);
} /* user_bad_exit */


void
user_good_exit (void)
{
  userproc_exit (0, 0);
} /* user_bad_exit */


/* SIGNAL handlers */

void
user_tim_handler (void)      /* SIGALRM  handler - only set during hibernate */
{
#ifdef DEBUG
  if (trace) {
    fputs ("USERPROC: Alarm ringing (SIGALRM).\n", stderr); fflush (stderr);
  }
#endif
  tim_sync = 1;
#if defined(SIGNAL_TYPE_SYSV)
  signal (SIGALRM, (SigParam)user_tim_handler);
#endif /* SIGNAL_TYPE */
}  /* user_tim_handler */


#if 0
void
user_panic_handler (void)    /* SIGPANIC  handler - only set by panicking user */
{
#ifdef DEBUG
  if (trace) {
    fputs ("USERPROC: Panic from user (SIGPANIC).\n", stderr); fflush (stderr);
  }
#endif
  panic_sync = 1;
}  /* user_panic_handler */
#endif

void
user_kbd_handler (void)      /* SIGKBD handler - available any time */
{
#ifdef DEBUG
  if (trace) {
    fprintf (stderr, "USERPROC: Character #%d arriving.\n", trace_count);
    ++trace_count;
  }
#endif
  kbd_char_signalled = 1;
  kbd_sync = 1;
#if defined(SIGNAL_TYPE_SYSV)
  signal (SIGKBD, (SigParam)user_kbd_handler);
#endif /* SIGNAL_TYPE */
}  /* user_kbd_handler */


#if defined(SIGNAL_TYPE_SYSV)
void user_trap_handler (int sig)
{
  fprintf (stderr, "KROC: Range error / STOP executed (signal %d)\n", sig);
  userproc_exit (1, 1);
} /* user_trap_handler */
#else
void user_trap_handler (int sig, int code, struct sigcontext *scp, char *addr)
{
/*      fprintf (stderr, "sig = 0x%x, code = 0x%x\n", sig, code);       */
  fputs ("KROC: ", stderr);
  switch (code & 0x1f) {           /* why? */
    case _SETERR:
      fputs ("Set error", stderr);
      break;
    case _OFLOW:
      fputs ("Integer overflow", stderr);
      break;
    case _RANGE:
      fputs ("Range error", stderr);
      break;
    case _FLOAT:
      fputs ("Floating-point error", stderr);
      break;
    case _UNREC:
      fputs ("Unrecognized instruction", stderr);
      break;
    case _UNIMP:
      fputs ("Unimplemented instruction", stderr);
      break;
    default:
      fprintf (stderr, "Unrecognized trap number 0x%x", code);
  }
  fputs ("\n", stderr);
  fflush (stderr);
  userproc_exit (code, 1);
} /* user_trap_handler */
#endif /* SIGNAL_TYPE */


#ifdef FLOAT
#if defined(SIGNAL_TYPE_SYSV)
void user_fp_handler (int sig)
{
  fprintf (stderr, "KROC: Floating-point error\n");
  userproc_exit (42, 1);
}
#else
void user_fp_handler (int sig, int code, struct sigcontext *scp, char *addr)
{
  switch(code) {
#ifdef FPE_INTOVF_TRAP
    case FPE_INTOVF_TRAP:
      fprintf (stderr, "KROC: FP error - integer overflow\n");
      break;
#endif      
#ifdef FPE_STARTSIG_TRAP
    case FPE_STARTSIG_TRAP:
      fprintf (stderr, "KROC: FP error - process using fp\n");
      break;
#endif      
#ifdef FPE_INTDIV_TRAP
    case FPE_INTDIV_TRAP:
      fprintf (stderr, "KROC: Floating Point error - integer divide by zero\n");
      break;
#endif      
#ifdef FPE_FLTINEX_TRAP
    case FPE_FLTINEX_TRAP:
      fprintf (stderr, "KROC: Floating Point error - floating inexact result\n");
      break;
#endif      
#ifdef FPE_FLTDIV_TRAP
    case FPE_FLTDIV_TRAP:
      fprintf (stderr, "KROC: Floating Point error - floating divide by zero\n");
      break;
#endif      
#ifdef FPE_FLTUND_TRAP
    case FPE_FLTUND_TRAP:
      fprintf (stderr, "KROC: Floating Point error - floating underflow\n");
      break;
#endif      
#ifdef FPE_FLTOPERR_TRAP
    case FPE_FLTOPERR_TRAP:
      fprintf (stderr, "KROC: Floating Point error - floating operand error\n");
      break;
#endif      
#ifdef FPE_FLTOVF_TRAP
    case FPE_FLTOVF_TRAP:
      fprintf (stderr, "KROC: Floating Point error - floating overflow\n");
      break;
#endif      
    default:
      fprintf (stderr, "KROC: Unknown floating point error number 0x%x\n", code);
  }
  userproc_exit (42, 1);
}
#endif /* SIGNAL_TYPE */
/* user_fp_handler */

#endif /* FLOAT */

char
read_the_kbd (void)             /* called from assembler */
{
  if (read (from_kbdproc, &TTin_buf, 1) != 1) {
    fputs ("USERPROC: Can't read tty char from KBDPROC\n", stderr);
    fflush (stderr);
    userproc_exit (3, 0);
  }
#ifdef DEBUG
  if (trace) {
    fprintf (stderr, "USERPROC: '%c' (0x%x) was read from KBDPROC\n",
             TTin_buf, TTin_buf);
    fflush (stderr);
  }
#endif
  return (TTin_buf);            /* dcw */
} /* read_the_kbd */


void
acknowledge (void)                      /* called from assembler */
{       
  kbd_sync = 0;
#ifdef DEBUG
  if (trace) {
    fputs ("USERPROC: Acknowledge KBDPROC\n", stderr); fflush (stderr);
  }
#endif

  if (write (to_kbdproc, &TTin_buf, 1) != 1) {
#ifdef DEBUG
    if (trace) {
      fputs ("USERPROC: Can't (pipe) acknowledge keyboard\n", stderr);
      userproc_exit (3, 0);
    }
#endif
  }
}  /* acknowledge */


void
user_to_kbd_sync (void)
{
  char go_char;

  /* give go-ahead to keyboard process */
#ifdef DEBUG
  if (trace) {
    fputs ("USERPROC: Signalling to KBDPROC\n", stderr); fflush (stderr);
  }
#endif
  if (write (to_kbdproc, "c", 1) != 1) {
    fputs ("USERPROC: Can't write go-ahead\n", stderr); fflush (stderr);
    userproc_exit (3, 0);
  }

  /* wait for keyboard process's go-ahead */
#ifdef DEBUG
  if (trace) {
    fputs ("USERPROC: Waiting for KBDPROC\n", stderr); fflush (stderr);
  }
#endif

  if ((read (from_kbdproc, &go_char, 1) != 1) || (go_char != 'p')) {
    fprintf (stderr, "USERPROC: can't read KBDPROC's go-ahead '%c'\n", go_char);
    userproc_exit (3, 0);
  }
#ifdef DEBUG
  if (trace) {
    fputs ("USERPROC: Got KBDPROC go-ahead\n", stderr); fflush (stderr);
  }
#endif
} /* user_to_kbd_sync */


static void
user_stop_signal (void)
{
  /* Advanced Programming in the UNIX Environment, Stevens; P321 */
#ifdef  DEBUG
  if (trace) {
    fputs("USERPROC: Stopped\n", stderr); fflush(stderr);
  }
#endif

  /* set continue handler for after stop */
  signal (SIGCONT, (SigParam)user_cont_signal);

#ifdef  DEBUG
  if (trace) {
    fputs("USERPROC: Before kill()\n",stderr); fflush(stderr);
  }
#endif

  if (stdin_is_tty) {
    restore_tty_state ();
#ifdef  DEBUG
    if (trace) {
      fputs("USERPROC: Restored tty state\n", stderr); fflush(stderr);
    }
#endif
  }

  /* Stop the KBDPROC */
  kill (kbdproc_pid, SIGSTOP);

  /* Do the default SIGTSTP action - stop this process */
  kill (getpid(), SIGSTOP);

#ifdef  DEBUG
  if (trace) {
    fputs("USERPROC: After kill()\n", stderr); fflush(stderr);
  }
#endif

#if defined(SIGNAL_TYPE_SYSV)
  /* Re-establish signal handler */
  signal (SIGTSTP, (SigParam)user_stop_signal);
#endif
}


static void
user_cont_signal (void)
{
  /* Advanced Programming in the UNIX Environment, Stevens; P321 */
  sigset_t mask;

#ifdef DEBUG
  if (trace) {
    fprintf(stderr, "USERPROC: In user_cont_signal()\n");  fflush(stderr);
  }
#endif

  /* unblock SIGCONT since it is blocked while we're handling it */
  sigemptyset (&mask);
  sigaddset (&mask, SIGCONT);
  sigprocmask (SIG_UNBLOCK, &mask, NULL);

  if (stdin_is_tty) {
    /* Save terminal state */
#ifdef  DEBUG
    if (trace) {
      fputs("USERPROC: Saving tty state\n", stderr); fflush(stderr);
    }
#endif
    save_tty_state ();
  }

  /* Do default SIGCONT action - continue the process */
  signal (SIGCONT, SIG_DFL);
  kill (getpid(), SIGCONT);
}


static void
set_user_process_signals (void)
{
  /* Signals used by KROC kernel */

  /* Modifies kernel kbd_sync variable */
  signal (SIGKBD, (SigParam)user_kbd_handler);
  /* Modifies kernel tim_sync variable */
  signal (SIGALRM, (SigParam)user_tim_handler);
#if 0
  /* Called by user if they panic */
  signal (SIGPANIC, (SigParam)user_panic_handler);
#endif

  /* Process termination signals */
  signal (SIGHUP, (SigParam)user_good_exit);  /* hangup */
#if 0
  signal (SIGEMT, (SigParam)user_bad_exit);  /* emulator trap */
#endif
  signal (SIGTERM, (SigParam)user_good_exit); /* software termination */

  /* Only catch if not being ignored */
  if (signal (SIGINT, SIG_IGN) != SIG_IGN)
    signal (SIGINT, (SigParam)user_good_exit);    /* interrupt (^C) */
  if (signal (SIGQUIT, SIG_IGN) != SIG_IGN)
    signal (SIGQUIT, (SigParam)user_good_exit);   /* quit */


  /* 'Error' signals */
  signal (SIGILL, (SigParam)user_trap_handler);  /* illegal instruction */
  signal (SIGBUS, (SigParam)user_trap_handler);  /* bus error */
  signal (SIGSEGV, (SigParam)user_trap_handler); /* segmentation violation */
#if defined(TARGET_CPU_ALPHA)
  signal (SIGTRAP, (SigParam)user_trap_handler);
#endif
  signal (SIGFPE, (SigParam)user_fp_handler);


  sigemptyset (&zero_mask);
  sigemptyset (&sig_mask);
  sigaddset (&sig_mask, SIGKBD);
  sigaddset (&sig_mask, SIGALRM);

  /* stop signal generated from keyboard (^Z) */
  /* Only catch this if running with job-control */
  if (signal (SIGTSTP, SIG_IGN) == SIG_DFL) {
    signal (SIGTSTP, (SigParam)user_stop_signal);
  }
}


void
impossible (int n, int occam_iptr)  /* (probably not) called from scheduler */
{       
  fprintf (stderr,
    "\nKROC: *** Impossible thing number %d happened at occam address %08X***\n",
           n, occam_iptr);
  fputs ("KROC: Please report thing number with your bug report\n", stderr);
  userproc_exit (4, 0);
} /* impossible */


void
status (int erfl)                   /* state of error flag */
{       
  if (erfl) {
    fputs ("KROC: Error flag set\n", stderr);
    fflush (stderr);
  }
} /* status */


void
dead (int erfl)                     /* called by scheduler */
{       
  fputs ("\nKROC: Program deadlocked (no processes to run)\n", stderr);
  fflush (stderr);              /* still in raw mode */
  status (erfl);
  userproc_exit (2, 0);
} /* dead */


unsigned int
utime (void)                            /* user (processor) time */
{
#if defined(HAVE_GETRUSAGE)
  struct rusage usage;          /* Resolution is 10000 usec */

  (void) getrusage (RUSAGE_SELF, &usage);
  return usage.ru_utime.tv_sec*1000000 + usage.ru_utime.tv_usec;
#elif defined(HAVE_TIMES)
  #include <sys/times.h>

  struct tms buffer;

  (void)times(&buffer);  /* Counts in 1/HZ seconds (where HZ is 60, for SunOS) */
  return buffer.tms_utime * (1000000 / HZ);
#else
  return rtime(); /* No processor time available */
#endif
} /* utime */


#if defined(HOST_OS_SUNOS)
unsigned int
rtime (void)                    /* real (clock) time    */
{
  struct timeval tp;            /* resolution 1 usec?   */
  (void) gettimeofday (&tp, 0);
  return tp.tv_sec*1000000 + tp.tv_usec;
} /* rtime */
#elif defined(HOST_OS_OSF3)
unsigned int
rtime (void)                    /* real (clock) time    */
{
  struct timespec ts;
  unsigned long msec;
  unsigned int occamtime;

  if (getclock(TIMEOFDAY, &ts)) {
    fprintf(stderr, "getclock(TIMEOFDAY) failed - ");
    perror(NULL);
    userproc_exit(1, 0);
  };
  /* Do not change this calculation - merging the lines will not work
     due to C's handling of int / unsigned int arithmetic */
  msec = ((unsigned int)ts.tv_nsec)/1000;
  msec += ts.tv_sec * 1000000;
  occamtime = (unsigned int)(msec & 0xffffffff);
  return occamtime;
} /* rtime */
#endif


/*
  This fixed constant of 10ms is added to each setitimer call to
  kludge (get round) a Solaris 2.3 (or later?) bug in returning too early.
  */
#if defined(HOST_OS_SOLARIS23)
#include <sys/param.h>
static int quantum; 
#endif


#ifndef HAVE_UALARM
#include <sys/time.h>
int
ualarm(int delay, int interval)
{
  struct itimerval itv;
#if defined(HOST_OS_SOLARIS23)
  delay += quantum; 
#endif
  itv.it_value.tv_sec =delay / 1000000;
  itv.it_value.tv_usec = delay % 1000000;

  itv.it_interval.tv_sec=0;
  itv.it_interval.tv_usec=interval;
  return setitimer(ITIMER_REAL, &itv, (struct itimerval*)NULL);
}
#endif /* HAVE_UALARM */


/* Called from assembler kernel */
int
safe_pause (void)
{       
  sigset_t old_mask;

  if (sigprocmask (SIG_BLOCK, &sig_mask, &old_mask) < 0) { /* critical region */
    fputs ("USERPROC: sigprocmask (SIG_BLOCK, ... , ...) failed.\n", stderr);
    userproc_exit (3, 0);
  }

#if defined(HOST_OS_OSF3)
    while ((kbd_sync & 1) == 0 || (tim_sync & 1) == 0)
#else
    while (combined_sync == 0)  /* all in one word now */
#endif
    { /* signal hasn't happened yet */
    if (sigsuspend (&zero_mask) != -1) {                /* wait for it */
      fputs ("USERPROC: sigsuspend (...) failed.\n", stderr);
      userproc_exit (3, 0);
    }

#if defined(HOST_OS_OSF3)
    if ((kbd_sync & 1) == 1 || (tim_sync & 1) == 1)
#else
    if (combined_sync != 0)  /* all in one word now */
#endif
      break;
  } /* while */


  if (sigprocmask (SIG_SETMASK, &old_mask, NULL) < 0) {   /* end of critical */
    fputs ("USERPROC: sigprocmask (SIG_SETMASK, ..., ...) failed.\n", stderr);
    userproc_exit (3, 0);
  }

  return 0;
} /* safe_pause */


void
user_process (int kbd_pid, int to_kbd_fd, int from_kbd_fd, bool is_a_tty)
{
  int erfl;

  /* user_process - run occam */

  kbdproc_pid = kbd_pid;
  to_kbdproc = to_kbd_fd;
  from_kbdproc = from_kbd_fd;
  stdin_is_tty = is_a_tty;

#if defined(HOST_OS_SOLARIS23)
  quantum=1000000/HZ;
#endif

  set_user_process_signals ();

  kbd_char_signalled = 0;
#ifdef DEBUG
  if (trace) {
    fputs ("USERPROC: About to synchronise with KBDPROC\n", stderr);
    fflush (stderr);
  }
#endif

  user_to_kbd_sync ();

#ifdef DEBUG
  if (trace) {
    fputs ("USERPROC: Synchronised ok\n", stderr);
    fflush (stderr);
  }
#endif

  /* really do the occam thing */
  erfl = enter ();              /* run the occam program */

  fflush (stdout);              /* clear buffer */
  status (erfl);
  userproc_exit (erfl, 0);

} /* user_process */
